/*---------------------------------------------------------------------------*/
/* Program:  binom.c                                                         */
/*                                                                           */
/* Purpose:  This program performs binomial smoothing of an image.  The      */
/*           number of smoothing iterations is set using the -n switch.      */
/*                                                                           */
/* Author:   John Gauch, Edu Metz                                            */
/*                                                                           */
/* Date:     April 20, 1994 - Original program.                              */
/*           July  23, 1996 - Color support added.                           */
/*           July 30, 1997  - Simplified color, vector, scalar code.         */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994-1996               */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Program variables */
int Debug = FALSE;
int Number = 3;

/* Image variables */
char Name1[50];
char Name2[50];
IM_TYPE *Image1;
IM_TYPE *Image2;
int PixType, Xdim, Ydim, Zdim, DimCnt;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs binomial smoothing of a scalar image.     */
/*---------------------------------------------------------------------------*/
void BinomScalar()
{
   FLOAT_TYPE **Data1;
   FLOAT_TYPE **Data2;
   FLOAT_TYPE **Temp;
   int i, x, y;

   /* Read input image */
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Perform binomial smoothing */
   for (i = 0; i < Number; i++)
   {
      /* Average pixels */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	    Data2[y][x] = (Data1[y - 1][x - 1] + Data1[y - 1][x + 1]
			   + Data1[y + 1][x - 1] + Data1[y + 1][x + 1]
			   + (Data1[y - 1][x] + Data1[y + 1][x]
			      + Data1[y][x - 1] + Data1[y][x + 1]) * 2
			   + Data1[y][x] * 4) / 16;

      /* Fix boundary */
      for (y = 0; y < Ydim; y++)
      {
	 Data2[y][0] = Data1[y][1];
	 Data2[y][Xdim - 1] = Data1[y][Xdim - 2];
      }
      for (x = 0; x < Xdim; x++)
      {
	 Data2[0][x] = Data1[1][x];
	 Data2[Ydim - 1][x] = Data1[Ydim - 2][x];
      }

      Temp = Data1;
      Data1 = Data2;
      Data2 = Temp;
   }

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data1[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs binomial smoothing of a vector image.     */
/*---------------------------------------------------------------------------*/
void BinomVector()
{
   FLOAT_TYPE ***Data1;
   FLOAT_TYPE ***Data2;
   FLOAT_TYPE ***Temp;
   int i, x, y, z;

   /* Read input image */
   Data1 = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE ***) im_alloc3D(Image2, FLOAT);

   /* Perform binomial smoothing */
   for (i = 0; i < Number; i++)
   {
      /* Average pixels */
      for (z = 0; z < Zdim; z++)
	 for (y = 1; y < (Ydim - 1); y++)
	    for (x = 1; x < (Xdim - 1); x++)
	       Data2[z][y][x] = (Data1[z][y - 1][x - 1] + Data1[z][y - 1][x + 1]
			   + Data1[z][y + 1][x - 1] + Data1[z][y + 1][x + 1]
				 + (Data1[z][y - 1][x] + Data1[z][y + 1][x]
			      + Data1[z][y][x - 1] + Data1[z][y][x + 1]) * 2
				 + Data1[z][y][x] * 4) / 16;

      /* Fix boundary */
      for (z = 0; z < Zdim; z++)
	 for (y = 0; y < Ydim; y++)
	 {
	    Data2[z][y][0] = Data1[z][y][1];
	    Data2[z][y][Xdim - 1] = Data1[z][y][Xdim - 2];
	 }
      for (z = 0; z < Zdim; z++)
	 for (x = 0; x < Xdim; x++)
	 {
	    Data2[z][0][x] = Data1[z][1][x];
	    Data2[z][Ydim - 1][x] = Data1[z][Ydim - 2][x];
	 }

      Temp = Data1;
      Data1 = Data2;
      Data2 = Temp;
   }

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data1[0][0][0]));
   im_free3D((char ***) Data1);
   im_free3D((char ***) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs binomial smoothing of a color image.      */
/*---------------------------------------------------------------------------*/
void BinomColor()
{
   COLOR_TYPE **Data1;
   COLOR_TYPE **Data2;
   COLOR_TYPE **Temp;
   int i, x, y, temp;

   /* Read input image */
   Data1 = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (COLOR_TYPE **) im_alloc2D(Image2, COLOR);

   /* Perform binomial smoothing */
   for (i = 0; i < Number; i++)
   {
      /* Average pixels */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    temp = ((int) Data1[y - 1][x - 1].r + (int) Data1[y - 1][x + 1].r
		 + (int) Data1[y + 1][x - 1].r + (int) Data1[y + 1][x + 1].r
		    + ((int) Data1[y - 1][x].r + (int) Data1[y + 1][x].r
		    + (int) Data1[y][x - 1].r + (int) Data1[y][x + 1].r) * 2
		    + (int) Data1[y][x].r * 4) / 16;
	    if (temp < 0)
	       temp = 0;
	    else if (temp > 255)
	       temp = 255;
	    Data2[y][x].r = temp;

	    temp = ((int) Data1[y - 1][x - 1].g + (int) Data1[y - 1][x + 1].g
		 + (int) Data1[y + 1][x - 1].g + (int) Data1[y + 1][x + 1].g
		    + ((int) Data1[y - 1][x].g + (int) Data1[y + 1][x].g
		    + (int) Data1[y][x - 1].g + (int) Data1[y][x + 1].g) * 2
		    + (int) Data1[y][x].g * 4) / 16;
	    if (temp < 0)
	       temp = 0;
	    else if (temp > 255)
	       temp = 255;
	    Data2[y][x].g = temp;

	    temp = ((int) Data1[y - 1][x - 1].b + (int) Data1[y - 1][x + 1].b
		 + (int) Data1[y + 1][x - 1].b + (int) Data1[y + 1][x + 1].b
		    + ((int) Data1[y - 1][x].b + (int) Data1[y + 1][x].b
		    + (int) Data1[y][x - 1].b + (int) Data1[y][x + 1].b) * 2
		    + (int) Data1[y][x].b * 4) / 16;
	    if (temp < 0)
	       temp = 0;
	    else if (temp > 255)
	       temp = 255;
	    Data2[y][x].b = temp;
	 }

      /* Fix boundary */
      for (y = 0; y < Ydim; y++)
      {
	 Data2[y][0].r = Data1[y][1].r;
	 Data2[y][Xdim - 1].r = Data1[y][Xdim - 2].r;
	 Data2[y][0].g = Data1[y][1].g;
	 Data2[y][Xdim - 1].g = Data1[y][Xdim - 2].g;
	 Data2[y][0].b = Data1[y][1].b;
	 Data2[y][Xdim - 1].b = Data1[y][Xdim - 2].b;
      }

      for (x = 0; x < Xdim; x++)
      {
	 Data2[0][x].r = Data1[1][x].r;
	 Data2[Ydim - 1][x].r = Data1[Ydim - 2][x].r;
	 Data2[0][x].g = Data1[1][x].g;
	 Data2[Ydim - 1][x].g = Data1[Ydim - 2][x].g;
	 Data2[0][x].b = Data1[1][x].b;
	 Data2[Ydim - 1][x].b = Data1[Ydim - 2][x].b;
      }

      Temp = Data1;
      Data1 = Data2;
      Data2 = Temp;
   }

   /* Write information to output image */
   im_write(Image2, COLOR, (char *) &(Data1[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("BINOM Program - KUIM Version 2.1\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'n':
	 if (sscanf(argv[++i], "%d", &Number) == 0)
	    Error("Could not get number of iterations");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: binom [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-n #]  Number of smoothing iterations (3)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Determine PixType */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);

   /* Handle color images */
   if (PixType == COLOR || PixType == PSEUDO || PixType == JPEG_RGB)
      BinomColor();

   /* Handle scalar images */
   else if (DimCnt == 2)
      BinomScalar();

   /* Handle vector images */
   else if (DimCnt == 3)
      BinomVector();

   /* Unsupported image type */
   else
      Error("Can not process image");
   return (0);
}
