/*---------------------------------------------------------------------------*/
/* Program:  blob_match.c                                                    */
/*                                                                           */
/* Purpose:  This program reads two blob colored images and uses             */
/*           "templates" from the first image to match "objects" in the      */
/*           second image.  The program outputs an image with the "matched"  */
/*           objects in their appropriate positions.  Both statistics-based  */
/*           and correlation-based matching are supported.                   */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     February 22, 1994 - Original program.                           */
/*           August 2, 1997    - Fixed bug in template match.                */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define BLOBS 1000
double sqrt();

/* Global program variables */
SHORT_TYPE **Data1;
SHORT_TYPE **Data2;
SHORT_TYPE **Data3;
int Debug = FALSE;
int Statistics = TRUE;
int Template = FALSE;
int Threshold = 10000;
int Search = 0;
int Xdim1, Ydim1, Xdim2, Ydim2;

/* Bounding rectangle and statistics for template blobs */
int Tdim = 0;
int Tcnt[BLOBS];
int TminX[BLOBS];
int TmaxX[BLOBS];
int TminY[BLOBS];
int TmaxY[BLOBS];
double TmeanX[BLOBS];
double TmeanY[BLOBS];
double TmomXX[BLOBS];
double TmomXY[BLOBS];
double TmomYY[BLOBS];
double TmomXXX[BLOBS];
double TmomXXY[BLOBS];
double TmomXYY[BLOBS];
double TmomYYY[BLOBS];

/* Bounding rectangle and statistics for object blobs */
int Mdim = 0;
int Mcnt[BLOBS];
int MminX[BLOBS];
int MmaxX[BLOBS];
int MminY[BLOBS];
int MmaxY[BLOBS];
double MmeanX[BLOBS];
double MmeanY[BLOBS];
double MmomXX[BLOBS];
double MmomXY[BLOBS];
double MmomYY[BLOBS];
double MmomXXX[BLOBS];
double MmomXXY[BLOBS];
double MmomXYY[BLOBS];
double MmomYYY[BLOBS];
double AmomXX, AmomXY, AmomYY, AmomXXX, AmomXYY, AmomXXY, AmomYYY;
double VmomXX, VmomXY, VmomYY, VmomXXX, VmomXYY, VmomXXY, VmomYYY;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine gets the bounds for each blob.                     */
/*---------------------------------------------------------------------------*/
void GetBounds()
{
   int t, m, x, y;

   /* Initialize bounds and statistics arrays */
   for (t = 0; t < BLOBS; t++)
   {
      TminX[t] = Xdim1;
      TminY[t] = Ydim1;
      TmaxX[t] = TmaxY[t] = Tcnt[t] = 0;
      TmeanX[t] = TmeanY[t] = 0.0;
      TmomXX[t] = TmomXY[t] = TmomYY[t] = 0.0;
      TmomXXX[t] = TmomXXY[t] = TmomXYY[t] = TmomYYY[t] = 0.0;
   }
   for (m = 0; m < BLOBS; m++)
   {
      MminX[m] = Xdim2;
      MminY[m] = Ydim2;
      MmaxX[m] = MmaxY[m] = Mcnt[m] = 0;
      MmeanX[m] = MmeanY[m] = 0.0;
      MmomXX[m] = MmomXY[m] = MmomYY[m] = 0.0;
      MmomXXX[m] = MmomXXY[m] = MmomXYY[m] = MmomYYY[m] = 0.0;
   }

   /* Calculate bounding rectangles in template image */
   for (y = 0; y < Ydim1; y++)
      for (x = 0; x < Xdim1; x++)
	 if (Data1[y][x] > 0)
	 {
	    t = Data1[y][x];
	    if (t > Tdim)
	       Tdim = t;
	    if (x < TminX[t])
	       TminX[t] = x;
	    if (y < TminY[t])
	       TminY[t] = y;
	    if (x > TmaxX[t])
	       TmaxX[t] = x;
	    if (y > TmaxY[t])
	       TmaxY[t] = y;
	    TmeanX[t] += x;
	    TmeanY[t] += y;
	    Tcnt[t]++;
	 }

   /* Calculate bounding rectangles in object image */
   for (y = 0; y < Ydim2; y++)
      for (x = 0; x < Xdim2; x++)
	 if (Data2[y][x] > 0)
	 {
	    m = Data2[y][x];
	    if (m > Mdim)
	       Mdim = m;
	    if (x < MminX[m])
	       MminX[m] = x;
	    if (y < MminY[m])
	       MminY[m] = y;
	    if (x > MmaxX[m])
	       MmaxX[m] = x;
	    if (y > MmaxY[m])
	       MmaxY[m] = y;
	    MmeanX[m] += x;
	    MmeanY[m] += y;
	    Mcnt[m]++;
	 }

   /* Normalize 1st moments */
   for (t = 0; t <= Tdim; t++)
      if (Tcnt[t] > 0)
      {
	 TmeanX[t] /= Tcnt[t];
	 TmeanY[t] /= Tcnt[t];
      }
   for (m = 1; m <= Mdim; m++)
      if (Mcnt[m] > 0)
      {
	 MmeanX[m] /= Mcnt[m];
	 MmeanY[m] /= Mcnt[m];
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine gets the statistics for each blob.                 */
/*---------------------------------------------------------------------------*/
void GetStats()
{
   int t, m, x, y;
   double X, Y;

   /* Calculate 2nd and 3rd moments in template image */
   for (y = 0; y < Ydim1; y++)
      for (x = 0; x < Xdim1; x++)
	 if (Data1[y][x] > 0)
	 {
	    t = Data1[y][x];
	    X = x - TmeanX[t];
	    Y = y - TmeanY[t];
	    TmomXX[t] += X * X;
	    TmomXY[t] += X * Y;
	    TmomYY[t] += Y * Y;
	    TmomXXX[t] += X * X * X;
	    TmomXXY[t] += X * X * Y;
	    TmomXYY[t] += X * Y * Y;
	    TmomYYY[t] += Y * Y * Y;
	 }

   /* Calculate 2nd and 3rd moments in object image */
   for (y = 0; y < Ydim2; y++)
      for (x = 0; x < Xdim2; x++)
	 if (Data2[y][x] > 0)
	 {
	    m = Data2[y][x];
	    X = x - MmeanX[m];
	    Y = y - MmeanY[m];
	    MmomXX[m] += X * X;
	    MmomXY[m] += X * Y;
	    MmomYY[m] += Y * Y;
	    MmomXXX[m] += X * X * X;
	    MmomXXY[m] += X * X * Y;
	    MmomXYY[m] += X * Y * Y;
	    MmomYYY[m] += Y * Y * Y;
	 }

   /* Normalize 2nd and 3rd moments */
   for (t = 0; t <= Tdim; t++)
      if (Tcnt[t] > 0)
      {
	 TmomXX[t] /= Tcnt[t];
	 TmomXY[t] /= Tcnt[t];
	 TmomYY[t] /= Tcnt[t];
	 TmomXXX[t] /= Tcnt[t];
	 TmomXXY[t] /= Tcnt[t];
	 TmomXYY[t] /= Tcnt[t];
	 TmomYYY[t] /= Tcnt[t];
      }
   for (m = 0; m <= Mdim; m++)
      if (Mcnt[m] > 0)
      {
	 MmomXX[m] /= Mcnt[m];
	 MmomXY[m] /= Mcnt[m];
	 MmomYY[m] /= Mcnt[m];
	 MmomXXX[m] /= Mcnt[m];
	 MmomXXY[m] /= Mcnt[m];
	 MmomXYY[m] /= Mcnt[m];
	 MmomYYY[m] /= Mcnt[m];
      }

   /* Calculate mean of template stats */
   AmomXX = AmomXY = AmomYY = AmomXXX = AmomXYY = AmomXXY = AmomYYY = 0.0;
   for (t = 0; t <= Tdim; t++)
   {
      AmomXX += TmomXX[t];
      AmomXY += TmomXY[t];
      AmomYY += TmomYY[t];
      AmomXXX += TmomXXX[t];
      AmomXXY += TmomXXY[t];
      AmomXYY += TmomXYY[t];
      AmomYYY += TmomYYY[t];
   }
   AmomXX /= Tdim;
   AmomXY /= Tdim;
   AmomYY /= Tdim;
   AmomXXX /= Tdim;
   AmomXXY /= Tdim;
   AmomXYY /= Tdim;
   AmomYYY /= Tdim;

   /* Calculate variance of template stats */
   VmomXX = VmomXY = VmomYY = VmomXXX = VmomXYY = VmomXXY = VmomYYY = 0.0;
   for (t = 0; t <= Tdim; t++)
   {
      VmomXX += (AmomXX - TmomXX[t]) * (AmomXX - TmomXX[t]);
      VmomXY += (AmomXY - TmomXY[t]) * (AmomXY - TmomXY[t]);
      VmomYY += (AmomYY - TmomYY[t]) * (AmomYY - TmomYY[t]);
      VmomXXX += (AmomXXX - TmomXXX[t]) * (AmomXXX - TmomXXX[t]);
      VmomXXY += (AmomXXY - TmomXXY[t]) * (AmomXXY - TmomXXY[t]);
      VmomXYY += (AmomXYY - TmomXYY[t]) * (AmomXYY - TmomXYY[t]);
      VmomYYY += (AmomYYY - TmomYYY[t]) * (AmomYYY - TmomYYY[t]);
   }
   VmomXX = sqrt(VmomXX / Tdim);
   VmomXY = sqrt(VmomXY / Tdim);
   VmomYY = sqrt(VmomYY / Tdim);
   VmomXXX = sqrt(VmomXXX / Tdim);
   VmomXXY = sqrt(VmomXXY / Tdim);
   VmomXYY = sqrt(VmomXYY / Tdim);
   VmomYYY = sqrt(VmomYYY / Tdim);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine matches blob statistics.                           */
/*---------------------------------------------------------------------------*/
int StatMatch(int m, int t)
{
   double Ttemp, Mtemp, Match = 0.0;

   Ttemp = (TmomXX[t] - AmomXX) / VmomXX;
   Mtemp = (MmomXX[m] - AmomXX) / VmomXX;
   Match += (Ttemp - Mtemp) * (Ttemp - Mtemp);

   Ttemp = (TmomXY[t] - AmomXY) / VmomXY;
   Mtemp = (MmomXY[m] - AmomXY) / VmomXY;
   Match += (Ttemp - Mtemp) * (Ttemp - Mtemp);

   Ttemp = (TmomYY[t] - AmomYY) / VmomYY;
   Mtemp = (MmomYY[m] - AmomYY) / VmomYY;
   Match += (Ttemp - Mtemp) * (Ttemp - Mtemp);

   Ttemp = (TmomXXX[t] - AmomXXX) / VmomXXX;
   Mtemp = (MmomXXX[m] - AmomXXX) / VmomXXX;
   Match += (Ttemp - Mtemp) * (Ttemp - Mtemp);

   Ttemp = (TmomXXY[t] - AmomXXY) / VmomXXY;
   Mtemp = (MmomXXY[m] - AmomXXY) / VmomXXY;
   Match += (Ttemp - Mtemp) * (Ttemp - Mtemp);

   Ttemp = (TmomXYY[t] - AmomXYY) / VmomXYY;
   Mtemp = (MmomXYY[m] - AmomXYY) / VmomXYY;
   Match += (Ttemp - Mtemp) * (Ttemp - Mtemp);

   Ttemp = (TmomYYY[t] - AmomYYY) / VmomYYY;
   Mtemp = (MmomYYY[m] - AmomYYY) / VmomYYY;
   Match += (Ttemp - Mtemp) * (Ttemp - Mtemp);

   /* Return match value */
   return ((int) (1000 * Match));
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine matches blob templates.                            */
/*---------------------------------------------------------------------------*/
int TempMatch(int m, int t)	/* JOHN */
{
   int dx, dy;
   int x1, y1, x2, y2;
   int cx1, cy1, cx2, cy2;
   int Match = 0, Miss = 0;
   float Result = 0, Best = 0;

   /* Calculate center of templates */
   cx1 = (TmaxX[t] + TminX[t]) / 2;
   cy1 = (TmaxY[t] + TminY[t]) / 2;
   cx2 = (MmaxX[m] + MminX[m]) / 2;
   cy2 = (MmaxY[m] + MminY[m]) / 2;

   /* Loop trying different alignments of templates */
   for (dy = -Search; dy <= Search; dy++)
      for (dx = -Search; dx <= Search; dx++)
      {
	 /* Compare two templates */
	 Match = Miss = 0;
	 y2 = cy2 - cy1 + TminY[t] + dy;
	 for (y1 = TminY[t]; y1 <= TmaxY[t]; y1++, y2++)
	 {
	    x2 = cx2 - cx1 + TminX[t] + dx;
	    for (x1 = TminX[t]; x1 <= TmaxX[t]; x1++, x2++)
	    {
	       if ((x2 >= MminX[m]) && (x2 <= MmaxX[m]) &&
		   (y2 >= MminY[m]) && (y2 <= MmaxY[m]))
	       {
		  if ((Data1[y1][x1] == t) && (Data2[y2][x2] == m))
		     Match++;
	       }
	    }
	 }

	 /* Compute percentage match */
	 Miss = Tcnt[t] - Match;
	 Result = (Match - Miss) / (float) Mcnt[m];
	 if (Result > Best)
	    Best = Result;
      }

   /* Return match value */
   return ((int) ((1 - Best) * 1000));
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs blob matching.                            */
/*---------------------------------------------------------------------------*/
void MyMain()
{
   int m, t, BestBlob;
   int Match, BestMatch;
   int x1, y1, x3, y3;

   /* Initialize output image */
   for (y3 = 0; y3 < Ydim2; y3++)
      for (x3 = 0; x3 < Xdim2; x3++)
	 Data3[y3][x3] = 0;

   /* Preprocess image to get blob information */
   GetBounds();
   if (Statistics == TRUE)
      GetStats();

   /* Compare each template blob to each object blob */
   for (m = 1; m <= Mdim; m++)
      if (Mcnt[m] > 100)
      {
	 /* Search for highest quality match */
	 BestBlob = 1;
	 if (Statistics == TRUE)
	    BestMatch = StatMatch(m, 1);
	 else if (Template == TRUE)
	    BestMatch = TempMatch(m, 1);
	 for (t = 2; t <= Tdim; t++)
	 {
	    if (Statistics == TRUE)
	       Match = StatMatch(m, t);
	    else if (Template == TRUE)
	       Match = TempMatch(m, t);
	    if (Match < BestMatch)
	    {
	       BestMatch = Match;
	       BestBlob = t;
	    }
	 }

	 /* Print debug info */
	 if (Debug == TRUE)
	    printf("object %d matches %d error %d\n", m, BestBlob, BestMatch);

	 /* Save best match in output image */
	 if (BestMatch < Threshold)
	 {
	    y3 = (int) (MmeanY[m] - TmeanY[BestBlob] + TminY[BestBlob]);
	    for (y1 = TminY[BestBlob]; y1 <= TmaxY[BestBlob]; y1++, y3++)
	    {
	       x3 = (int) (MmeanX[m] - TmeanX[BestBlob] + TminX[BestBlob]);
	       for (x1 = TminX[BestBlob]; x1 <= TmaxX[BestBlob]; x1++, x3++)
		  if ((x3 >= 0) && (x3 < Xdim2) && (y3 >= 0) && (y3 < Ydim2) &&
		      (Data1[y1][x1] == BestBlob))
		     Data3[y3][x3] = -1;
	    }
	 }
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   int PixType, Zdim, DimCnt;

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("BLOB_MATCH Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'r':
	 if (sscanf(argv[++i], "%d", &Search) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'm':
	 if (sscanf(argv[++i], "%d", &Threshold) == 0)
	    Error("Could not get integer argument");
	 break;
      case 's':
	 Statistics = TRUE;
	 Template = FALSE;
	 break;
      case 't':
	 Template = TRUE;
	 Statistics = FALSE;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr,
	   "Usage: blob_match [options] templatefile objectfile outfile\n");
      fprintf(stderr, "       [-d]  Print debugging information\n");
      fprintf(stderr, "       [-s]  Statistics based matching\n");
      fprintf(stderr, "       [-t]  Template based matching\n");
      fprintf(stderr, "       [-r #]  Radius of template search region (0)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name3) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim1, &Ydim1, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   im_read(Image1, SHORT, (char *) &(Data1[0][0]));

   /* Read input image */
   Image2 = im_open(Name2, &PixType, &Xdim2, &Ydim2, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data2 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);
   im_read(Image2, SHORT, (char *) &(Data2[0][0]));

   /* Create output image */
   Image3 = im_create(Name3, SHORT, Xdim2, Ydim2, Zdim);
   Data3 = (SHORT_TYPE **) im_alloc2D(Image3, SHORT);
   MyMain();

   /* Write information to output image */
   im_write(Image3, SHORT, (char *) &(Data3[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   im_free2D((char **) Data3);
   return (0);
}
