/*---------------------------------------------------------------------------*/
/* Program:  cell_count.c                                                    */
/*                                                                           */
/* Purpose:  This program processes a sequence of images to count            */
/*           the number of cells which have moved out of the field of        */
/*           view.  This is done by (1) subtracting one image from an        */
/*           image taken some time later, (2) smoothing to reduce noise,     */
/*           (3) thresholding to identify areas of significant change,       */
/*           and finally (4) counting the number of blobs above a given      */
/*           size threshold.  These should correspond to cells.              */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     October 24, 1995                                                */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Global variables */
int Size = 0;
int Xdim = 0;
int Ydim = 0;
FLOAT_TYPE **Data2;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine identifies a connected blob.                       */
/*---------------------------------------------------------------------------*/
void BlobColor(int x, int y, int Color)
{
   if ((Data2[y][x] > 0) && (Data2[y][x] != Color))
   {
      Size++;
      Data2[y][x] = (float) Color;
      if (x + 1 < Xdim)
	 BlobColor(x + 1, y, Color);
      if (x > 0)
	 BlobColor(x - 1, y, Color);
      if (y + 1 < Ydim)
	 BlobColor(x, y + 1, Color);
      if (y > 0)
	 BlobColor(x, y - 1, Color);
      if ((x + 1 < Xdim) && (y + 1 < Ydim))
	 BlobColor(x + 1, y + 1, Color);
      if ((x > 0) && (y > 0))
	 BlobColor(x - 1, y - 1, Color);
      if ((x > 0) && (y + 1 < Ydim))
	 BlobColor(x - 1, y + 1, Color);
      if ((x + 1 < Xdim) && (y > 0))
	 BlobColor(x + 1, y - 1, Color);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   FLOAT_TYPE **Data1;
   int PixType, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int AddedCells = FALSE;
   int i = 0, x, y;
   int Offset = 10;
   int Minimum = 10;
   int Threshold = 80;
   int First = 0;
   int Last = 1000;
   int Count[5000];
   int Index = 0;

   /* Interpret program options */
   fprintf(stderr, "CELL_COUNT Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'o':
	 if (sscanf(argv[++i], "%d", &Offset) == 0)
	    Error("Could not get Offset argument");
	 break;
      case 'm':
	 if (sscanf(argv[++i], "%d", &Minimum) == 0)
	    Error("Could not get Minimum argument");
	 break;
      case 't':
	 if (sscanf(argv[++i], "%d", &Threshold) == 0)
	    Error("Could not get Threshold argument");
	 break;
      case 'r':
	 if (sscanf(argv[++i], "%d", &First) == 0)
	    Error("Could not get First argument");
	 if (sscanf(argv[++i], "%d", &Last) == 0)
	    Error("Could not get Last argument");
	 break;
      case 'a':
	 AddedCells = TRUE;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i > 1)
   {
      fprintf(stderr, "Usage: cell_count [options]\n");
      fprintf(stderr, "\t[-d]      Print debugging information\n");
      fprintf(stderr, "\t[-a]      Check for cells which are added\n");
      fprintf(stderr, "\t[-o #]    Time offset between images (%d frames)\n",
	      Offset);
      fprintf(stderr, "\t[-m #]    Minimum blob size (%d pixels)\n", Minimum);
      fprintf(stderr, "\t[-t #]    Blob detection threshold (%d)\n", Threshold);
      fprintf(stderr, "\t[-r # #]  Range of images to consider (%d to %d)\n",
	      First, Last);
      exit(1);
   }

   /* Get file prefix from argument list */
   /* if (sscanf(argv[i++],"%s",Name0)==0)
      Error("Could not get input file prefix"); */

   /* Loop processing sequence of images */
   for (i = First + Offset; i <= Last; i = i + Offset)
   {
      /* Generating file names */
      sprintf(Name1, "%04d.jpg", i - Offset);
      sprintf(Name2, "%04d.jpg", i);
      sprintf(Name3, "out.%04d.%04d", i - Offset, i);
      if (Debug == TRUE)
	 printf("processing %s and %s\n", Name1, Name2);

      /* Read first image */
      Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
      if (DimCnt != 2)
	 Error("Can not process 1D or 3D images");
      Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
      im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

      /* Read second image */
      Image2 = im_open(Name2, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
      if (DimCnt != 2)
	 Error("Can not process 1D or 3D images");
      Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
      im_read(Image2, FLOAT, (char *) &(Data2[0][0]));

      /* Calculate difference image */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	    Data2[y][x] -= Data1[y][x];

      /* Reverse image if checking for added cells */
      if (AddedCells == TRUE)
	 for (y = 0; y < Ydim; y++)
	    for (x = 0; x < Xdim; x++)
	       Data2[y][x] = -Data2[y][x];

      /* Smooth image with 3x3 averaging */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	    Data2[y][x] = (Data2[y - 1][x - 1] + Data2[y - 1][x] + Data2[y - 1][x + 1]
			   + Data2[y][x - 1] + Data2[y][x] + Data2[y][x + 1]
			   + Data2[y + 1][x - 1] + Data2[y + 1][x] + Data2[y + 1][x + 1]) / 9;

      /* Threshold the image */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	    if (Data2[y][x] >= Threshold)
	       Data2[y][x] = 1;
	    else
	       Data2[y][x] = 0;

      /* Clean up bottom of image */
      for (y = Ydim - 10; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	    Data2[y][x] = 0;

      /* Locate and count blobs */
      Count[Index] = 0;
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	    if (Data2[y][x] == 1)
	    {
	       Size = 0;
	       BlobColor(x, y, 2);
	       if (Size > Minimum)
	       {
		  if (Debug == TRUE)
		     printf("cell %d %d size %d\n", x, y, Size);
		  Count[Index]++;
	       }
	       else
		  BlobColor(x, y, 0);
	    }
      Index++;

      /* Create output image */
      if (Debug == TRUE)
      {
	 /* Mark cell boundary */
	 for (y = 1; y < Ydim; y++)
	    for (x = 1; x < Xdim; x++)
	       if ((Data2[y][x] != Data2[y - 1][x]) ||
		   (Data2[y][x] != Data2[y][x - 1]))
		  Data1[y][x] = 0;

	 /* Write image */
	 if (Debug == TRUE)
	    printf("generating %s\n", Name3);
	 Image3 = im_create(Name3, JPEG_GRAY, Xdim, Ydim, Zdim);
	 im_write(Image3, FLOAT, (char *) &(Data1[0][0]));
      }

      /* Free image space */
      im_free2D((char **) Data1);
      im_free2D((char **) Data2);
   }

   /* Output cell counting results */
   for (i = 0; i < Index; i++)
      printf("%5d %7d\n", i, Count[i]);
   return (0);
}
