/*---------------------------------------------------------------------------*/
/* Program:  complex2float.c                                                 */
/*                                                                           */
/* Purpose:  This program converts a complex KUIM image into a float         */
/*           KUIM image containing either the real, imaginary, amplitude     */
/*           or phase of the complex value.                                  */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     April 6, 1994                                                   */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

#define LOG_AMPLITUDE   0
#define AMPLITUDE   1
#define PHASE      2
#define REAL      3
#define IMAGINARY   4

double sqrt();
double log();
double atan2();

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   COMPLEX_TYPE *Data1;
   FLOAT_TYPE *Data2;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Count;
   int Output = LOG_AMPLITUDE;
   int Debug = FALSE;
   int i = 0;

   /* Interpret program options */
   printf("COMPLEX2FLOAT Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'A':
	 Output = LOG_AMPLITUDE;
	 break;
      case 'a':
	 Output = AMPLITUDE;
	 break;
      case 'p':
	 Output = PHASE;
	 break;
      case 'r':
	 Output = REAL;
	 break;
      case 'i':
	 Output = IMAGINARY;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: complex2float [options] infile outfile\n");
      fprintf(stderr, "       [-d]  Print debugging information\n");
      fprintf(stderr, "       [-A]  Output LOG AMPLITUDE\n");
      fprintf(stderr, "       [-a]  Output AMPLITUDE\n");
      fprintf(stderr, "       [-p]  Output PHASE\n");
      fprintf(stderr, "       [-r]  Output REAL part\n");
      fprintf(stderr, "       [-i]  Output IMAGINARY part\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   Data1 = (COMPLEX_TYPE *) im_alloc1D(Image1, COMPLEX);
   im_read(Image1, COMPLEX, (char *) Data1);

   /* Create output image */
   Image2 = im_create(Name2, FLOAT, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE *) im_alloc1D(Image2, FLOAT);
   Count = Xdim * Ydim * Zdim;

   /* Convert complex to float */
   if (Output == REAL)
      for (i = 0; i < Count; i++)
	 Data2[i] = (FLOAT_TYPE) Data1[i].re;

   else if (Output == IMAGINARY)
      for (i = 0; i < Count; i++)
	 Data2[i] = (FLOAT_TYPE) Data1[i].im;

   else if (Output == LOG_AMPLITUDE)
      for (i = 0; i < Count; i++)
	 Data2[i] = (FLOAT_TYPE) (log(1.0 + sqrt((double)
		 (Data1[i].re * Data1[i].re + Data1[i].im * Data1[i].im))));

   else if (Output == AMPLITUDE)
      for (i = 0; i < Count; i++)
	 Data2[i] = (FLOAT_TYPE) (sqrt((double)
		  (Data1[i].re * Data1[i].re + Data1[i].im * Data1[i].im)));

   else if (Output == PHASE)
      for (i = 0; i < Count; i++)
	 Data2[i] = (FLOAT_TYPE) atan2((double) Data1[i].im, (double) Data1[i].im);

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) Data2);
   im_free1D((char *) Data1);
   im_free1D((char *) Data2);
   return (0);
}
