/*---------------------------------------------------------------------------*/
/* Purpose:  This program performs general purpose image convolution.  The   */
/*           convolution mask is read from an ascii data file of the form:   */
/*                                                                           */
/*           Rows Cols                                                       */
/*           m[0][0] m[0][1] ... m[0][Cols-1]                                */
/*           m[1][0] m[1][1] ... m[1][Cols-1]                                */
/*           ...     ...     ...     ...                                     */
/*           m[Rows-1][0] m[Rows-1][1] ... m[Rows-1][Cols-1]                 */
/*                                                                           */
/*           The convolution mask is centered at the origin by shifting      */
/*           Rows/2 pixels left and Cols/2 pixels up.                        */
/*                                                                           */
/* Author:   John Gauch, Edu Metz                                            */
/*                                                                           */
/* Date:     April 2, 1994 - Original program.                               */
/*           July 27, 1994 - Color support added.                            */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

   /* Image variables */
char Name0[50];
char Name1[50];
char Name2[50];
IM_TYPE *Image0;
IM_TYPE *Image1;
IM_TYPE *Image2;
int PixType, Xdim, Ydim, Zdim, DimCnt;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does 2D convolve.                                  */
/*---------------------------------------------------------------------------*/
void GreyConvolve2d(int Rows, int Cols, FLOAT_TYPE ** Data0)
{
   FLOAT_TYPE **Data1;
   FLOAT_TYPE **Data2;
   int xshift, yshift;
   int i, j, x, y, X, Y;
   float Total;

   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, FLOAT, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Perform image convolution */
   xshift = Cols / 2;
   yshift = Rows / 2;
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Total = 0.0;
	 for (i = 0; i < Rows; i++)
	    for (j = 0; j < Cols; j++)
	    {
	       X = x - j + xshift;
	       Y = y - i + yshift;
	       if (X < 0 || X >= Xdim)
		  X = ((X % Xdim) + Xdim) % Xdim;
	       if (Y < 0 || Y >= Ydim)
		  Y = ((Y % Ydim) + Ydim) % Ydim;
	       Total += Data1[Y][X] * Data0[i][j];
	    }
	 Data2[y][x] = Total;
      }

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does 3D convolve.                                  */
/*---------------------------------------------------------------------------*/
void GreyConvolve3d(int Rows, int Cols, FLOAT_TYPE ** Data0)
{
   FLOAT_TYPE ***Data1;
   FLOAT_TYPE ***Data2;
   int xshift, yshift;
   int i, j, x, y, z, X, Y;
   float Total;

   Data1 = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE ***) im_alloc3D(Image2, FLOAT);

   /* Perform image convolution */
   xshift = Cols / 2;
   yshift = Rows / 2;
   for (z = 0; z < Zdim; z++)
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    Total = 0.0;
	    for (i = 0; i < Rows; i++)
	       for (j = 0; j < Cols; j++)
	       {
		  X = x - j + xshift;
		  Y = y - i + yshift;
		  if (X < 0 || X >= Xdim)
		     X = ((X % Xdim) + Xdim) % Xdim;
		  if (Y < 0 || Y >= Ydim)
		     Y = ((Y % Ydim) + Ydim) % Ydim;
		  Total += Data1[z][Y][X] * Data0[i][j];
	       }
	    Data2[z][y][x] = Total;
	 }

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0][0]));
   im_free3D((char ***) Data1);
   im_free3D((char ***) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does 2D color convolve.                            */
/*---------------------------------------------------------------------------*/
void ColorConvolve2d(int Rows, int Cols, FLOAT_TYPE ** Data0)
{
   COLOR_TYPE **Data1;
   COLOR_TYPE **Data2;
   int xshift, yshift;
   int i, j, x, y, X, Y;
   float TotalRed, TotalGreen, TotalBlue;

   Data1 = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, COLOR, Xdim, Ydim, Zdim);
   Data2 = (COLOR_TYPE **) im_alloc2D(Image2, COLOR);

   /* Perform image convolution */
   xshift = Cols / 2;
   yshift = Rows / 2;
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 TotalRed = TotalGreen = TotalBlue = 0.0;
	 for (i = 0; i < Rows; i++)
	    for (j = 0; j < Cols; j++)
	    {
	       X = x - j + xshift;
	       Y = y - i + yshift;
	       if (X < 0 || X >= Xdim)
		  X = ((X % Xdim) + Xdim) % Xdim;
	       if (Y < 0 || Y >= Ydim)
		  Y = ((Y % Ydim) + Ydim) % Ydim;
	       TotalRed += (float) Data1[Y][X].r * Data0[i][j];
	       TotalGreen += (float) Data1[Y][X].g * Data0[i][j];
	       TotalBlue += (float) Data1[Y][X].b * Data0[i][j];
	    }
	 Data2[y][x].r = (unsigned char) TotalRed;
	 Data2[y][x].g = (unsigned char) TotalGreen;
	 Data2[y][x].b = (unsigned char) TotalBlue;
      }

   /* Write information to output image */
   im_write(Image2, COLOR, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does 3D color convolve.                            */
/*---------------------------------------------------------------------------*/
void ColorConvolve3d(int Rows, int Cols, FLOAT_TYPE ** Data0)
{
   COLOR_TYPE ***Data1;
   COLOR_TYPE ***Data2;
   int xshift, yshift;
   int i, j, x, y, z, X, Y;
   float TotalRed, TotalGreen, TotalBlue;

   Data1 = (COLOR_TYPE ***) im_alloc3D(Image1, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (COLOR_TYPE ***) im_alloc3D(Image2, COLOR);

   /* Perform image convolution */
   xshift = Cols / 2;
   yshift = Rows / 2;
   for (z = 0; z < Zdim; z++)
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    TotalRed = TotalGreen = TotalBlue = 0.0;
	    for (i = 0; i < Rows; i++)
	       for (j = 0; j < Cols; j++)
	       {
		  X = x - j + xshift;
		  Y = y - i + yshift;
		  if (X < 0 || X >= Xdim)
		     X = ((X % Xdim) + Xdim) % Xdim;
		  if (Y < 0 || Y >= Ydim)
		     Y = ((Y % Ydim) + Ydim) % Ydim;
		  TotalRed += (float) Data1[z][Y][X].r * Data0[i][j];
		  TotalGreen += (float) Data1[z][Y][X].g * Data0[i][j];
		  TotalBlue += (float) Data1[z][Y][X].b * Data0[i][j];
	       }
	    Data2[z][y][x].r = (unsigned char) TotalRed;
	    Data2[z][y][x].g = (unsigned char) TotalGreen;
	    Data2[z][y][x].b = (unsigned char) TotalBlue;
	 }

   /* Write information to output image */
   im_write(Image2, COLOR, (char *) &(Data2[0][0][0]));
   im_free3D((char ***) Data1);
   im_free3D((char ***) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Program variables */
   int Debug = FALSE;
   int i = 0, j, Rows, Cols;
   int Normalize = TRUE;
   float Total = 0.0;
   FLOAT_TYPE **Data0;
   FILE *fd;
   float Temp;

   /* Interpret program options */
   printf("CONVOLVE Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'n':
	 Normalize = FALSE;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr, "Usage: convolve [options] maskfile infile outfile\n");
      fprintf(stderr, "       [-d]  Print debugging information\n");
      fprintf(stderr, "       [-n]  No mask normalization\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name0) == 0)
      Error("Could not get mask file name");
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Open mask file */
   if ((fd = fopen(Name0, "r")) == NULL)
      Error("Could not open mask file");
   if (fscanf(fd, "%d %d", &Rows, &Cols) != 2)
      Error("Could not read convolution mask size");
   if (Debug == TRUE)
      printf("mask size = %d %d\n", Rows, Cols);
   Image0 = im_create("/dev/null", FLOAT, Cols, Rows, 1);
   Data0 = (FLOAT_TYPE **) im_alloc2D(Image0, FLOAT);

   /* Read convolution mask */
   for (i = 0; i < Rows; i++)
      for (j = 0; j < Cols; j++)
      {
	 if (fscanf(fd, "%f", &Temp) != 1)
	    Error("Could not read convolution mask");
	 if (Debug == TRUE)
	    printf("read %d %d = %f\n", i, j, Temp);
	 Total += Temp;
	 Data0[i][j] = Temp;
      }

   /* Normalize mask */
   if (Total == 0.0)
      Total = 1.0;
   if (Normalize == TRUE)
      for (i = 0; i < Rows; i++)
	 for (j = 0; j < Cols; j++)
	    Data0[i][j] /= Total;

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt == 1)
      Error("Can not process 1D images");
   if ((PixType == COLOR) || (PixType == PSEUDO) || (PixType == JPEG_RGB))
   {
      if (DimCnt == 3)
	 ColorConvolve3d(Rows, Cols, Data0);
      else if (DimCnt == 2)
	 ColorConvolve2d(Rows, Cols, Data0);
   }
   else
   {
      if (DimCnt == 3)
	 GreyConvolve3d(Rows, Cols, Data0);
      else if (DimCnt == 2)
	 GreyConvolve2d(Rows, Cols, Data0);
   }
   return (0);
}
