/*---------------------------------------------------------------------------*/
/* Program:  fading.c                                                        */
/*                                                                           */
/* Purpose:  This program interpolate the fading between two KUIM images   . */
/*                                                                           */
/* Author:   John Gauch, Edu Metz                                            */
/*                                                                           */
/* Date:     October 26, 1994                                                */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Image variables */
char Name1[50];
char Name2[50];
char Name3[50];
IM_TYPE *Image1;
IM_TYPE *Image2;
IM_TYPE *Image3;
int PixType1, PixType2, DimCnt1, DimCnt2;
int Xdim, Ydim, Zdim, z, y, x;
int FadingSize = 8;

/*---------------------------------------------------------------------------*/
/* Purpose: Interpolate the fading between two Grey images                   */
/*---------------------------------------------------------------------------*/
void GreyFading()
{
   FLOAT_TYPE **Data1;
   FLOAT_TYPE **Data2;
   FLOAT_TYPE ***Data3;
   float Intermediate;

   /* Read imput image 1 and 2 */
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   im_read(Image2, FLOAT, (char *) &(Data2[0][0]));

   /* Create output image */
   Image3 = im_create(Name3, FLOAT, Xdim, Ydim, FadingSize + 2);
   Data3 = (FLOAT_TYPE ***) im_alloc3D(Image3, FLOAT);

   /* Initialize the start and stop of the fading sequence */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Data3[0][y][x] = Data1[y][x];
	 Data3[FadingSize + 1][y][x] = Data2[y][x];
      }

   /* Estimate the fading values */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Intermediate = (Data3[FadingSize + 1][y][x] - Data3[0][y][x])
	    / (FadingSize + 1);
	 for (z = 1; z < FadingSize + 1; z++)
	    Data3[z][y][x] = Data3[0][y][x] + (Intermediate * z);
      }

   /* write information to output image */
   im_write(Image3, FLOAT, (char *) &(Data3[0][0][0]));
   im_free3D((char ***) Data3);
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}


/*---------------------------------------------------------------------------*/
/* Purpose:  Interpolate the fading between two Color images                 */
/*---------------------------------------------------------------------------*/
void ColorFading()
{
   COLOR_TYPE **Data1;
   COLOR_TYPE **Data2;
   COLOR_TYPE ***Data3;
   float IntermedRed, IntermedGreen, IntermedBlue;

   /* Read imput image 1 and 2 */
   Data1 = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0]));
   Data2 = (COLOR_TYPE **) im_alloc2D(Image2, COLOR);
   im_read(Image2, COLOR, (char *) &(Data2[0][0]));

   /* Create output image */
   Image3 = im_create(Name3, COLOR, Xdim, Ydim, FadingSize + 2);
   Data3 = (COLOR_TYPE ***) im_alloc3D(Image3, COLOR);

   /* Initialize the start and stop of the fading sequence */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Data3[0][y][x].r = Data1[y][x].r;
	 Data3[0][y][x].g = Data1[y][x].g;
	 Data3[0][y][x].b = Data1[y][x].b;
	 Data3[FadingSize + 1][y][x].r = Data2[y][x].r;
	 Data3[FadingSize + 1][y][x].g = Data2[y][x].g;
	 Data3[FadingSize + 1][y][x].b = Data2[y][x].b;
      }

   /* Estimate the fading values */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 IntermedRed = ((float) Data3[FadingSize + 1][y][x].r -
			(float) Data3[0][y][x].r) / (FadingSize + 1);
	 IntermedGreen = ((float) Data3[FadingSize + 1][y][x].g -
			  (float) Data3[0][y][x].g) / (FadingSize + 1);
	 IntermedBlue = ((float) Data3[FadingSize + 1][y][x].b -
			 (float) Data3[0][y][x].b) / (FadingSize + 1);
	 for (z = 1; z < FadingSize + 1; z++)
	 {
	    Data3[z][y][x].r = (unsigned char)
	       (Data3[0][y][x].r + (IntermedRed * z));
	    Data3[z][y][x].g = (unsigned char)
	       (Data3[0][y][x].g + (IntermedGreen * z));
	    Data3[z][y][x].b = (unsigned char)
	       (Data3[0][y][x].b + (IntermedBlue * z));
	 }
      }

   /* write information to output image */
   im_write(Image3, COLOR, (char *) &(Data3[0][0][0]));
   im_free3D((char ***) Data3);
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
/* Program variables */
   int i = 0;
   int Debug = FALSE;

   /* Interpret program options */
   printf("FADING Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 's':
	 if (sscanf(argv[++i], "%d", &FadingSize) == 0)
	    Error("Can not get the size of the fading block");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr,
	      "Usage: fading [options] infile1 infile2 outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-s #]  Size of the fading block (8)\n");
      exit(1);
   }

   /* Get image file names from argument list */

   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name3) == 0)
      Error("Could not get output file name");

   /* Read input 1,2 images */
   Image1 = im_open(Name1, &PixType1, &Xdim, &Ydim, &Zdim, &DimCnt1);
   Image2 = im_open(Name2, &PixType2, &Xdim, &Ydim, &Zdim, &DimCnt2);

   if (DimCnt1 != 2 || DimCnt2 != 2)
      Error("Can not process 1D or 3D images");

   if ((PixType1 == COLOR || PixType1 == PSEUDO || PixType1 == JPEG_RGB) &&
       (PixType2 == COLOR || PixType2 == PSEUDO || PixType2 == JPEG_RGB))
      ColorFading();
   else
      GreyFading();
   return (0);
}
