/*---------------------------------------------------------------------------*/
/* Program:  gauss.c                                                         */
/*                                                                           */
/* Purpose:  This program performs Gaussian blurring in the frequency        */
/*           domain.  Images are automatically converted to COMPLEX          */
/*           type, filtered, and converted back to FLOAT type.               */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     April 4, 1994   - Original program.                             */
/*           August 14, 1996 - Fixed stddev calculations.                    */
/*           July 31, 1997   - Added support for color images.               */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994-1996               */
/*---------------------------------------------------------------------------*/
#include <IM.h>

double exp();
float StdDev = 1.0;
float StdDevX = 1.0;
float StdDevY = 1.0;
float StdDevZ = 1.0;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs 1D Gaussian filtering.                    */
/*---------------------------------------------------------------------------*/
void Gauss1D(COMPLEX_TYPE * Data, int Xdim)
{
   int x, x1, halfx;
   float denomX, value;

   /* Convert to frequency domain */
   if (fft_1d(Data, Xdim, -1) == INVALID)
      Error("Could not perform FFT");

   /* Prepare to filter */
   denomX = -2.0 * StdDevX * StdDevX;
   halfx = Xdim / 2;

   /* Loop over X axis */
   for (x = 0; x < Xdim; x++)
   {
      if (x < halfx)
	 x1 = x;
      else
	 x1 = Xdim - x;

      /* Calculate value of Gaussian */
      value = (float) exp(x1 * x1 / denomX);

      /* Do pointwise multiplication */
      Data[x].re *= value;
      Data[x].im *= value;
   }

   /* Convert to spatial domain */
   if (fft_1d(Data, Xdim, 1) == INVALID)
      Error("Could not perform FFT");
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs 2D Gaussian filtering.                    */
/*---------------------------------------------------------------------------*/
void Gauss2D(COMPLEX_TYPE ** Data, int Xdim, int Ydim)
{
   int x, y, x1, y1, halfx, halfy;
   float denomX, denomY, value;

   /* Convert to frequency domain */
   if (fft_2d(&(Data[0][0]), Xdim, Ydim, -1) == INVALID)
      Error("Could not perform FFT");

   /* Prepare to filter */
   denomX = -2.0 * StdDevX * StdDevX;
   denomY = -2.0 * StdDevY * StdDevY;
   halfx = Xdim / 2;
   halfy = Ydim / 2;

   /* Loop over Y axis */
   for (y = 0; y < Ydim; y++)
   {
      if (y < halfy)
	 y1 = y;
      else
	 y1 = Ydim - y;

      /* Loop over X axis */
      for (x = 0; x < Xdim; x++)
      {
	 if (x < halfx)
	    x1 = x;
	 else
	    x1 = Xdim - x;

	 /* Calculate value of Gaussian */
	 value = (float) exp(x1 * x1 / denomX + y1 * y1 / denomY);

	 /* Do pointwise multiplication */
	 Data[y][x].re *= value;
	 Data[y][x].im *= value;
      }
   }

   /* Convert to spatial domain */
   if (fft_2d(&(Data[0][0]), Xdim, Ydim, 1) == INVALID)
      Error("Could not perform inverse FFT");
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs 3D Gaussian filtering.                    */
/*---------------------------------------------------------------------------*/
void Gauss3D(COMPLEX_TYPE *** Data, int Xdim, int Ydim, int Zdim)
{
   int x, y, z, x1, y1, z1, halfx, halfy, halfz;
   float denomX, denomY, denomZ, value;

   /* Convert to frequency domain */
   if (fft_3d(&(Data[0][0][0]), Xdim, Ydim, Zdim, -1) == INVALID)
      Error("Could not perform FFT");

   /* Prepare to filter */
   denomX = -2.0 * StdDevX * StdDevX;
   denomY = -2.0 * StdDevY * StdDevY;
   denomZ = -2.0 * StdDevZ * StdDevZ;
   halfx = Xdim / 2;
   halfy = Ydim / 2;
   halfz = Zdim / 2;

   /* Loop over Z axis */
   for (z = 0; z < Zdim; z++)
   {
      if (z < halfz)
	 z1 = z;
      else
	 z1 = Zdim - z;

      /* Loop over Y axis */
      for (y = 0; y < Ydim; y++)
      {
	 if (y < halfy)
	    y1 = y;
	 else
	    y1 = Ydim - y;

	 /* Loop over X axis */
	 for (x = 0; x < Xdim; x++)
	 {
	    if (x < halfx)
	       x1 = x;
	    else
	       x1 = Xdim - x;

	    /* Calculate value of Gaussian */
	    value = (float) exp(x1 * x1 / denomX + y1 * y1 / denomY + z1 * z1 / denomZ);

	    /* Do pointwise multiplication */
	    Data[z][y][x].re *= value;
	    Data[z][y][x].im *= value;
	 }
      }
   }

   /* Convert to spatial domain */
   if (fft_3d(&(Data[0][0][0]), Xdim, Ydim, Zdim, 1) == INVALID)
      Error("Could not perform inverse FFT");
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   COMPLEX_TYPE *Data1;
   COMPLEX_TYPE **Data2;
   COMPLEX_TYPE ***Data3;
   COLOR_TYPE **Data4;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int Spatial = TRUE;
   int i = 0, x, y, z;

   /* Interpret program options */
   printf("GAUSS Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 's':
      case 'S':
	 Spatial = TRUE;
	 if (sscanf(argv[++i], "%f", &StdDev) == 0)
	    Error("Could not get spatial width");
	 break;
      case 'f':
      case 'F':
	 Spatial = FALSE;
	 if (sscanf(argv[++i], "%f", &StdDev) == 0)
	    Error("Could not get frequency width");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: gauss [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-s #]  Width of spatial gaussian (1)\n");
      fprintf(stderr, "       [-f #]  Width of frequency gaussian\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image and create output image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);

   /* Initialize stddev properly */
   if (Spatial == TRUE)
   {
      StdDevX = Xdim / (2.0 * M_PI * StdDev);
      StdDevY = Ydim / (2.0 * M_PI * StdDev);
      StdDevZ = Zdim / (2.0 * M_PI * StdDev);
   }
   else
      StdDevX = StdDevY = StdDevZ = StdDev;

   /* Handle color images */
   if (PixType == COLOR || PixType == PSEUDO || PixType == JPEG_RGB)
   {
      Image3 = im_create("/dev/null", COMPLEX, Xdim, Ydim, Zdim = 3);
      Data3 = (COMPLEX_TYPE ***) im_alloc3D(Image3, COMPLEX);
      Data4 = (COLOR_TYPE **) im_alloc2D(Image2, COLOR);
      im_read(Image1, COLOR, (char *) &(Data4[0][0]));
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    Data3[0][y][x].re = (float) Data4[y][x].r;
	    Data3[1][y][x].re = (float) Data4[y][x].g;
	    Data3[2][y][x].re = (float) Data4[y][x].b;
	 }
      for (z = 0; z < Zdim; z++)
	 Gauss2D(Data3[z], Xdim, Ydim);
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    Data4[y][x].r = (BYTE_TYPE) Data3[0][y][x].re;
	    Data4[y][x].g = (BYTE_TYPE) Data3[1][y][x].re;
	    Data4[y][x].b = (BYTE_TYPE) Data3[2][y][x].re;
	 }
      im_write(Image2, COLOR, (char *) &(Data4[0][0]));
      im_free3D((char ***) Data3);
      im_free2D((char **) Data4);
   }

   /* Handle 1D images */
   else if (DimCnt == 1)
   {
      Data1 = (COMPLEX_TYPE *) im_alloc1D(Image1, COMPLEX);
      im_read(Image1, COMPLEX, (char *) Data1);
      Gauss1D(Data1, Xdim);
      im_write(Image2, COMPLEX, (char *) Data1);
      im_free1D((char *) Data1);
   }

   /* Handle 2D images */
   else if (DimCnt == 2)
   {
      Data2 = (COMPLEX_TYPE **) im_alloc2D(Image1, COMPLEX);
      im_read(Image1, COMPLEX, (char *) &(Data2[0][0]));
      Gauss2D(Data2, Xdim, Ydim);
      im_write(Image2, COMPLEX, (char *) &(Data2[0][0]));
      im_free2D((char **) Data2);
   }

   /* Handle 3D images */
   else if (DimCnt == 3)
   {
      Data3 = (COMPLEX_TYPE ***) im_alloc3D(Image1, COMPLEX);
      im_read(Image1, COMPLEX, (char *) &(Data3[0][0][0]));
      for (z = 0; z < Zdim; z++)
	 Gauss2D(Data3[z], Xdim, Ydim);
      /* Replaced call to Gauss3D(Data3, Xdim, Ydim, Zdim); */
      im_write(Image2, COMPLEX, (char *) &(Data3[0][0][0]));
      im_free3D((char ***) Data3);
   }

   return (0);
}
