/*---------------------------------------------------------------------------*/
/* Program:  huff_encode.c                                                   */
/*                                                                           */
/* Purpose:  This program performs Huffman encoding of KUIM images.          */
/*           The encoded pixels are output to a compressed binary file.      */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     October 26, 1995 - Original program.                            */
/*           November 1, 1995 - Extended to any grey level range.            */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define MAX_SIZE 10000

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   SHORT_TYPE **Data1;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int i = 0, x, y;
   char Bit[MAX_SIZE];
   char Code[MAX_SIZE];
   short int Prob[MAX_SIZE];
   int Parent[MAX_SIZE];
   int Loop, Min, Max, Range, Pos, Len;
   FILE *fd;
   int Buffer;
   int BufLen;
   int TotalLen = 0;

   /* Interpret program options */
   fprintf(stderr, "HUFF_ENCODE Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: huff_encode [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   im_read(Image1, SHORT, (char *) &(Data1[0][0]));

   /* Find min and max pixel values */
/*-------------------------------*/
   Min = Max = Data1[0][0];
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 if (Min > Data1[y][x])
	    Min = Data1[y][x];
	 else if (Max < Data1[y][x])
	    Max = Data1[y][x];
      }
   Range = Max - Min + 1;

   /* Subtract min to get range 0..max-min */
/*--------------------------------------*/
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 Data1[y][x] -= Min;

   /* Calculate pixel probabilities */
/*-------------------------------*/
   for (i = 0; i < MAX_SIZE; i++)
   {
      Bit[i] = 0;
      Prob[i] = 0;
      Parent[i] = -1;
   }
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 Prob[Data1[y][x]]++;

   /* Write Huffman file header */
/*---------------------------*/
   if ((fd = fopen(Name2, "w")) == NULL)
      Error("Could not open output file");
   fwrite(&Xdim, 1, sizeof(Xdim), fd);
   fwrite(&Ydim, 1, sizeof(Ydim), fd);
   fwrite(&Min, 1, sizeof(Min), fd);
   fwrite(&Max, 1, sizeof(Max), fd);
   fwrite(Prob, Range, sizeof(Prob[0]), fd);

   /* Build the Huffman code tree */
/*-----------------------------*/
   for (Loop = 0; Loop < (Range - 1); Loop++)
   {
      /* Process lowest probability symbol */
      Pos = -1;
      Min = Xdim * Ydim;
      for (i = 0; i < Range + Loop; i++)
	 if ((Prob[i] < Min) && (Parent[i] == -1))
	 {
	    Min = Prob[i];
	    Pos = i;
	 }
      Bit[Pos] = 0;
      Prob[Range + Loop] += Min;
      Parent[Pos] = Range + Loop;

      /* Process second lowest probability symbol */
      Pos = -1;
      Min = Xdim * Ydim;
      for (i = 0; i < Range + Loop; i++)
	 if ((Prob[i] < Min) && (Parent[i] == -1))
	 {
	    Min = Prob[i];
	    Pos = i;
	 }
      Bit[Pos] = 1;
      Prob[Range + Loop] += Min;
      Parent[Pos] = Range + Loop;
   }

   /* Print Huffman code tree */
/*-------------------------*/
   Min = Max - Range + 1;
   if (Debug == TRUE)
      for (Loop = 0; Loop < Range; Loop++)
      {
	 Len = 0;
	 Pos = Loop;
	 printf("Value %d Probability %d Code ", Loop + Min, Prob[Loop]);
	 while (Parent[Pos] >= 0)
	 {
	    Code[Len++] = Bit[Pos];
	    Pos = Parent[Pos];
	 }
	 TotalLen += Len * Prob[Loop];
	 while (Len > 0)
	    printf("%c", '0' + Code[--Len]);
	 printf("\n");
      }
   if (Debug == TRUE)
      printf("Average pixel code length %f bits\n", TotalLen / (float) (Xdim * Ydim));

   /* Encode image data using Huffman codes */
/*---------------------------------------*/
   Buffer = 0;
   BufLen = 0;
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 /* Build up the Huffman code */
	 Len = 0;
	 Pos = Data1[y][x];
	 while (Parent[Pos] >= 0)
	 {
	    Code[Len++] = Bit[Pos];
	    Pos = Parent[Pos];
	 }

	 /* Write Huffman code one bit at a time */
	 while (Len > 0)
	 {
	    Buffer = (Buffer << 1) | Code[--Len];
	    BufLen++;

	    /* Flush buffer when full */
	    if (BufLen == 32)
	    {
	       fwrite(&Buffer, 1, sizeof(Buffer), fd);
	       Buffer = 0;
	       BufLen = 0;
	    }
	 }
      }

   /* Flush last buffer */
   if (BufLen != 0)
   {
      Buffer = Buffer << (32 - BufLen);
      fwrite(&Buffer, 1, sizeof(Buffer), fd);
   }

   /* Finish program */
   im_free2D((char **) Data1);
   fclose(fd);
   return (0);
}
