/*---------------------------------------------------------------------------*/
/* Program:  ice_edge.c                                                      */
/*                                                                           */
/* Purpose:  This program follows ice interfaces in radar echogram           */
/*           images by simply tracking intensity maxima in the Y             */
/*           direction.  A small (1x7 pixel) window is used for this         */
/*           local search.  It is assumed that the input image has           */
/*           been smoothed somewhat.  The starting point for the edge        */
/*           tracking must be input using the -s switch.                     */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     March 17, 1996                                                  */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1996                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   FLOAT_TYPE **Data1;
   FLOAT_TYPE **Data2;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int X, Y, nx;
   int i = 0, x, y;
   float Best;

   /* Interpret program options */
   printf("ICE_EDGE Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 's':
	 if (sscanf(argv[++i], "%d", &X) == 0)
	    Error("Could not get integer argument");
	 if (sscanf(argv[++i], "%d", &Y) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: ice_edge [options] infile outfile\n");
      fprintf(stderr, "       [-d]      Print debugging information\n");
      fprintf(stderr, "       [-s # #]  Starting X Y position\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Walk uphill in X from start X Y position */
   while ((X > 0) && (Data1[Y][X - 1] > Data1[Y][X]))
      X--;
   while ((X < Xdim - 1) && (Data1[Y][X + 1] > Data1[Y][X]))
      X++;
   printf("starting from %d %d\n", X, Y);
   Data2[Y][X] = 1;

   /* Follow edge in increasing Y direction */
   for (x = X, y = Y + 1; y < Ydim; y++)
   {
      Best = Data2[y][x];
      if (Data1[y][x - 3] > Best)
      {
	 nx = x - 3;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x - 2] > Best)
      {
	 nx = x - 2;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x - 1] > Best)
      {
	 nx = x - 1;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x + 1] > Best)
      {
	 nx = x + 1;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x + 2] > Best)
      {
	 nx = x + 2;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x + 3] > Best)
      {
	 nx = x + 3;
	 Best = Data1[y][nx];
      }
      x = nx;
      Data2[y][x] = 1;
   }

   /* Follow edge in decreasing Y direction */
   for (x = X, y = Y - 1; y > 0; y--)
   {
      Best = Data2[y][x];
      if (Data1[y][x - 3] > Best)
      {
	 nx = x - 3;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x - 2] > Best)
      {
	 nx = x - 2;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x - 1] > Best)
      {
	 nx = x - 1;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x + 1] > Best)
      {
	 nx = x + 1;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x + 2] > Best)
      {
	 nx = x + 2;
	 Best = Data1[y][nx];
      }
      if (Data1[y][x + 3] > Best)
      {
	 nx = x + 3;
	 Best = Data1[y][nx];
      }
      x = nx;
      Data2[y][x] = 1;
   }

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   return (0);
}
