/*---------------------------------------------------------------------------*/
/* Program:  ideal.c                                                         */
/*                                                                           */
/* Purpose:  This program demonstrates ideal filtering in the frequency      */
/*           domain.  Images are automatically converted to COMPLEX type     */
/*           to be filtered and converted back to FLOAT type to be output.   */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     October 6, 1995                                                 */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   COMPLEX_TYPE **Data;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int LowF = 0;
   int HighF = 512;
   int Debug = FALSE;
   int i = 0, x, y, u, v;
   float dist;

   /* Interpret program options */
   printf("IDEAL Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'l':
	 if (sscanf(argv[++i], "%d", &HighF) == 0)
	    Error("Could not get frequency argument");
	 break;
      case 'h':
	 if (sscanf(argv[++i], "%d", &LowF) == 0)
	    Error("Could not get frequency argument");
	 break;
      case 'b':
	 if (sscanf(argv[++i], "%d", &LowF) == 0)
	    Error("Could not get frequency argument");
	 if (sscanf(argv[++i], "%d", &HighF) == 0)
	    Error("Could not get frequency argument");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: ideal [options] infile outfile\n");
      fprintf(stderr, "       [-l #]      Lowpass filter  (0..#) kept\n");
      fprintf(stderr, "       [-h #]      Highpass filter (#..max) kept\n");
      fprintf(stderr, "       [-b #1 #2]  Bandpass filter (#1..#2) kept\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image and create output image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Image2 = im_create(Name2, FLOAT, Xdim, Ydim, Zdim);
   Data = (COMPLEX_TYPE **) im_alloc2D(Image1, COMPLEX);
   im_read(Image1, COMPLEX, (char *) &(Data[0][0]));

   /* Convert image to frequency domain */
   if (fft_2d(&(Data[0][0]), Xdim, Ydim, -1) == INVALID)
      Error("Could not perform forward FFT");

   /* Perform ideal lowpass or highpass filtering */
   for (v = (-Ydim / 2); v < Ydim / 2; v++)
      for (u = (-Xdim / 2); u < Xdim / 2; u++)
      {
	 /* Calculate distance to origin */
	 dist = (float) sqrt((double) (u * u + v * v));
	 if ((dist < LowF) || (dist > HighF))
	 {
	    x = (u + Xdim) % Xdim;
	    y = (v + Ydim) % Ydim;
	    Data[y][x].re = 0.0;
	    Data[y][x].im = 0.0;
	 }
      }

   /* Convert image back to spatial domain */
   if (fft_2d(&(Data[0][0]), Xdim, Ydim, 1) == INVALID)
      Error("Could not perform inverse FFT");

   /* Write information to output image */
   im_write(Image2, COMPLEX, (char *) &(Data[0][0]));
   im_free2D((char **) Data);
   return (0);
}
