/*---------------------------------------------------------------------------*/
/* Program:  JPEG_io.c                                                       */
/*                                                                           */
/* Purpose:  This program reads and writes jpg data files.                   */
/*           Uses code borrowed from djpeg.c and cjpeg.c                     */
/*                                                                           */
/* Author:   John Gauch / Chunyen Liu                                        */
/*                                                                           */
/* Date:     November 30, 1995                                               */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <jinclude.h>
#include <jpeglib.h>

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine checks if an image is in JPEG format.              */
/*---------------------------------------------------------------------------*/
int JPEG_test(fp)
     FILE *fp;
{
   unsigned char c1, c2;
   int ok = 0;

   c1 = fgetc(fp);
   c2 = fgetc(fp);
   if (c1 == 0xff && c2 == 0xd8)
      ok = 1;
   return ok;
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This reads in JPEG header.                                      */
/*---------------------------------------------------------------------------*/
void JPEG_read_header(Image, PixType, Xdim, Ydim, Zdim, DimCnt)
     IM_TYPE *Image;
     int *PixType, *Xdim, *Ydim, *Zdim, *DimCnt;
{
   /* JPEG declarations */
   struct jpeg_decompress_struct dinfo;
   struct jpeg_error_mgr jerr;

   /* Initialize the JPEG decompression object with default error handling. */
   dinfo.err = jpeg_std_error(&jerr);
   jpeg_create_decompress(&dinfo);

   /* Specify data source for decompression */
   jpeg_stdio_src(&dinfo, Image->Fd);

   /* Read file header, set default decompression parameters */
   (void) jpeg_read_header(&dinfo, TRUE);

   jpeg_calc_output_dimensions(&dinfo);

   if (dinfo.out_color_space == JCS_GRAYSCALE)
   {
      *PixType = Image->PixType = JPEG_GRAY;
   }
   else
   {
      *PixType = Image->PixType = JPEG_RGB;
   }

   *Xdim = Image->Xdim = dinfo.output_width;
   *Ydim = Image->Ydim = dinfo.output_height;
   *Zdim = Image->Zdim = 1;
   *DimCnt = 2;
   jpeg_destroy_decompress(&dinfo);
   rewind(Image->Fd);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This reads JPEG data.                                           */
/*---------------------------------------------------------------------------*/
void JPEG_read_data(Image, Data)
     IM_TYPE *Image;
     char *Data;
{
   int i = 0, index = 0, GreyScale = TRUE, Value;

   /* JPEG declarations */
   int BitsPerPixel, cshift, ColorMapSize, numberOfColors;
   struct jpeg_decompress_struct dinfo;
   struct jpeg_error_mgr jerr;
   JSAMPARRAY buffer;
   JSAMPROW ptr;
   int rowStride;

   /* Initialize the JPEG decompression object with default error handling. */
   dinfo.err = jpeg_std_error(&jerr);
   jpeg_create_decompress(&dinfo);

   /* Specify data source for decompression */
   jpeg_stdio_src(&dinfo, Image->Fd);

   /* Read file header, set default decompression parameters */
   (void) jpeg_read_header(&dinfo, TRUE);

   /* Force quantization if color or if > 8 bits input */
   if (dinfo.out_color_space != JCS_GRAYSCALE &&
       dinfo.out_color_space != JCS_RGB)
      Error("Incorrect color space");

   if (dinfo.out_color_space != JCS_GRAYSCALE || dinfo.data_precision > 8)
   {
      dinfo.quantize_colors = TRUE;
      if (dinfo.desired_number_of_colors > 256)
	 dinfo.desired_number_of_colors = 256;
   }

   jpeg_calc_output_dimensions(&dinfo);

   /* Start decompressor */
   jpeg_start_decompress(&dinfo);

   if (dinfo.quantize_colors == TRUE)
      numberOfColors = dinfo.actual_number_of_colors;
   else
      numberOfColors = 256;

   if (numberOfColors > 256)
      Error("Too many colors");

   cshift = dinfo.data_precision - 8;
   BitsPerPixel = 1;
   while (numberOfColors > (1 << BitsPerPixel))
      BitsPerPixel++;
   ColorMapSize = 1 << BitsPerPixel;

   /* Set up color map */
   if (dinfo.out_color_space == JCS_RGB)
      for (i = 0; i < ColorMapSize; i++)
	 if (i < numberOfColors)
	 {
	    /* normal RGB color map */
	    Image->Red[i] = GETJSAMPLE(dinfo.colormap[0][i]) >> cshift;
	    Image->Green[i] = GETJSAMPLE(dinfo.colormap[1][i]) >> cshift;
	    Image->Blue[i] = GETJSAMPLE(dinfo.colormap[2][i]) >> cshift;
	    GreyScale = ((Image->Red[i] == Image->Green[i])
		       && (Image->Green[i] == Image->Blue[i]) && GreyScale);
	 }
	 else
	 {
	    /* fill out the map to a power of 2 */
	    Image->Red[i] = Image->Green[i] = Image->Blue[i] = 0;
	 }

   /* JSAMPLES per row in output buffer */
   rowStride = dinfo.output_width * dinfo.output_components;
   buffer = (*dinfo.mem->alloc_sarray) ((j_common_ptr) & dinfo, JPOOL_IMAGE,
				    (JDIMENSION) rowStride, (JDIMENSION) 1);

   /* Process JPEG scanlines */
   while (dinfo.output_scanline < dinfo.output_height)
   {
      (void) jpeg_read_scanlines(&dinfo, buffer, 1);
      ptr = buffer[0];
      for (i = 0; i < (int) dinfo.output_width; i++)
      {
	 Value = GETJSAMPLE(*ptr++);
	 Data[index++] = (unsigned char) Value;
      }
   }

   /* Finish decompression and release memory */
   (void) jpeg_finish_decompress(&dinfo);
   jpeg_destroy_decompress(&dinfo);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This writes JPEG data.                                          */
/*---------------------------------------------------------------------------*/
void JPEG_write_data(Image, Data)
     IM_TYPE *Image;
     char *Data;
{
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   int i, quality = 75, smoothness = 0, Value, index = 0;
   JSAMPARRAY colormap, buffer;
   JSAMPROW ptr;

   /* Initialize the JPEG compression object with default error handling. */
   cinfo.err = jpeg_std_error(&jerr);
   jpeg_create_compress(&cinfo);

   /* Initialize JPEG parameters which may be overridden later */
   cinfo.in_color_space = JCS_RGB;
   jpeg_set_defaults(&cinfo);

   /* Set up color map */
   colormap = (*cinfo.mem->alloc_sarray) ((j_common_ptr) & cinfo,
					  JPOOL_IMAGE, 256, 3);
   if (Image->PixType == JPEG_RGB)
      for (i = 0; i < 256; i++)
      {
	 colormap[0][i] = Image->Red[i];
	 colormap[1][i] = Image->Green[i];
	 colormap[2][i] = Image->Blue[i];
      }

   /* Obtain file size and colorspace */
   if (Image->PixType == JPEG_RGB)
   {
      cinfo.input_components = 3;
      cinfo.in_color_space = JCS_RGB;
   }
   else
   {
      cinfo.input_components = 1;
      cinfo.in_color_space = JCS_GRAYSCALE;
   }
   cinfo.image_width = Image->Xdim;
   cinfo.image_height = Image->Ydim;
   cinfo.data_precision = BITS_IN_JSAMPLE;
   buffer = (*cinfo.mem->alloc_sarray) ((j_common_ptr) & cinfo, JPOOL_IMAGE,
					(JDIMENSION) (cinfo.image_width * cinfo.input_components), (JDIMENSION) 1);

   /* Fix colorspace-dependent defaults */
   jpeg_default_colorspace(&cinfo);

   /* Customize parameters */
   cinfo.smoothing_factor = smoothness;
   jpeg_set_quality(&cinfo, quality, TRUE);

   /* Specify data destination for compression */
   jpeg_stdio_dest(&cinfo, Image->Fd);

   /* Start compressor */
   jpeg_start_compress(&cinfo, TRUE);

   /* Process JPEG scanlines */
   while (cinfo.next_scanline < cinfo.image_height)
   {
      ptr = buffer[0];
      for (i = 0; i < (int) cinfo.image_width; i++)
      {
	 Value = (unsigned char) Data[index++];
	 if (Image->PixType == JPEG_RGB)
	 {
	    *ptr++ = colormap[0][Value];
	    *ptr++ = colormap[1][Value];
	    *ptr++ = colormap[2][Value];
	 }
	 else
	    *ptr++ = Value;
      }
      (void) jpeg_write_scanlines(&cinfo, buffer, 1);
   }

   /* Finish compression and release memory */
   jpeg_finish_compress(&cinfo);
   jpeg_destroy_compress(&cinfo);
}
