/*---------------------------------------------------------------------------*/
/* Program:  link_extrema.c                                                  */
/*                                                                           */
/* Purpose:  This program uses gaussian blurring to generate a stack of      */
/*           grey-scale images.  Then the extrema are linked from one        */
/*           level to the next using a simple hill climbing algorithm.       */
/*           To control the level of blurring, the user must specify         */
/*           the standard deviation of the gaussian convolution kernel.      */
/*                                                                           */
/* Inputs:   An input image file.                                            */
/*           A file containing standard deviations for blurring.             */
/*                                                                           */
/* Outputs:  A output file representing the extrema hierarchy.               */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     February 13, 1995 - Converted to KUIM format.                   */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define MaxCount 50000

/* Global program variables */
int Gradient = TRUE;
int Debug = TRUE;
int Xchild[MaxCount];
int Ychild[MaxCount];
int Level[MaxCount];
int Parent[MaxCount];
int Xparent[MaxCount];
int Yparent[MaxCount];
int Count;
int BlurCount = 25;
int TableSize = 0;
float StdDev[100], MinCnt[100];
FILE *Fd;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine performs Gaussian filtering.                       */
/*---------------------------------------------------------------------------*/
void Gauss(COMPLEX_TYPE ** Data, int Xdim, int Ydim, float StdDev)
{
   int x, y, x1, y1, halfx, halfy;
   double numerator, denominator, value;

   /* Convert to frequency domain */
   if (fft_2d(&(Data[0][0]), Xdim, Ydim, -1) == INVALID)
      Error("Could not perform FFT");

   /* Prepare to filter */
   denominator = -2.0 * StdDev * StdDev;
   halfx = Xdim / 2;
   halfy = Ydim / 2;

   /* Loop over Y axis */
   for (y = 0; y < Ydim; y++)
   {
      if (y < halfy)
	 y1 = y;
      else
	 y1 = Ydim - y;

      /* Loop over X axis */
      for (x = 0; x < Xdim; x++)
      {
	 if (x < halfx)
	    x1 = x;
	 else
	    x1 = Xdim - x;

	 /* Calculate value of Gaussian */
	 numerator = x1 * x1 + y1 * y1;
	 value = exp(numerator / denominator);

	 /* Do pointwise multiplication */
	 Data[y][x].re *= (float) value;
	 Data[y][x].im *= (float) value;
      }
   }

   /* Convert to spatial domain */
   if (fft_2d(&(Data[0][0]), Xdim, Ydim, 1) == INVALID)
      Error("Could not perform inverse FFT");
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates the gradient magnitude.                 */
/*---------------------------------------------------------------------------*/
void GradMagnitude(COMPLEX_TYPE ** Data, FLOAT_TYPE ** Temp, int Xdim, int Ydim)
{
   int x, y;
   float Dx, Dy;

   /* Calculate internal point gradients */
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 Dx = (Data[y + 1][x + 1].re + 2 * Data[y][x + 1].re
	       + Data[y - 1][x + 1].re - Data[y + 1][x - 1].re
	       - 2 * Data[y][x - 1].re - Data[y - 1][x - 1].re) / 8.0;
	 Dy = (Data[y + 1][x + 1].re + 2 * Data[y + 1][x].re
	       + Data[y + 1][x - 1].re - Data[y - 1][x + 1].re
	       - 2 * Data[y - 1][x].re - Data[y - 1][x - 1].re) / 8.0;
	 Temp[y][x] = (FLOAT_TYPE) sqrt((double) (Dx * Dx + Dy * Dy));
      }
   for (x = 0; x < Xdim; x++)
   {
      Temp[0][x] = Temp[1][x];
      Temp[Ydim - 1][x] = Temp[Ydim - 2][x];
   }
   for (y = 0; y < Ydim; y++)
   {
      Temp[y][0] = Temp[y][1];
      Temp[y][Xdim - 1] = Temp[y][Xdim - 2];
   }
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 Data[y][x].re = Temp[y][x];
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine locates minima in an image.                        */
/*---------------------------------------------------------------------------*/
void FindMinima(COMPLEX_TYPE ** Data, int Xdim, int Ydim, float StdDev)
{
   int x, y, X, Y;
   int MinIndex;
   float MinValue;

   /* Make (0,0) a pseudo minima for boundary */
   Level[1] = 0;
   Xchild[1] = 0;
   Ychild[1] = 0;
   Parent[1] = -1;
   Xparent[1] = 0;
   Yparent[1] = 0;

   /* Find local minima */
   Count = 1;
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 /* Search 3x3 neighborhood */
	 MinValue = Data[y][x].re;
	 MinIndex = 4;
	 for (Y = 0; Y < 3; Y++)
	    for (X = 0; X < 3; X++)
	       if (Data[y + Y - 1][x + X - 1].re <= MinValue)
	       {
		  MinValue = Data[y + Y - 1][x + X - 1].re;
		  MinIndex = Y * 3 + X;
	       }

	 /* Save minima information */
	 if (MinIndex == 4)
	 {
	    Count++;
	    Level[Count] = 0;
	    Xchild[Count] = x;
	    Ychild[Count] = y;
	    Parent[Count] = -1;
	    Xparent[Count] = -1;
	    Yparent[Count] = -1;
	 }
      }
   fprintf(Fd, "%f %d\n", StdDev, Count);
   if (Debug == TRUE)
      printf("\t%f\t%d\n", StdDev, Count);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates watershed regions of an image.          */
/*---------------------------------------------------------------------------*/
void Watershed(COMPLEX_TYPE ** Data, SHORT_TYPE ** Water, int Xdim, int Ydim)
{
   int x, y, X, Y;
   int MinCount, MinIndex;
   float MinValue;

   /* Mark boundary as pseudo-minima */
   for (y = 0; y < Ydim; y++)
      Water[y][0] = Water[y][Xdim - 1] = 1;
   for (x = 0; x < Xdim; x++)
      Water[0][x] = Water[Ydim - 1][x] = 1;

   /* Find local minima and gradient for each pixel */
   MinCount = 1;
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 MinValue = Data[y][x].re;
	 MinIndex = 4;
	 for (Y = 0; Y < 3; Y++)
	    for (X = 0; X < 3; X++)
	       if (Data[y + Y - 1][x + X - 1].re <= MinValue)
	       {
		  MinValue = Data[y + Y - 1][x + X - 1].re;
		  MinIndex = Y * 3 + X;
	       }
	 if (MinIndex == 4)
	    Water[y][x] = (++MinCount);
	 else
	    Water[y][x] = -MinIndex;
      }

   /* Follow gradient downhill for each pixel */
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 X = x;
	 Y = y;
	 while (Water[Y][X] <= 0)
	 {
	    switch (Water[Y][X])
	    {
	    case 0:
	       X--;
	       Y--;
	       break;
	    case -1:
	       Y--;
	       break;
	    case -2:
	       X++;
	       Y--;
	       break;
	    case -3:
	       X--;
	       break;
	    case -5:
	       X++;
	       break;
	    case -6:
	       X--;
	       Y++;
	       break;
	    case -7:
	       Y++;
	       break;
	    case -8:
	       X++;
	       Y++;
	       break;
	    }
	 }
	 Water[y][x] = Water[Y][X];
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine links minima from one level to the next.           */
/*---------------------------------------------------------------------------*/
void LinkMinima(COMPLEX_TYPE ** Data, SHORT_TYPE ** Water, int Xdim,
		int Ydim, float StdDev)
{
   int i, x, y, X, Y;
   int MinIndex;
   float MinValue;
   int MinCount;

   /* Mark boundary as pseudo-minima */
   for (y = 0; y < Ydim; y++)
      Water[y][0] = Water[y][Xdim - 1] = 1;
   for (x = 0; x < Xdim; x++)
      Water[0][x] = Water[Ydim - 1][x] = 1;

   /* Find local minima and gradient for each pixel */
   MinCount = 1;
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 MinValue = Data[y][x].re;
	 MinIndex = 4;
	 for (Y = 0; Y < 3; Y++)
	    for (X = 0; X < 3; X++)
	       if (Data[y + Y - 1][x + X - 1].re <= MinValue)
	       {
		  MinValue = Data[y + Y - 1][x + X - 1].re;
		  MinIndex = Y * 3 + X;
	       }
	 if (MinIndex == 4)
	    Water[y][x] = (++MinCount);
	 else
	    Water[y][x] = -MinIndex;
      }
   fprintf(Fd, "%f %d\n", StdDev, MinCount);
   if (Debug == TRUE)
      printf("\t%f\t%d\n", StdDev, MinCount);

   /* Process list of previous minima */
   for (i = 1; i <= Count; i++)
      if (Level[i] == 0)
      {
	 /* Walk gradient downhill to minima */
	 X = Xchild[i];
	 Y = Ychild[i];
	 while (Water[Y][X] <= 0)
	 {
	    switch (Water[Y][X])
	    {
	    case 0:
	       X--;
	       Y--;
	       break;
	    case -1:
	       Y--;
	       break;
	    case -2:
	       X++;
	       Y--;
	       break;
	    case -3:
	       X--;
	       break;
	    case -5:
	       X++;
	       break;
	    case -6:
	       X--;
	       Y++;
	       break;
	    case -7:
	       Y++;
	       break;
	    case -8:
	       X++;
	       Y++;
	       break;
	    }
	 }

	 /* Check for link to image boundary */
	 if ((X == 0) || (Y == 0) || (X == Xdim - 1) || (Y == Ydim - 1))
	    X = Y = 0;

	 /* Save link information */
	 Xparent[i] = X;
	 Yparent[i] = Y;
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine finds shortest links between minima.               */
/*---------------------------------------------------------------------------*/
void ShortestLinks(int BlurLevel)
{
   int i, j, MinIndex;
   int Dx, Dy, Dist, MinDist;

   /* Identify annihilations */
   for (i = 1; i <= Count; i++)
      if (Level[i] == 0)
      {
	 /* Find first link distance */
	 Dx = Xchild[i] - Xparent[i];
	 Dy = Ychild[i] - Yparent[i];
	 MinDist = Dx * Dx + Dy * Dy;
	 MinIndex = i;

	 /* Search for shorter links */
	 for (j = 1; j <= Count; j++)
	    if ((Xparent[i] == Xparent[j]) &&
		(Yparent[i] == Yparent[j]) &&
		(Level[j] == 0))
	    {
	       Dx = Xchild[j] - Xparent[j];
	       Dy = Ychild[j] - Yparent[j];
	       Dist = Dx * Dx + Dy * Dy;
	       if (Dist < MinDist)
	       {
		  MinDist = Dist;
		  MinIndex = j;
	       }
	    }

	 /* Process annihilations */
	 if (i != MinIndex)
	 {
	    Level[i] = BlurLevel;
	    Parent[i] = MinIndex;
	 }
	 else
	 {
	    Xchild[i] = Xparent[i];
	    Ychild[i] = Yparent[i];
	 }
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine finds scale between two regions.                   */
/*---------------------------------------------------------------------------*/
int FindScale(int region1, int region2)
{
   int parent1[100], scale1[100], count1;
   int parent2[100], scale2[100], count2;

   /* Record sequence of parents for region1 */
   parent1[0] = region1;
   scale1[0] = 0;
   count1 = 1;
   while (Parent[region1] > 0)
   {
      scale1[count1] = Level[region1];
      parent1[count1] = Parent[region1];
      region1 = Parent[region1];
      count1++;
   }

   /* Record sequence of parents for region2 */
   parent2[0] = region2;
   scale2[0] = 0;
   count2 = 1;
   while (Parent[region2] > 0)
   {
      scale2[count2] = Level[region2];
      parent2[count2] = Parent[region2];
      region2 = Parent[region2];
      count2++;
   }

   /* Search for lowest scale region in common */
   for (region1 = 0; region1 < count1; region1++)
      for (region2 = 0; region2 < count2; region2++)
	 if (parent1[region1] == parent2[region2])
	 {
	    if (scale1[region1] > scale2[region2])
	       return (scale1[region1]);
	    else
	       return (scale2[region2]);
	 }
   return (BlurCount + 10);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine determines the watershed boundary scales.          */
/*---------------------------------------------------------------------------*/
void BoundaryScale(SHORT_TYPE ** Water, SHORT_TYPE ** Bound, int Xdim, int Ydim)
{
   int x, y;
   int temp1, temp2;

   /* Locate internal boundaries */
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 temp1 = temp2 = 0;
	 if (Water[y][x] != Water[y][x + 1])
	    temp1 = FindScale(Water[y][x], Water[y][x + 1]);
	 if (Water[y][x] != Water[y + 1][x])
	    temp2 = FindScale(Water[y][x], Water[y + 1][x]);
	 if (temp1 > temp2)
	    Bound[y][x] = temp1;
	 else
	    Bound[y][x] = temp2;
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine initializes the stddev-mincnt table.               */
/*---------------------------------------------------------------------------*/
void ReadTable(char *Name, int Xdim, int Ydim)
{
   /* Read input file until EOF */
   if ((Fd = fopen(Name, "r")) != NULL)
      while ((fscanf(Fd, "%f %f", &StdDev[TableSize], &MinCnt[TableSize])) != EOF)
      {
	 if (Debug == TRUE)
	    printf("%f\t%f\n", StdDev[TableSize], MinCnt[TableSize]);
	 TableSize++;
      }

   /* Handle EMPTY input file */
   if (TableSize < 2)
   {
      StdDev[0] = Xdim / 2.0;
      StdDev[1] = Xdim / 64.0;
      StdDev[1] = 4.0;
      MinCnt[0] = Xdim * Ydim / 15.0;
      MinCnt[1] = Xdim * Ydim / 600.0;
      MinCnt[1] = 100;
      TableSize = 2;
   }

   /* Create output data file */
   if ((Fd = fopen(Name, "w")) == NULL)
      Error("Can not write data file");
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine looks up a sigma for a given mincnt.               */
/*---------------------------------------------------------------------------*/
float GetSigma(float Minima)
{
   int i = 0;

   while ((i < TableSize) && (MinCnt[i] > Minima))
      i++;
   if (i == 0)
      i = 1;

   return (StdDev[i - 1] + (StdDev[i] - StdDev[i - 1]) *
	   (Minima - MinCnt[i - 1]) / (MinCnt[i] - MinCnt[i - 1]));
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   FLOAT_TYPE **Data1;
   COMPLEX_TYPE **Blur1;
   FLOAT_TYPE **Temp1;
   SHORT_TYPE **Link1;
   SHORT_TYPE **Bound1;
   SHORT_TYPE **Water1;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int i = 0, x, y;
   float Sigma;
   float Minima;
   float Percent;
   int BlurLevel;

   /* Interpret program options */
   fprintf(stderr, "LINK_EXTREMA program - Version 3.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      case 'g':
	 Gradient = FALSE;
	 break;
      case 'n':
	 if (sscanf(argv[++i], "%d", &BlurCount) == 0)
	    Error("Could not get BlurCount argument");
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr,
	      "Usage: link_extrema [options] infile outprefix\n");
      fprintf(stderr, "\t[-d]    Debugging flag\n");
      fprintf(stderr, "\t[-g]    Turn off gradient watersheds\n");
      fprintf(stderr, "\t[-n #]  Number of blurring steps (25)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output prefix name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Read the stddev-mincnt data file */
   sprintf(Name3, "%s.data", Name2);
   ReadTable(Name3, Xdim, Ydim);

   /* Create output image and temporary images */
   sprintf(Name3, "%s.out", Name2);
   Image2 = im_create(Name3, SHORT, Xdim, Ydim, Zdim);
   Blur1 = (COMPLEX_TYPE **) im_alloc2D(Image2, COMPLEX);
   Temp1 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   Link1 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);
   Bound1 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);
   Water1 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);

   /* Prepare to blur image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Blur1[y][x].re = Data1[y][x];
	 Blur1[y][x].im = 0;
      }

   /* Perform first level of blurring */
   Percent = (float) exp(log((double) MinCnt[TableSize - 1]
			     / MinCnt[0]) / BlurCount);
   Minima = MinCnt[0];
   Sigma = StdDev[0];
   if (Debug == TRUE)
      printf("percent %f\n", Percent);
   if (Debug == TRUE)
      printf("%d\t%d", 0, (int) Minima);
   Gauss(Blur1, Xdim, Ydim, Sigma);

   /* Locate initial minima and watersheds */
   if (Gradient == TRUE)
      GradMagnitude(Blur1, Temp1, Xdim, Ydim);
   FindMinima(Blur1, Xdim, Ydim, Sigma);
   Watershed(Blur1, Water1, Xdim, Ydim);

   /* Output Watershed region image */
   sprintf(Name3, "%s.regions", Name2);
   Image3 = im_create(Name3, SHORT, Xdim, Ydim, Zdim);
   im_write(Image3, SHORT, (char *) &(Water1[0][0]));

   /* Perform blurring and linking cycle */
   for (BlurLevel = 1; BlurLevel <= BlurCount; BlurLevel++)
   {
      /* Prepare to blur image */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    Blur1[y][x].re = Data1[y][x];
	    Blur1[y][x].im = 0;
	 }

      /* Perform next level of blurring */
      Minima *= Percent;
      Sigma = GetSigma(Minima);
      if (Debug == TRUE)
	 printf("%d\t%d ", BlurLevel, (int) Minima);
      Gauss(Blur1, Xdim, Ydim, Sigma);

      /* Locate and link minima */
      if (Gradient == TRUE)
	 GradMagnitude(Blur1, Temp1, Xdim, Ydim);
      LinkMinima(Blur1, Link1, Xdim, Ydim, Sigma);
      ShortestLinks(BlurLevel);
   }

   /* Print current region hierarchy */
   fclose(Fd);
   sprintf(Name3, "%s.links", Name2);
   Fd = fopen(Name3, "w");
   for (i = 0; i < Count; i++)
      if (Parent[i] > 0)
	 fprintf(Fd, "%d %d %d\n", i, Parent[i], Level[i]);
      else
	 fprintf(Fd, "%d %d %d\n", i, Parent[i], BlurCount);

   /* Generate watershed boundary image */
   BoundaryScale(Water1, Bound1, Xdim, Ydim);

   /* Write output image */
   im_write(Image2, SHORT, (char *) &(Bound1[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Blur1);
   im_free2D((char **) Temp1);
   im_free2D((char **) Link1);
   im_free2D((char **) Water1);
   im_free2D((char **) Bound1);
   fclose(Fd);
   return (0);
}
