/*--------------------------------------------------------------------------*/
/* Program:  ml_segment.c                                                   */
/*                                                                          */
/* Purpose:  This program performs maximum likelihood segmentation of       */
/*           images where the user specifies the desired number of classes. */
/*           The program currently works with either a 2D scalar image      */
/*           or a 3D multispectral image (eg. an MRI image).                */
/*                                                                          */
/*           NOTE: The 2D algorithm can be made much faster.                */
/*                                                                          */
/* Author:   Yu Zhao, John Gauch, Felix Fernandes                           */
/*                                                                          */
/* Date:     July 29, 1996     - Original program.                          */
/*           February 24, 1997 - Revised version.                           */
/*           April, 26 1997    - Support for 2D and 3D images.              */
/*                                                                          */
/* Note:     Copyright (C) The University of Kansas, 1997                   */
/*--------------------------------------------------------------------------*/
#include <IM.h>
#define KMAX 10

/* Gaussian distribution parameters */
struct StatsRec
{
   float W;
   float U[3];
   float V[3][3];
}
Stats[KMAX];

/* Global variables */
FLOAT_TYPE **Input2d;
FLOAT_TYPE ***Input3d;
SHORT_TYPE **Output;
FLOAT_TYPE ***Prob;
int Xdim, Ydim, Zdim;
int Debug = FALSE;
int Number = 4;
int Loops = 25;
int Step = 2;
float Max = 100;

/*---------------------------------------------------------------------------*/
/* Purpose:  This function estimates model parameters for 2D images.         */
/*---------------------------------------------------------------------------*/
void Estimate2d()
{
   int x, y, k, n;
   double p[KMAX], DeltaU0, DETV;
   double g, LogG, TWOPI = 2 * M_PI;
   double Difference;

   /* Get maximum intensity value in input image */
/*--------------------------------------------*/
   Max = 0;
   for (y = 0; y < Ydim; y += Step)
      for (x = 0; x < Xdim; x += Step)
	 if (Input2d[y][x] > Max)
	    Max = Input2d[y][x];

   /* Set initial model parameters */
/*------------------------------*/
   for (k = 0; k < Number; k++)
   {
      Stats[k].W = (k + 1) / ((Number + 1) * Number * 0.5);
      Stats[k].U[0] = k * Max / Number + Max / (2 * Number);
      Stats[k].V[0][0] = Max;
      if (Debug == TRUE)
	 printf("Class %d Weight %f Mean %f Variance %f\n",
		k, Stats[k].W, Stats[k].U[0], Stats[k].V[0][0]);
   }

   /* Calculate initial p[k], g, and LogG */
/*----------------------------------*/
   LogG = 0;
   for (y = 0; y < Ydim; y += Step)
      for (x = 0; x < Xdim; x += Step)
      {
	 /* Loop over all classes */
	 g = 0;
	 for (k = 0; k < Number; k++)
	 {
	    /* Get determinant of V matrix */
	    DETV = Stats[k].V[0][0];

	    /* Get distance from mean */
	    DeltaU0 = Input2d[y][x] - Stats[k].U[0];

	    /* Get class probability */
	    p[k] = exp(-0.5 * DeltaU0 * DeltaU0 / Stats[k].V[0][0])
	       / sqrt(DETV * TWOPI);

	    /* Increment g */
	    g += p[k] * Stats[k].W;
	 }
	 /* Update LogG */
	 LogG += log(g);

	 /* Save normalized probability */
	 for (k = 0; k < Number; k++)
	    Prob[k][y][x] = p[k] * Stats[k].W / g;
      }

   /* Print initial info */
   if (Debug == TRUE)
      printf("Iteration %d logG %lf\n", 0, LogG);

   /* Iteratively improve model parameters */
/*--------------------------------------*/
   for (n = 0; n < Loops; n++)
   {
      /* Refine W U V for each class */
/*-----------------------------*/
      for (k = 0; k < Number; k++)
      {
	 /* Get new estimates of W and U */
	 Stats[k].W = 0;
	 Stats[k].U[0] = 0;
	 for (y = 0; y < Ydim; y += Step)
	    for (x = 0; x < Xdim; x += Step)
	    {
	       Stats[k].W += Prob[k][y][x];
	       Stats[k].U[0] += Prob[k][y][x] * Input2d[y][x];
	    }

	 /* Normalize U values */
	 Stats[k].U[0] /= Stats[k].W;

	 /* Get new estimate of V */
	 Stats[k].V[0][0] = 0;
	 for (y = 0; y < Ydim; y += Step)
	    for (x = 0; x < Xdim; x += Step)
	    {
	       Difference = Input2d[y][x] - Stats[k].U[0];
	       Stats[k].V[0][0] += Prob[k][y][x] * Difference * Difference;
	    }

	 /* Normalize W and V values */
	 Stats[k].V[0][0] /= Stats[k].W;
	 if (Stats[k].V[0][0] < 1)
	    Stats[k].V[0][0] = 1;
	 Stats[k].W /= (Xdim * Ydim) / (Step * Step);

	 /* Print new info */
	 if (Debug == TRUE)
	 {
	    printf("Class %d Weight %f Mean %f Variance %f\n",
		   k, Stats[k].W, Stats[k].U[0], Stats[k].V[0][0]);
	 }
      }

      /* Calculate new p[k], g, and LogG */
/*------------------------------*/
      LogG = 0;
      for (y = 0; y < Ydim; y += Step)
	 for (x = 0; x < Xdim; x += Step)
	 {
	    /* Loop over all classes */
	    g = 0;
	    for (k = 0; k < Number; k++)
	    {
	       /* Get determinant of V matrix */
	       DETV = Stats[k].V[0][0];

	       /* Get distance from mean */
	       DeltaU0 = Input2d[y][x] - Stats[k].U[0];

	       /* Get class probability */
	       p[k] = exp(-0.5 * DeltaU0 * DeltaU0 / Stats[k].V[0][0])
		  / sqrt(DETV * TWOPI);

	       /* Increment g */
	       g += p[k] * Stats[k].W;
	    }
	    /* Update LogG */
	    LogG += log(g);

	    /* Save normalized probability */
	    for (k = 0; k < Number; k++)
	       Prob[k][y][x] = p[k] * Stats[k].W / g;
	 }

      /* Print initial info */
      if (Debug == TRUE)
	 printf("Iteration %d logG %lf\n", 1 + n, LogG);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This function estimates model parameters for 3D images.         */
/*---------------------------------------------------------------------------*/
void Estimate3d()
{
   int x, y, i, j, k, n;
   double p[KMAX], DeltaU0, DeltaU1, DeltaU2, DETV;
   double g, LogG, TWOPI = 2 * M_PI;

   /* Set initial model parameters */
/*------------------------------*/
   for (k = 0; k < Number; k++)
   {
      Stats[k].W = (k + 1) / ((Number + 1) * Number * 0.5);
      if (Debug == TRUE)
	 printf("Class %d Weight %f\n", k, Stats[k].W);
      for (i = 0; i < 3; i++)
      {
	 Stats[k].U[i] = k * Max / Number + Max / (2 * Number);
	 for (j = 0; j < 3; j++)
	    Stats[k].V[i][j] = 0.0;
	 Stats[k].V[i][i] = Max;
	 if (Debug == TRUE)
	    printf("Dimension %d Mean %f Variance %f\n",
		   i, Stats[k].U[i], Stats[k].V[i][i]);
      }
   }

   /* Calculate initial p[k], g, and LogG */
/*----------------------------------*/
   LogG = 0;
   for (y = 0; y < Ydim; y += Step)
      for (x = 0; x < Xdim; x += Step)
      {
	 /* Loop over all classes */
	 g = 0;
	 for (k = 0; k < Number; k++)
	 {
	    /* Get determinant of V matrix */
	    DETV = Stats[k].V[0][0] * Stats[k].V[1][1] * Stats[k].V[2][2];

	    /* Get distance from mean */
	    DeltaU0 = Input3d[0][y][x] - Stats[k].U[0];
	    DeltaU1 = Input3d[1][y][x] - Stats[k].U[1];
	    DeltaU2 = Input3d[2][y][x] - Stats[k].U[2];

	    /* Get class probability */
	    p[k] = exp((DeltaU0 * DeltaU0 / Stats[k].V[0][0]
			+ DeltaU1 * DeltaU1 / Stats[k].V[1][1]
			+ DeltaU2 * DeltaU2 / Stats[k].V[2][2])
		       * (-0.5))
	       / (sqrt(DETV) * sqrt(TWOPI) * TWOPI);

	    /* Increment g */
	    g += p[k] * Stats[k].W;
	 }
	 /* Update LogG */
	 LogG += log(g);

	 /* Save normalized probability */
	 for (k = 0; k < Number; k++)
	    Prob[k][y][x] = p[k] * Stats[k].W / g;
      }

   /* Print initial info */
   if (Debug == TRUE)
      printf("Iteration %d logG %lf\n", 0, LogG);

   /* Iteratively improve model parameters */
/*--------------------------------------*/
   for (n = 0; n < Loops; n++)
   {
      /* Refine W U V for each class */
/*-----------------------------*/
      for (k = 0; k < Number; k++)
      {
	 /* Get new estimates of W and U */
	 Stats[k].W = 0;
	 for (i = 0; i < 3; i++)
	    Stats[k].U[i] = 0;
	 for (y = 0; y < Ydim; y += Step)
	    for (x = 0; x < Xdim; x += Step)
	    {
	       Stats[k].W += Prob[k][y][x];
	       for (i = 0; i < 3; i++)
		  Stats[k].U[i] += Prob[k][y][x] * Input3d[i][y][x];
	    }

	 /* Normalize U values */
	 for (i = 0; i < 3; i++)
	    Stats[k].U[i] /= Stats[k].W;

	 /* Get new estimate of V */
	 for (i = 0; i < 3; i++)
	    for (j = 0; j < 3; j++)
	       Stats[k].V[i][j] = 0;
	 for (y = 0; y < Ydim; y += Step)
	    for (x = 0; x < Xdim; x += Step)
	    {
	       for (i = 0; i < 3; i++)
		  for (j = 0; j < 3; j++)
		     Stats[k].V[i][j] += Prob[k][y][x] *
			(Input3d[i][y][x] - Stats[k].U[i]) *
			(Input3d[j][y][x] - Stats[k].U[j]);
	    }

	 /* Normalize W and V values */
	 for (i = 0; i < 3; i++)
	    for (j = 0; j < 3; j++)
	    {
	       Stats[k].V[i][j] /= Stats[k].W;
	       if (Stats[k].V[i][j] < 1)
		  Stats[k].V[i][j] = 1;
	    }
	 Stats[k].W /= (Xdim * Ydim) / (Step * Step);

	 /* Print new info */
	 if (Debug == TRUE)
	 {
	    printf("Class %d Weight %f\n", k, Stats[k].W);
	    for (i = 0; i < 3; i++)
	       printf("Dimension %d Mean %f Variance %f\n",
		      i, Stats[k].U[i], Stats[k].V[i][i]);
	 }
      }

      /* Calculate new p[k], g, and LogG */
/*------------------------------*/
      LogG = 0;
      for (y = 0; y < Ydim; y += Step)
	 for (x = 0; x < Xdim; x += Step)
	 {
	    /* Loop over all classes */
	    g = 0;
	    for (k = 0; k < Number; k++)
	    {
	       /* Get determinant of V matrix */
	       DETV = Stats[k].V[0][0] * Stats[k].V[1][1] * Stats[k].V[2][2];

	       /* Get distance from mean */
	       DeltaU0 = Input3d[0][y][x] - Stats[k].U[0];
	       DeltaU1 = Input3d[1][y][x] - Stats[k].U[1];
	       DeltaU2 = Input3d[2][y][x] - Stats[k].U[2];

	       /* Get class probability */
	       p[k] = exp((DeltaU0 * DeltaU0 / Stats[k].V[0][0]
			   + DeltaU1 * DeltaU1 / Stats[k].V[1][1]
			   + DeltaU2 * DeltaU2 / Stats[k].V[2][2])
			  * (-0.5))
		  / (sqrt(DETV) * sqrt(TWOPI) * TWOPI);

	       /* Increment g */
	       g += p[k] * Stats[k].W;
	    }
	    /* Update LogG */
	    LogG += log(g);

	    /* Save normalized probability */
	    for (k = 0; k < Number; k++)
	       Prob[k][y][x] = p[k] * Stats[k].W / g;
	 }

      /* Print initial info */
      if (Debug == TRUE)
	 printf("Iteration %d logG %lf\n", 1 + n, LogG);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This function performs ML segmentation for 2D images.           */
/*---------------------------------------------------------------------------*/
void Segment2d()
{
   int x, y, k, MaxK;
   double p, MaxP, DeltaU0, DETV;
   double TWOPI = 2 * M_PI;

   /* Process all pixels in the image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 /* Calculate maximum likelihood class */
	 MaxP = MaxK = -1;
	 for (k = 0; k < Number; k++)
	 {
	    /* Get determinant of V matrix */
	    DETV = Stats[k].V[0][0];

	    /* Get distance from mean */
	    DeltaU0 = Input2d[y][x] - Stats[k].U[0];

	    /* Get class probability */
	    p = exp(-0.5 * DeltaU0 * DeltaU0 / Stats[k].V[0][0])
	       / sqrt(DETV * TWOPI);

	    /* Check to see if maximum */
	    if (p > MaxP)
	    {
	       MaxP = p;
	       MaxK = k;
	    }
	 }
	 Output[y][x] = MaxK;
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This function performs ML segmentation for 3D images.           */
/*---------------------------------------------------------------------------*/
void Segment3d()
{
   int x, y, k, MaxK;
   double p, MaxP, DeltaU0, DeltaU1, DeltaU2, DETV;
   double TWOPI = 2 * M_PI;

   /* Process all pixels in the image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 /* Calculate maximum likelihood class */
	 MaxP = MaxK = -1;
	 for (k = 0; k < Number; k++)
	 {
	    /* Get determinant of V matrix */
	    DETV = Stats[k].V[0][0] * Stats[k].V[1][1] * Stats[k].V[2][2];

	    /* Get distance from mean */
	    DeltaU0 = Input3d[0][y][x] - Stats[k].U[0];
	    DeltaU1 = Input3d[1][y][x] - Stats[k].U[1];
	    DeltaU2 = Input3d[2][y][x] - Stats[k].U[2];

	    /* Get class probability */
	    p = exp((DeltaU0 * DeltaU0 / Stats[k].V[0][0]
		     + DeltaU1 * DeltaU1 / Stats[k].V[1][1]
		     + DeltaU2 * DeltaU2 / Stats[k].V[2][2])
		    * (-0.5))
	       / (sqrt(DETV) * sqrt(TWOPI) * TWOPI);

	    /* Check to see if maximum */
	    if (p > MaxP)
	    {
	       MaxP = p;
	       MaxK = k;
	    }
	 }
	 Output[y][x] = MaxK;
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   int PixType, DimCnt;

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("ML_SEGMENT Program - KUIM Version 3.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'k':
	 if (sscanf(argv[++i], "%d", &Number) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'l':
	 if (sscanf(argv[++i], "%d", &Loops) == 0)
	    Error("Could not get integer argument");
	 break;
      case 's':
	 if (sscanf(argv[++i], "%d", &Step) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: ml_segment [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-k #]  Number of output classes (4)\n");
      fprintf(stderr, "       [-l #]  Number of ML iterations (25)\n");
      fprintf(stderr, "       [-s #]  Step size for sampling (2)\n");
      return (1);
   }
   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt == 2)
   {
      Input2d = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
      im_read(Image1, FLOAT, (char *) &(Input2d[0][0]));
   }
   else if ((DimCnt == 3) && (Zdim == 3))
   {
      Input3d = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
      im_read(Image1, FLOAT, (char *) &(Input3d[0][0][0]));
   }
   else
      Error("Can not process image");

   /* Create output image */
   Image2 = im_create(Name2, SHORT, Xdim, Ydim, 1);
   Output = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);

   /* Create probability image */
   Image3 = im_create("/dev/null", FLOAT, Xdim, Ydim, Number);
   Prob = (FLOAT_TYPE ***) im_alloc3D(Image3, FLOAT);

   /* Perform ML segmentation of image */
   if (DimCnt == 2)
   {
      Estimate2d();
      Segment2d();
      im_free2D((char **) Input2d);
   }
   else if ((DimCnt == 3) && (Zdim == 3))
   {
      Estimate3d();
      Segment3d();
      im_free3D((char ***) Input3d);
   }

   /* Write information to output image */
   im_write(Image2, SHORT, (char *) &(Output[0][0]));
   im_write(Image3, FLOAT, (char *) &(Prob[0][0][0]));
   im_free2D((char **) Output);
   im_free3D((char ***) Prob);
   return (0);
}
