/*---------------------------------------------------------------------------*/
/* Program:  blob_color.c                                                    */
/*                                                                           */
/* Purpose:  This is a non recursive blob coloring program.                  */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     February 13, 1997                                               */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1997                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define SIZE 1000

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(argc, argv)
     int argc;
     char *argv[];
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   SHORT_TYPE **data1;
   SHORT_TYPE **data2;
   int Xdim, Ydim, Zdim;
   int PixType, DimCnt;

   /* Program variables */
   int parent[SIZE];
   int Xmean[SIZE];
   int Ymean[SIZE];
   int Size[SIZE];
   int Debug = FALSE;
   int Reverse = FALSE;
   int Color = 1;
   int Count = 0;
   int i = 0, x, y, c1, c2;

   /* Interpret program options */
   printf("BLOB_COLOR Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'r':
	 Reverse = TRUE;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: blob_color [options] infile outfile\n");
      fprintf(stderr, "       [-d]  Print debugging information\n");
      fprintf(stderr, "       [-r]  Reverse 0/1 before blob coloring\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Open input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can only process 2D images");
   data1 = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   im_read(Image1, SHORT, (char *) &(data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, SHORT, Xdim, Ydim, Zdim);
   data2 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);

   /* Preprocess the image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 if (data1[y][x] > 1)
	    data1[y][x] = 1;
	 if (data1[y][x] < 0)
	    data1[y][x] = 0;
	 if (Reverse == TRUE)
	    data1[y][x] = 1 - data1[y][x];
	 data2[y][x] = 0;
      }

   /* Mark the blobs in the input image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 if (data1[y][x] == 1)
	 {
	    /* Handle merger of two blobs */
	    if ((x > 0) && (data2[y][x - 1] > 0) &&
		(y > 0) && (data2[y - 1][x] > 0) &&
		(data2[y][x - 1] != data2[y - 1][x]))
	    {
	       /* Make one parent point to the other */
	       c1 = data2[y][x - 1];
	       c2 = data2[y - 1][x];
	       while (parent[c1] > 0)
		  c1 = parent[c1];
	       while (parent[c2] > 0)
		  c2 = parent[c2];
	       if (c1 != c2)
		  parent[c1] = c2;

	       /* Mark pixel with blob color */
	       c1 = data2[y][x - 1];
	       Size[c1]++;
	       Xmean[c1] += x;
	       Ymean[c1] += y;
	       data2[y][x] = c1;
	       if (Debug == TRUE)
		  printf("merge %d and %d\n", data2[y - 1][x], data2[y][x - 1]);
	    }

	    /* Handle blob to the left */
	    else if ((x > 0) && (data2[y][x - 1] > 0))
	    {
	       c1 = data2[y][x - 1];
	       Size[c1]++;
	       Xmean[c1] += x;
	       Ymean[c1] += y;
	       data2[y][x] = c1;
	    }

	    /* Handle blob above */
	    else if ((y > 0) && (data2[y - 1][x] > 0))
	    {
	       c1 = data2[y - 1][x];
	       Size[c1]++;
	       Xmean[c1] += x;
	       Ymean[c1] += y;
	       data2[y][x] = c1;
	    }

	    /* Handle new blob */
	    else
	    {
	       Size[Color] = 1;
	       Xmean[Color] = 0;
	       Ymean[Color] = 0;
	       parent[Color] = -1;
	       data2[y][x] = Color++;
	       if (Debug == TRUE)
		  printf("Start of blob %d at position %d %d\n", Color, x, y);
	    }
	 }

   /* Renumber blobs and update statistics */
   Count = -1;
   for (c1 = 1; c1 < Color; c1++)
   {
      if (parent[c1] < 0)
	 parent[c1] = Count--;
      else
      {
	 c2 = c1;
	 while (parent[c2] > 0)
	    c2 = parent[c2];
	 Size[c2] += Size[c1];
	 Xmean[c2] += Xmean[c1];
	 Ymean[c2] += Ymean[c1];
      }
   }

   /* Output blob statistics */
   for (c1 = 1; c1 < Color; c1++)
      if (parent[c1] < 0)
	 printf("Blob %d Size %d Center %d %d\n", -parent[c1], Size[c1],
		Xmean[c1] / Size[c1], Ymean[c1] / Size[c1]);

   /* Save final blob colors */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 c1 = data2[y][x];
	 while (parent[c1] > 0)
	    c1 = parent[c1];
	 data2[y][x] = -parent[c1];
      }

   /* Write information to output image */
   im_write(Image2, SHORT, (char *) &(data2[0][0]));
   im_free2D((char **) data1);
   im_free2D((char **) data2);
   exit(0);
}
