/*--------------------------------------------------------------------------*/
/* Program:  optic_flow.c                                                   */
/*                                                                          */
/* Purpose:  This program calculates the optic flow field for a pair        */
/*           of images taken from a motion sequence.  Either template       */
/*           matching or Horn's variational approach can be used.           */
/*                                                                          */
/* Author:   John Gauch, Edu Metz, Jacques Bride                            */
/*                                                                          */
/* Usage:    optic_flow [options] infile1 infile2 outfileU outfileV         */
/*                      [-d] Print debugging information                    */
/*                                                                          */
/* Date:     March 24, 1994  -  Original program.                           */
/*           June 3, 1997    -  Updated Algorithm.                          */
/*                                                                          */
/* Note:     Copyright (C) The University of Kansas, 1994-1997              */
/*--------------------------------------------------------------------------*/

#include <IM.h>
#define TEMPLATE   1
#define VARIATIONAL 2

IM_TYPE *Image1;
IM_TYPE *Image2;
IM_TYPE *ImageU;
IM_TYPE *ImageV;
FLOAT_TYPE **DataU;
FLOAT_TYPE **DataV;

int Debug = FALSE;
int Xdim, Ydim, Zdim, PixType, DimCnt;
int Method = VARIATIONAL;
int SearchSize = 3;
int BlockSize = 3;
int Count = 30;
float Alpha = 0.05;

/*---------------------------------------------------------------------------*/
/* Purpose:  Block matching optic flow for 2D Grey images.                   */
/*---------------------------------------------------------------------------*/
void GreyMatch2d()
{
   int x, y, dx, dy, bx, by, minU, minV;
   float Temp, Diff, MinDiff;
   FLOAT_TYPE **Data1, **Data2;

   /* Read input */
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));
   im_read(Image2, FLOAT, (char *) &(Data2[0][0]));

   /* Initialize output images */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Calculate optic flow */
   for (y = SearchSize + BlockSize; y < Ydim - SearchSize - BlockSize; y++)
      for (x = SearchSize + BlockSize; x < Xdim - SearchSize - BlockSize; x++)
      {
	 /* Set initial match error */
	 MinDiff = 0;
	 minU = minV = 0;
	 for (by = -BlockSize; by <= BlockSize; by++)
	    for (bx = -BlockSize; bx <= BlockSize; bx++)
	    {
	       Temp = Data2[y + by][x + bx]
		  - Data1[y + by][x + bx];
	       if (Temp < 0)
		  MinDiff -= Temp;
	       else
		  MinDiff += Temp;
	    }

	 /* Search for best block match at each point */
	 for (dy = -SearchSize; dy <= SearchSize; dy++)
	    for (dx = -SearchSize; dx <= SearchSize; dx++)
	    {
	       Diff = 0;
	       for (by = -BlockSize; by <= BlockSize; by++)
		  for (bx = -BlockSize; bx <= BlockSize; bx++)
		  {
		     Temp = Data2[y + by][x + bx]
			- Data1[y + by - dy][x + bx - dx];
		     if (Temp < 0)
			Diff -= Temp;
		     else
			Diff += Temp;
		  }
	       if (Diff < MinDiff)
	       {
		  MinDiff = Diff;
		  minU = dx;
		  minV = dy;
	       }
	    }
	 DataU[y][x] = minU;
	 DataV[y][x] = minV;
      }

   /* Write information to output image */
   im_write(ImageU, FLOAT, (char *) &(DataU[0][0]));
   im_write(ImageV, FLOAT, (char *) &(DataV[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  Block matching optic flow for 3D Grey images.                   */
/*---------------------------------------------------------------------------*/
void GreyMatch3d()
{
   int x, y, z, dx, dy, bx, by, minU, minV;
   float Temp, Diff, MinDiff;
   FLOAT_TYPE ***Data1, ***Data2;

   /* Read input */
   Data1 = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
   Data2 = (FLOAT_TYPE ***) im_alloc3D(Image2, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0][0]));
   im_read(Image2, FLOAT, (char *) &(Data2[0][0][0]));

   /* Initialize output images */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Calculate optic flow */
   for (y = SearchSize + BlockSize; y < Ydim - SearchSize - BlockSize; y++)
      for (x = SearchSize + BlockSize; x < Xdim - SearchSize - BlockSize; x++)
      {
	 /* Set initial match error */
	 MinDiff = minU = minV = 0;
	 for (by = -BlockSize; by <= BlockSize; by++)
	    for (bx = -BlockSize; bx <= BlockSize; bx++)
	       for (z = 0; z < Zdim; z++)
	       {
		  Temp = Data2[z][y + by][x + bx]
		     - Data1[z][y + by][x + bx];
		  if (Temp < 0)
		     MinDiff -= Temp;
		  else
		     MinDiff += Temp;
	       }

	 /* Search for best block match at each point */
	 for (dy = -SearchSize; dy <= SearchSize; dy++)
	    for (dx = -SearchSize; dx <= SearchSize; dx++)
	    {
	       Diff = 0;
	       for (by = -BlockSize; by <= BlockSize; by++)
		  for (bx = -BlockSize; bx <= BlockSize; bx++)
		     for (z = 0; z < Zdim; z++)
		     {
			Temp = Data2[z][y + by][x + bx]
			   - Data1[z][y + by - dy][x + bx - dx];
			if (Temp < 0)
			   Diff -= Temp;
			else
			   Diff += Temp;
		     }
	       if (Diff < MinDiff)
	       {
		  MinDiff = Diff;
		  minU = dx;
		  minV = dy;
	       }
	    }
	 DataU[y][x] = minU;
	 DataV[y][x] = minV;
      }

   /* Write information to output image */
   im_write(ImageU, FLOAT, (char *) &(DataU[0][0]));
   im_write(ImageV, FLOAT, (char *) &(DataV[0][0]));
   im_free3D((char ***) Data1);
   im_free3D((char ***) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  Block matching optic flow for 2D Color images.                  */
/*---------------------------------------------------------------------------*/
void ColorMatch2d()
{
   int x, y, dx, dy, bx, by, minU, minV;
   float TempR, TempG, TempB, Diff, MinDiff;
   COLOR_TYPE **Data1, **Data2;

   /* Read input */
   Data1 = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
   Data2 = (COLOR_TYPE **) im_alloc2D(Image2, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0]));
   im_read(Image2, COLOR, (char *) &(Data2[0][0]));

   /* Initialize output images */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Calculate optic flow */
   for (y = SearchSize + BlockSize; y < Ydim - SearchSize - BlockSize; y++)
      for (x = SearchSize + BlockSize; x < Xdim - SearchSize - BlockSize; x++)
      {
	 /* Set initial match error */
	 MinDiff = minU = minV = 0;
	 for (by = -BlockSize; by <= BlockSize; by++)
	    for (bx = -BlockSize; bx <= BlockSize; bx++)
	    {
	       TempR = (float) Data2[y + by][x + bx].r -
		  (float) Data1[y + by][x + bx].r;
	       TempG = (float) Data2[y + by][x + bx].g -
		  (float) Data1[y + by][x + bx].g;
	       TempB = (float) Data2[y + by][x + bx].b -
		  (float) Data1[y + by][x + bx].b;
	       if (TempR < 0)
		  TempR = -TempR;
	       if (TempG < 0)
		  TempG = -TempG;
	       if (TempB < 0)
		  TempB = -TempB;
	       MinDiff += TempR + TempG + TempB;
	    }

	 /* Search for best block match at each point */
	 for (dy = -SearchSize; dy <= SearchSize; dy++)
	    for (dx = -SearchSize; dx <= SearchSize; dx++)
	    {
	       Diff = 0;
	       for (by = -BlockSize; by <= BlockSize; by++)
		  for (bx = -BlockSize; bx <= BlockSize; bx++)
		  {
		     TempR = (float) Data2[y + by][x + bx].r -
			(float) Data1[y + by - dy][x + bx - dx].r;
		     TempG = (float) Data2[y + by][x + bx].g -
			(float) Data1[y + by - dy][x + bx - dx].g;
		     TempB = (float) Data2[y + by][x + bx].b -
			(float) Data1[y + by - dy][x + bx - dx].b;
		     if (TempR < 0)
			TempR = -TempR;
		     if (TempG < 0)
			TempG = -TempG;
		     if (TempB < 0)
			TempB = -TempB;
		     Diff += TempR + TempG + TempB;
		  }
	       if (Diff < MinDiff)
	       {
		  MinDiff = Diff;
		  minU = dx;
		  minV = dy;
	       }
	    }
	 DataU[y][x] = minU;
	 DataV[y][x] = minV;
      }

   /* Write information to output image */
   im_write(ImageU, FLOAT, (char *) &(DataU[0][0]));
   im_write(ImageV, FLOAT, (char *) &(DataV[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  Variational optic flow for 2D Grey images.                      */
/*---------------------------------------------------------------------------*/
void GreyVariational2d()
{
   int x, y, i;
   float Ix, Iy, It;
   float Error, Step;
   FLOAT_TYPE **Data1, **Data2;
   FLOAT_TYPE **AveU, **AveV;

   /* Read input */
   AveU = (FLOAT_TYPE **) im_alloc2D(ImageU, FLOAT);
   AveV = (FLOAT_TYPE **) im_alloc2D(ImageV, FLOAT);
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));
   im_read(Image2, FLOAT, (char *) &(Data2[0][0]));

   /* Initialize output images */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Calculate optic flow */
   for (i = 0; i < Count; i++)
   {
      /* Calculate average (u,v) values */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    AveU[y][x] = (DataU[y - 1][x] + DataU[y + 1][x]
			  + DataU[y][x - 1] + DataU[y][x + 1]) / 4;
	    AveV[y][x] = (DataV[y - 1][x] + DataV[y + 1][x]
			  + DataV[y][x - 1] + DataV[y][x + 1]) / 4;
	 }

      /* Loop over image */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    /* Calculate partial derivatives */
	    Ix = (Data1[y][x + 1] - Data1[y][x]
		  + Data1[y + 1][x + 1] - Data1[y + 1][x]
		  + Data2[y][x + 1] - Data2[y][x]
		  + Data2[y + 1][x + 1] - Data2[y + 1][x]) / 4;
	    Iy = (Data1[y + 1][x] - Data1[y][x]
		  + Data1[y + 1][x + 1] - Data1[y][x + 1]
		  + Data2[y + 1][x] - Data2[y][x]
		  + Data2[y + 1][x + 1] - Data2[y][x + 1]) / 4;
	    It = (Data2[y][x] - Data1[y][x]
		  + Data2[y + 1][x] - Data1[y + 1][x]
		  + Data2[y][x + 1] - Data1[y][x + 1]
		  + Data2[y + 1][x + 1] - Data1[y + 1][x + 1]) / 4;

	    /* Calculate new (u,v) values */
	    Error = (Ix * AveU[y][x] + Iy * AveV[y][x] + It);
	    Step = Alpha / (1 + Alpha * Ix * Ix + Alpha * Iy * Iy);
	    DataU[y][x] = AveU[y][x] - Step * Error * Ix;
	    DataV[y][x] = AveV[y][x] - Step * Error * Iy;
	 }
   }

   /* Write information to output image */
   im_write(ImageU, FLOAT, (char *) &(DataU[0][0]));
   im_write(ImageV, FLOAT, (char *) &(DataV[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  Variational optic flow for 3D Grey images.                      */
/*---------------------------------------------------------------------------*/
void GreyVariational3d()
{
   int x, y;
   FLOAT_TYPE ***Data1, ***Data2;

   /* Read input */
   Data1 = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
   Data2 = (FLOAT_TYPE ***) im_alloc3D(Image2, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0][0]));
   im_read(Image2, FLOAT, (char *) &(Data2[0][0][0]));

   /* Initialize output images */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Calculate optic flow */
   /* TO BE ADDED */

   /* Write information to output image */
   im_write(ImageU, FLOAT, (char *) &(DataU[0][0]));
   im_write(ImageV, FLOAT, (char *) &(DataV[0][0]));
   im_free3D((char ***) Data1);
   im_free3D((char ***) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  Variational optic flow for 2D Color images.                     */
/*---------------------------------------------------------------------------*/
void ColorVariational2d()
{
   int x, y;
   COLOR_TYPE **Data1, **Data2;

   /* Read input */
   Data1 = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
   Data2 = (COLOR_TYPE **) im_alloc2D(Image2, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0]));
   im_read(Image2, COLOR, (char *) &(Data2[0][0]));

   /* Initialize output images */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Calculate optic flow */
   /* TO BE ADDED */

   /* Write information to output image */
   im_write(ImageU, FLOAT, (char *) &(DataU[0][0]));
   im_write(ImageV, FLOAT, (char *) &(DataV[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/** Purpose:  This is the main program.                                      */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char NameU[50];
   char NameV[50];

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("OPTIC_FLOW Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 's':
	 if (sscanf(argv[++i], "%d", &SearchSize) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'b':
	 if (sscanf(argv[++i], "%d", &BlockSize) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'c':
	 if (sscanf(argv[++i], "%d", &Count) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'w':
	 if (sscanf(argv[++i], "%f", &Alpha) == 0)
	    Error("Could not get float argument");
	 break;
      case 'T':
	 Method = TEMPLATE;
	 break;
      case 'V':
	 Method = VARIATIONAL;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 4)
   {
      fprintf(stderr, "Usage: optic_flow [options] in1 in2 outU outV\n");
      fprintf(stderr, "       [-T]    Use template matching\n");
      fprintf(stderr, "       [-s #]  Search region (template match)\n");
      fprintf(stderr, "       [-b #]  Block size (template match)\n");
      fprintf(stderr, "       [-V]    Use variational method\n");
      fprintf(stderr, "       [-c #]  Iteration count (variational)\n");
      fprintf(stderr, "       [-w #]  Smoothing weight (variational)\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      return (1);
   }
   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", NameU) == 0)
      Error("Could not get output file name");
   if (sscanf(argv[i++], "%s", NameV) == 0)
      Error("Could not get output file name");

   /* Open input images */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   Image2 = im_open(Name2, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);

   /* Create output images */
   ImageU = im_create(NameU, FLOAT, Xdim, Ydim, 1);
   ImageV = im_create(NameV, FLOAT, Xdim, Ydim, 1);
   DataU = (FLOAT_TYPE **) im_alloc2D(ImageU, FLOAT);
   DataV = (FLOAT_TYPE **) im_alloc2D(ImageV, FLOAT);

   /* Call appropriate optic flow function */
   if (Method == TEMPLATE)
   {
      if (DimCnt == 3)
	 GreyMatch3d();
      else if ((PixType == COLOR) || (PixType == PSEUDO))
	 ColorMatch2d();
      else
	 GreyMatch2d();
   }

   else if (Method == VARIATIONAL)
   {
      if (DimCnt == 2)
	 GreyVariational2d();
   }

   /* Write information to output image */
   im_free2D((char **) DataU);
   im_free2D((char **) DataV);
   return (0);
}
