/*---------------------------------------------------------------------------*/
/* Program:  photometric.c                                                   */
/*                                                                           */
/* Purpose:  This program calculates shape from shading using the            */
/*           photometric stereo approach.  Three shaded images which         */
/*           are illuminated from different points of view are input         */
/*           and the surface normal (-p, -q, 1) and depth z are              */
/*           calculated at each point assuming Lambertian reflectance.       */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     March 3, 1994                                                   */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Program variables */
FLOAT_TYPE **Data1;
FLOAT_TYPE **Data2;
FLOAT_TYPE **Data3;
FLOAT_TYPE **DataP;
FLOAT_TYPE **DataQ;
FLOAT_TYPE **DataZ;
int Debug = FALSE;
int Xdim, Ydim;
int Count = 20;
float Alpha = 2;
float Small = 0.02;
float Lx, Ly, Lz;
float Mx, My, Mz;
float Nx, Ny, Nz;
double sqrt();

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine uses 3 images for photometric stereo.              */
/*---------------------------------------------------------------------------*/
void Solve()
{
   int x, y;
   float i1, i2, i3;
   float a12, b12, c12, a31, b31, c31, a23, b23, c23;
   float t1, p1, q1, t2, p2, q2, t3, p3, q3;

   /* Scan whole image calculating (p,q) */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 /* Save image intensity values */
	 i1 = Data1[y][x];
	 i2 = Data2[y][x];
	 i3 = Data3[y][x];
	 if ((i1 > Small) && (i2 > Small) && (i3 > Small))
	 {
	    /* Intersect R1(p,q) with R2(p,q) for line12 */
	    a12 = (Lx * i2 - Mx * i1);
	    b12 = (Ly * i2 - My * i1);
	    c12 = (Mz * i1 - Lz * i2);

	    /* Intersect R2(p,q) with R3(p,q) for line23 */
	    a23 = (Mx * i3 - Nx * i2);
	    b23 = (My * i3 - Ny * i2);
	    c23 = (Nz * i2 - Mz * i3);

	    /* Intersect R3(p,q) with R1(p,q) for line31 */
	    a31 = (Lx * i3 - Nx * i1);
	    b31 = (Ly * i3 - Ny * i1);
	    c31 = (Nz * i1 - Lz * i3);

	    /* Intersect line12 and line31 to get (p1,q1) */
	    t1 = -(a31 * b12 - a12 * b31);
	    if (t1 != 0.0)
	       q1 = (a31 * c12 - a12 * c31) / t1;
	    else
	       q1 = 0.0;
	    t1 = -(b31 * a12 - b12 * a31);
	    if (t1 != 0.0)
	       p1 = (b31 * c12 - b12 * c31) / t1;
	    else
	       p1 = 0.0;

	    /* Intersect line12 and line23 to get (p2,q2) */
	    t2 = -(a23 * b12 - a12 * b23);
	    if (t2 != 0.0)
	       q2 = (a23 * c12 - a12 * c23) / t2;
	    else
	       q2 = 0.0;
	    t2 = -(b23 * a12 - b12 * a23);
	    if (t2 != 0.0)
	       p2 = (b23 * c12 - b12 * c23) / t2;
	    else
	       p2 = 0.0;

	    /* Intersect line31 and line23 to get (p3,q3) */
	    t3 = -(a23 * b31 - a31 * b23);
	    if (t3 != 0.0)
	       q3 = (a23 * c31 - a31 * c23) / t3;
	    else
	       q3 = 0.0;
	    t3 = -(b23 * a31 - b31 * a23);
	    if (t3 != 0.0)
	       p3 = (b23 * c31 - b31 * c23) / t3;
	    else
	       p3 = 0.0;

	    /* Save average (p,q) value */
	    DataP[y][x] = (p1 + p2 + p3) / 3;
	    DataQ[y][x] = (q1 + q2 + q3) / 3;
	 }
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine minimizes functional to fill in (p,q) values.      */
/*---------------------------------------------------------------------------*/
void Relax()
{
   int i, x, y;
   float i1, i2, i3, r1, r2, r3, p, q;
   float length, length3, dpr1, dqr1, dpr2, dqr2, dpr3, dqr3;

   /* Solve for (p,q) based on initial z */
   for (i = 0; i < Count; i++)
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    /* Save image intensity values */
	    i1 = Data1[y][x];
	    i2 = Data2[y][x];
	    i3 = Data3[y][x];

	    /* Fill in uncertain (p,q) values */
	    if ((i1 <= Small) || (i2 <= Small) || (i3 <= Small))
	    {
	       /* Calculate modeled intensity values */
	       p = DataP[y][x];
	       q = DataQ[y][x];
	       length = (float) sqrt((double) (p * p + q * q + 1));
	       r1 = (-Lx * p - Ly * q + Lz) / length;
	       r2 = (-Mx * p - My * q + Mz) / length;
	       r3 = (-Nx * p - Ny * q + Nz) / length;

	       /* Calculate partials of modeled intensity */
	       length3 = length * length * length;
	       dpr1 = Lx / length + p * (-Lx * p - Ly * q + Lz) / length3;
	       dqr1 = Ly / length + q * (-Lx * p - Ly * q + Lz) / length3;
	       dpr2 = Mx / length + p * (-Mx * p - My * q + Mz) / length3;
	       dqr2 = My / length + q * (-Mx * p - My * q + Mz) / length3;
	       dpr3 = Nx / length + p * (-Nx * p - Ny * q + Nz) / length3;
	       dqr3 = Ny / length + q * (-Nx * p - Ny * q + Nz) / length3;
	       if (r1 < 0.0)
		  r1 = dpr1 = dqr1 = 0.0;
	       if (r2 < 0.0)
		  r2 = dpr2 = dqr2 = 0.0;
	       if (r3 < 0.0)
		  r3 = dpr3 = dqr3 = 0.0;

	       /* Calculate new (p,q) values */
	       DataP[y][x] = (DataP[y - 1][x]
			      + DataP[y + 1][x]
			      + DataP[y][x - 1]
			      + DataP[y][x + 1]) / 4
		  + ((r1 - i1) * dpr1
		     + (r2 - i2) * dpr2
		     + (r3 - i3) * dpr3) * Alpha;

	       DataQ[y][x] = (DataQ[y - 1][x]
			      + DataQ[y + 1][x]
			      + DataQ[y][x - 1]
			      + DataQ[y][x + 1]) / 4
		  + ((r1 - i1) * dqr1
		     + (r2 - i2) * dqr2
		     + (r3 - i3) * dqr3) * Alpha;
	    }
	 }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates functional energy.                      */
/*---------------------------------------------------------------------------*/
void Energy()
{
   int x, y;
   float i1, i2, i3, r1, r2, r3, p, q, px, py, qx, qy, length;
   float partA, partB, partC, total;

   /* Solve for (p,q) based on initial z */
   partA = partB = partC = total = 0.0;
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 /* Save image intensity values */
	 i1 = Data1[y][x];
	 i2 = Data2[y][x];
	 i3 = Data3[y][x];

	 /* Calculate modeled intensity values */
	 p = DataP[y][x];
	 q = DataQ[y][x];
	 length = (float) sqrt((double) (p * p + q * q + 1));
	 r1 = (-Lx * p - Ly * q + Lz) / length;
	 r2 = (-Mx * p - My * q + Mz) / length;
	 r3 = (-Nx * p - Ny * q + Nz) / length;
	 if (r1 < 0.0)
	    r1 = 0.0;
	 if (r2 < 0.0)
	    r2 = 0.0;
	 if (r3 < 0.0)
	    r3 = 0.0;

	 /* Calculate (p,q) smoothness */
	 px = DataP[y][x + 1] - DataP[y][x];
	 py = DataP[y + 1][x] - DataP[y][x];
	 qx = DataQ[y][x + 1] - DataQ[y][x];
	 qy = DataQ[y + 1][x] - DataQ[y][x];

	 /* Calculate energy */
	 partA += (px * px + py * py + qx * qx + qy * qy);
	 partB += (py - qx) * (py - qx);
	 partC += Alpha * ((r1 - i1) * (r1 - i1) + (r2 - i2) * (r2 - i2) + (r3 - i3) * (r3 - i3));
      }

   total = partA + partB + partC;
   if (Debug == TRUE)
      printf("Smoothness %6.2f Continuity %6.2f Match %6.2f Total %6.2f\n",
	     partA, partB, partC, total);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine integrates to obtain z from (p,q).                 */
/*---------------------------------------------------------------------------*/
void Integrate()
{
   int x, y;

   /* Forwards X, forwards Y pass */
   for (y = 1; y < Ydim; y++)
      for (x = 1; x < Xdim; x++)
	 DataZ[y][x] = (DataZ[y][x - 1] + DataZ[y - 1][x]
			+ DataP[y][x - 1] + DataQ[y - 1][x]) / 2;

   /* Forwards X, backwards Y pass */
   for (y = Ydim / 2; y >= 0; y--)
      for (x = Xdim / 2; x < Xdim; x++)
	 DataZ[y][x] = (DataZ[y][x - 1] + DataZ[y + 1][x]
			+ DataP[y][x - 1] - DataQ[y + 1][x]) / 2;

   /* Backwards X, backwards Y pass */
   for (y = Ydim / 2; y >= 0; y--)
      for (x = Xdim / 2; x >= 0; x--)
	 DataZ[y][x] = (DataZ[y][x + 1] + DataZ[y + 1][x]
			- DataP[y][x + 1] - DataQ[y + 1][x]) / 2;

   /* Backwards X, forwards Y pass */
   for (y = Ydim / 2; y < Ydim; y++)
      for (x = Xdim / 2; x >= 0; x--)
	 DataZ[y][x] = (DataZ[y][x + 1] + DataZ[y - 1][x]
			- DataP[y][x + 1] + DataQ[y - 1][x]) / 2;
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   char Name4[50];
   char Name5[50];
   char Name6[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   IM_TYPE *Image4;
   IM_TYPE *Image5;
   IM_TYPE *Image6;
   int PixType, Zdim, DimCnt;
   char Title[nTITLE];

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("PHOTOMETRIC Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'a':
	 if (sscanf(argv[++i], "%f", &Alpha) == 0)
	    Error("Can not get smoothness weight");
	 break;
      case 'c':
	 if (sscanf(argv[++i], "%d", &Count) == 0)
	    Error("Can not get iteration count");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 6)
   {
      fprintf(stderr,
	      "Usage: photometric [options] infile1 infile2 infile3 outP outQ outZ\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-a #]  Smoothness weight (0.1)\n");
      fprintf(stderr, "       [-c #]  Number of smoothing iterations (10)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name3) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name4) == 0)
      Error("Could not get output file name");
   if (sscanf(argv[i++], "%s", Name5) == 0)
      Error("Could not get output file name");
   if (sscanf(argv[i++], "%s", Name6) == 0)
      Error("Could not get output file name");

   /* Read input image1 */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));
   im_get_title(Image1, Title);
   sscanf(Title, "%f %f %f", &Lx, &Ly, &Lz);
   if (Debug == TRUE)
      printf("Lx Ly Lz %f %f %f\n", Lx, Ly, Lz);

   /* Read input image2 */
   Image2 = im_open(Name2, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   im_read(Image2, FLOAT, (char *) &(Data2[0][0]));
   im_get_title(Image2, Title);
   sscanf(Title, "%f %f %f", &Mx, &My, &Mz);
   if (Debug == TRUE)
      printf("Mx My Mz %f %f %f\n", Mx, My, Mz);

   /* Read input image3 */
   Image3 = im_open(Name3, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data3 = (FLOAT_TYPE **) im_alloc2D(Image3, FLOAT);
   im_read(Image3, FLOAT, (char *) &(Data3[0][0]));
   im_get_title(Image3, Title);
   sscanf(Title, "%f %f %f", &Nx, &Ny, &Nz);
   if (Debug == TRUE)
      printf("Nx Ny Nz %f %f %f\n", Nx, Ny, Nz);

   /* Create output P,Q,Z images */
   Image4 = im_create(Name4, FLOAT, Xdim, Ydim, Zdim);
   Image5 = im_create(Name5, FLOAT, Xdim, Ydim, Zdim);
   Image6 = im_create(Name6, FLOAT, Xdim, Ydim, Zdim);
   DataP = (FLOAT_TYPE **) im_alloc2D(Image4, FLOAT);
   DataQ = (FLOAT_TYPE **) im_alloc2D(Image5, FLOAT);
   DataZ = (FLOAT_TYPE **) im_alloc2D(Image6, FLOAT);

   /* Solve for (p,q) and z */
   Solve();
   Relax();
   Integrate();
   Energy();

   /* Write information to output image */
   im_write(Image4, FLOAT, (char *) &(DataP[0][0]));
   im_write(Image5, FLOAT, (char *) &(DataQ[0][0]));
   im_write(Image6, FLOAT, (char *) &(DataZ[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   im_free2D((char **) Data3);
   im_free2D((char **) DataP);
   im_free2D((char **) DataQ);
   im_free2D((char **) DataZ);
   return (0);
}
