/*---------------------------------------------------------------------------*/
/* Program:  plot_line.c                                                     */
/*                                                                           */
/* Purpose:  This program plots a sequence of (x1,y1) (x2,y2) line           */
/*           segments in color on an image.  The line colors can be          */
/*           chosen globally or on a line by line basis.                     */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     June 7, 1996                                                    */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1996                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>

/* Global variables */
COLOR_TYPE **Data1;
COLOR_TYPE Line;
int Debug = FALSE;

/* Function prototype */
void PlotLine(int x1, int y1, int x2, int y2);

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name0[50];
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int i = 0, x1, y1, x2, y2;
   int temp = 0;
   FILE *fd;

   /* Initialize line color */
   Line.r = Line.g = Line.b = 200;

   /* Interpret program options */
   printf("PLOT_LINE Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'c':
	 if (sscanf(argv[++i], "%d", &temp) == 0)
	    Error("Could not get red component");
	 Line.r = temp;
	 if (sscanf(argv[++i], "%d", &temp) == 0)
	    Error("Could not get green component");
	 Line.g = temp;
	 if (sscanf(argv[++i], "%d", &temp) == 0)
	    Error("Could not get blue component");
	 Line.b = temp;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr, "Usage: plot_line [options] linefile infile outfile\n");
      fprintf(stderr, "       [-c R# G# B#]  Color of lines to be drawn\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name0) == 0)
      Error("Could not get line file name");
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can only process 2D images");
   Data1 = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, COLOR, Xdim, Ydim, Zdim);
   printf("Line color = %d %d %d\n", Line.r, Line.g, Line.b);

   /* Read the line file and draw lines */
   if ((fd = fopen(Name0, "r")) == NULL)
      Error("Could not open point file");
   while (fscanf(fd, "%d %d %d %d", &x1, &y1, &x2, &y2) != EOF)
      PlotLine(x1, y1, x2, y2);
   fclose(fd);

   /* Write information to output image */
   im_write(Image2, COLOR, (char *) &(Data1[0][0].r));
   im_free2D((char **) Data1);
   return (0);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine plots a given line segment.                        */
/*---------------------------------------------------------------------------*/
void PlotLine(int x1, int y1, int x2, int y2)
{
   int dx, dy, length, i, x, y;

   if (Debug == TRUE)
      printf("line %d %d to %d %d\n", x1, y1, x2, y2);
   dx = x2 - x1;
   dy = y2 - y1;
   length = (int) sqrt((double) (dx * dx + dy * dy));
   for (i = 0; i <= length; i++)
   {
      x = x1 + dx * i / length;
      y = y1 + dy * i / length;
      Data1[y][x].r = Line.r;
      Data1[y][x].g = Line.g;
      Data1[y][x].b = Line.b;
      if (Debug == TRUE)
	 printf("%d %d\n", x, y);
   }
}
