/*---------------------------------------------------------------------------*/
/* Program:  quantize.c                                                      */
/*                                                                           */
/* Purpose:  This program quantizes an input image to the specified          */
/*           number of grey levels using a variation on the Lloyd-Max        */
/*           least quared quantization algorithm.                            */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     November 14, 1995                                               */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1995                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
double sqrt(double);

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   SHORT_TYPE **Data1;
   SHORT_TYPE **Data2;
   INT_TYPE *Histo;
   INT_TYPE *Lookup;
   INT_TYPE *Quant;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int Diffuse = FALSE;
   int Number = 20;
   int Iterate = 20;
   int i = 0, j, x, y;
   int Min, Max, Range, Low, High, Err;
   int Loop, Mean, Count;

   /* Interpret program options */
   printf("QUANTIZE Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'n':
	 if (sscanf(argv[++i], "%d", &Number) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'i':
	 if (sscanf(argv[++i], "%d", &Iterate) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      case 'e':
	 Diffuse = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: quantize [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-e]    Perform error diffusion\n");
      fprintf(stderr, "       [-n #]  Number of quantization levels (20)\n");
      fprintf(stderr, "       [-i #]  Number of minimization iterations (20)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   im_read(Image1, SHORT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, PixType, Xdim, Ydim, Zdim);
   Data2 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);

   /* Find intensity range */
/*----------------------*/
   Min = Max = Data1[0][0];
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 if (Min > Data1[y][x])
	    Min = Data1[y][x];
	 else if (Max < Data1[y][x])
	    Max = Data1[y][x];
   Range = Max - Min + 1;

   /* Calculate histogram */
/*---------------------*/
   Image1 = im_create("/dev/null", INT, Range, 1, 1);
   Histo = (INT_TYPE *) im_alloc1D(Image1, INT);
   Lookup = (INT_TYPE *) im_alloc1D(Image1, INT);
   for (i = 0; i < Range; i++)
      Histo[i] = 0;
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 Histo[Data1[y][x] - Min]++;

   /* Select initial quantization levels */
/*------------------------------------*/
   Image1 = im_create("/dev/null", INT, Number, 1, 1);
   Quant = (INT_TYPE *) im_alloc1D(Image1, INT);
   for (j = 0; j < Number; j++)
      Quant[j] = (int) (0.5 + Min + (Max - Min) * j / (float) (Number - 1));

   /* Iteratively optimize quantization levels */
/*------------------------------------------*/
   for (Loop = 0; Loop < Iterate; Loop++)
   {
      Err = 0;
      Low = Min;
      for (j = 0; j < Number - 1; j++)
      {
	 Mean = 0;
	 Count = 0;
	 High = (Quant[j] + Quant[j + 1]) / 2;
	 for (i = Low; i <= High; i++)
	 {
	    Mean += i * Histo[i - Min];
	    Count += Histo[i - Min];
	    Err += (i - Quant[j]) * (i - Quant[j]) * Histo[i - Min];
	 }
	 if (Count > 0)
	    Quant[j] = (int) (0.5 + Mean / (float) Count);
	 else if (j > 0)
	    Quant[j] = (Quant[j - 1] + Quant[j + 1]) / 2;
	 Low = High + 1;
      }
      High = Max;
      Mean = 0;
      Count = 0;
      for (i = Low; i <= High; i++)
      {
	 Mean += i * Histo[i - Min];
	 Count += Histo[i - Min];
	 Err += (i - Quant[j]) * (i - Quant[j]) * Histo[i - Min];
      }
      if (Count > 0)
	 Quant[j] = (int) (0.5 + Mean / (float) Count);
      else
	 Quant[j] = (Quant[j - 1] + Quant[j]) / 2;
      if (Debug == TRUE)
	 printf("Iteration %d quantization error = %lf\n", Loop,
		sqrt((double) Err / (Xdim * Ydim)));
   }

   /* Generate quantization lookup table */
/*------------------------------------*/
   Low = Min;
   for (j = 0; j < Number - 1; j++)
   {
      High = (Quant[j] + Quant[j + 1]) / 2;
      for (i = Low; i <= High; i++)
	 Lookup[i - Min] = Quant[j];
      Low = High + 1;
   }
   High = Max;
   for (i = Low; i <= High; i++)
      Lookup[i - Min] = Quant[j];

   /* Quantize input image data */
/*---------------------------*/
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Data2[y][x] = Lookup[Data1[y][x] - Min];
	 if ((x < (Xdim - 1)) && (Diffuse == TRUE))
	 {
	    /* Distribute error to next pixel */
	    Err = Data1[y][x] - Data2[y][x];
	    Data1[y][x + 1] += Err;
	    if (Data1[y][x + 1] < Min)
	       Data1[y][x + 1] = Min;
	    if (Data1[y][x + 1] > Max)
	       Data1[y][x + 1] = Max;
	 }
      }

   /* Write information to output image */
   im_write(Image2, SHORT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   return (0);
}
