/*---------------------------------------------------------------------------*/
/* Program:  ridge.c                                                         */
/*                                                                           */
/* Purpose:  This program locates ridges in an image by searching for        */
/*           extrema of level curve curvature.  With the -c option, the      */
/*           level curve curvature is calculated.  With the -v option,       */
/*           the second fundamental form is computed.  The zeros of this     */
/*           function correspond to ridge tops and valley bottoms.           */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     September 22, 1994                                              */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define CURVATURE 1
#define VERTEX    2

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   FLOAT_TYPE **Data1;
   FLOAT_TYPE **Data2;
   FLOAT_TYPE **Data3;
   int PixType, Xdim, Ydim, Zdim, DimCnt;

   /* Program variables */
   int Debug = FALSE;
   int Output = CURVATURE;
   int i = 0, x, y;
   float Dx, Dy, Dxx, Dxy, Dyy, Cx, Cy;
   float Len, Len3;

   /* Interpret program options */
   printf("RIDGE Program - KUIM Version 1.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      case 'c':
	 Output = CURVATURE;
	 break;
      case 'v':
	 Output = VERTEX;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: ridge [options] infile outfile\n");
      fprintf(stderr, "       [-d]  Print debugging information\n");
      fprintf(stderr, "       [-c]  Output level curve curvature.\n");
      fprintf(stderr, "       [-v]  Output second fundamental form.\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 2)
      Error("Can not process 1D or 3D images");
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, FLOAT, Xdim, Ydim, Zdim);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
   Data3 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Calculate level curve curvature */
   if (Output >= CURVATURE)
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    /* Estimate derivatives */
	    Dx = (Data1[y + 1][x + 1] + 2 * Data1[y][x + 1] + Data1[y - 1][x + 1]
		  - Data1[y + 1][x - 1] - 2 * Data1[y][x - 1] - Data1[y - 1][x - 1]) / 8;

	    Dy = (Data1[y + 1][x + 1] + 2 * Data1[y + 1][x] + Data1[y + 1][x - 1]
		  - Data1[y - 1][x + 1] - 2 * Data1[y - 1][x] - Data1[y - 1][x - 1]) / 8;

	    Dxx = (Data1[y + 1][x + 1] + 2 * Data1[y][x + 1] + Data1[y - 1][x + 1]
	       - 2 * Data1[y + 1][x] - 4 * Data1[y][x] - 2 * Data1[y - 1][x]
		   + Data1[y + 1][x - 1] + 2 * Data1[y][x - 1] + Data1[y - 1][x - 1]) / 4;

	    Dyy = (Data1[y + 1][x + 1] - 2 * Data1[y][x + 1] + Data1[y - 1][x + 1]
	       + 2 * Data1[y + 1][x] - 4 * Data1[y][x] + 2 * Data1[y - 1][x]
		   + Data1[y + 1][x - 1] - 2 * Data1[y][x - 1] + Data1[y - 1][x - 1]) / 4;

	    Dxy = (Data1[y + 1][x + 1] - Data1[y - 1][x + 1]
		   - Data1[y + 1][x - 1] + Data1[y - 1][x - 1]) / 4;

	    /* Calculate level curve curvature */
	    Len = (float) sqrt((double) (Dx * Dx + Dy * Dy));
	    Len3 = Len * Len * Len;
	    if (Len3 > 0.0)
	       Data2[y][x] = -(Dy * Dy * Dxx - 2 * Dx * Dy * Dxy + Dx * Dx * Dyy) / Len3;
	    else
	       Data2[y][x] = -(Dy * Dy * Dxx - 2 * Dx * Dy * Dxy + Dx * Dx * Dyy);
	 }

   /* Handle boundary rows and columns */
   for (x = 0; x < Xdim; x++)
   {
      Data2[0][x] = Data2[1][x];
      Data2[Ydim - 1][x] = Data2[Ydim - 2][x];
   }
   for (y = 0; y < Ydim; y++)
   {
      Data2[y][0] = Data2[y][1];
      Data2[y][Xdim - 1] = Data2[y][Xdim - 2];
   }

   /* Calculate second fundamental form */
   if (Output >= VERTEX)
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    /* Estimate derivatives */
	    Dx = (Data1[y + 1][x + 1] + 2 * Data1[y][x + 1] + Data1[y - 1][x + 1]
		  - Data1[y + 1][x - 1] - 2 * Data1[y][x - 1] - Data1[y - 1][x - 1]) / 8;

	    Dy = (Data1[y + 1][x + 1] + 2 * Data1[y + 1][x] + Data1[y + 1][x - 1]
		  - Data1[y - 1][x + 1] - 2 * Data1[y - 1][x] - Data1[y - 1][x - 1]) / 8;

	    Cx = (Data2[y + 1][x + 1] + 2 * Data2[y][x + 1] + Data2[y - 1][x + 1]
		  - Data2[y + 1][x - 1] - 2 * Data2[y][x - 1] - Data2[y - 1][x - 1]) / 8;

	    Cy = (Data2[y + 1][x + 1] + 2 * Data2[y + 1][x] + Data2[y + 1][x - 1]
		  - Data2[y - 1][x + 1] - 2 * Data2[y - 1][x] - Data2[y - 1][x - 1]) / 8;

	    /* Calculate second fundamental form */
	    Len = Dx * Dx + Dy * Dy;
	    Len3 = Len * Len * Len;
	    if (Len3 > 0.0)
	       Data3[y][x] = (Cx * Dy - Cy * Dx) / Len3;
	    else
	       Data3[y][x] = Cx * Dy - Cy * Dx;
	 }

   /* Handle boundary rows and columns */
   for (x = 0; x < Xdim; x++)
   {
      Data3[0][x] = Data3[1][x];
      Data3[Ydim - 1][x] = Data3[Ydim - 2][x];
   }
   for (y = 0; y < Ydim; y++)
   {
      Data3[y][0] = Data3[y][1];
      Data3[y][Xdim - 1] = Data3[y][Xdim - 2];
   }

   /* Write information to output image */
   if (Output == CURVATURE)
      im_write(Image2, FLOAT, (char *) &(Data2[0][0]));
   else if (Output == VERTEX)
      im_write(Image2, FLOAT, (char *) &(Data3[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   im_free2D((char **) Data3);
   return (0);
}
