/*---------------------------------------------------------------------------*/
/* Program:  split_merge.c                                                   */
/*                                                                           */
/* Purpose:  This program performs the classic recursive split merge         */
/*           segmentation algorithm.  Regions are split/merged until         */
/*           they have a variance below a specified threshold.               */
/*                                                                           */
/* Author:   John Gauch, Bona Nasution                                       */
/*                                                                           */
/* Date:     April 25, 1994 - Original program.                              */
/*           June 26, 1997  - extended to handle color.                      */
/*           July 30, 1997  - Simplified color, vector, scalar code.         */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define MAXREGION 500000

/* Image variables */
char Name1[50];
char Name2[50];
IM_TYPE *Image1;
IM_TYPE *Image2;
int PixType, Xdim, Ydim, Zdim, DimCnt;

/* Global Variables */
float VarThresh = 225;
int Debug = FALSE;
int Region = 0;
int Parent[MAXREGION];
int Count[MAXREGION];
float Total[3][MAXREGION];
float Tot2[3][MAXREGION];

/* Function prototypes */
void Scalar();
void SplitScalar(int Xlow, int Xhigh, int Ylow, int Yhigh,
		 FLOAT_TYPE ** Data1, INT_TYPE ** Data2);
void MergeScalar(int Region1, int Region2);
void Color();
void SplitColor(int Xlow, int Xhigh, int Ylow, int Yhigh,
		COLOR_TYPE ** Data1, INT_TYPE ** Data2);
void MergeColor(int Region1, int Region2);
void Vector();
void SplitVector(int Xlow, int Xhigh, int Ylow, int Yhigh,
		 FLOAT_TYPE *** Data1, INT_TYPE ** Data2);
void MergeVector(int Region1, int Region2);

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("SPLIT_MERGE Program - KUIM Version 2.1\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'v':
	 if (sscanf(argv[++i], "%f", &VarThresh) == 0)
	    Error("Could not get variance threshold");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: split_merge [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-v #]  Desired region variance (225)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);

   /* Handle color images */
   if (PixType == COLOR || PixType == PSEUDO || PixType == JPEG_RGB)
      Color();

   /* Handle grey images */
   else if (DimCnt == 2)
      Scalar();

   /* Handle vector images */
   else if ((DimCnt == 3) && (Zdim == 3))
      Vector();

   /* Unsupported image type */
   else
      Error("Can not process image");
   return (0);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates the split_merge of a scalar image.      */
/*---------------------------------------------------------------------------*/
void Scalar()
{
   FLOAT_TYPE **Data1;
   INT_TYPE **Data2;
   int x, y;

   /* Read input image */
   Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, INT, Xdim, Ydim, Zdim);
   Data2 = (INT_TYPE **) im_alloc2D(Image2, INT);

   /* Perform segmentation */
   SplitScalar(0, Xdim, 0, Ydim, Data1, Data2);
   if (Debug == TRUE)
      printf("Split count = %d\n", Region);

   /* Mark each region with mean value of region */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Region = Data2[y][x];
	 while (Parent[Region] > 0)
	    Region = Parent[Region];
	 Data2[y][x] = (int) (Total[0][Region] / Count[Region]);
      }

   /* Write information to output image */
   im_write(Image2, INT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does the recursive splitting.                      */
/*---------------------------------------------------------------------------*/
void SplitScalar(int Xlow, int Xhigh, int Ylow, int Yhigh,
		 FLOAT_TYPE ** Data1, INT_TYPE ** Data2)
{
   int x, y, Xmid, Ymid;
   int Cnt = 0;
   float Value, total = 0, total2 = 0;
   float Mean = 0, Variance = 0;

   /* Calculate region mean and variance */
   if ((Xlow == Xhigh) || (Ylow == Yhigh))
      return;
   for (y = Ylow; y < Yhigh; y++)
      for (x = Xlow; x < Xhigh; x++)
      {
	 Value = Data1[y][x];
	 total += Value;
	 total2 += Value * Value;
      }
   Cnt = (Yhigh - Ylow) * (Xhigh - Xlow);
   Mean = total / Cnt;
   Variance = total2 / Cnt - Mean * Mean;

   /* Save region information */
   if (Variance <= VarThresh)
   {
      for (y = Ylow; y < Yhigh; y++)
	 for (x = Xlow; x < Xhigh; x++)
	    Data2[y][x] = (int) Region;

      if (Region >= MAXREGION)
	 Error("Too many regions");
      Parent[Region] = -1;
      Count[Region] = Cnt;
      Total[0][Region] = total;
      Tot2[0][Region] = total2;
      Region++;
   }

   /* Recursively split and merge regions */
   else
   {
      Xmid = (Xhigh + Xlow) / 2;
      Ymid = (Yhigh + Ylow) / 2;
      SplitScalar(Xlow, Xmid, Ylow, Ymid, Data1, Data2);
      SplitScalar(Xlow, Xmid, Ymid, Yhigh, Data1, Data2);
      SplitScalar(Xmid, Xhigh, Ylow, Ymid, Data1, Data2);
      SplitScalar(Xmid, Xhigh, Ymid, Yhigh, Data1, Data2);

      if (Xmid > 0)
	 for (y = Ylow; y < Yhigh; y++)
	    MergeScalar(Data2[y][Xmid], Data2[y][Xmid - 1]);
      if (Ymid > 0)
	 for (x = Xlow; x < Xhigh; x++)
	    MergeScalar(Data2[Ymid][x], Data2[Ymid - 1][x]);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine merges two regions.                                */
/*---------------------------------------------------------------------------*/
void MergeScalar(int Region1, int Region2)
{
   float NewMean, NewVar;

   /* Find parents of regions */
   if ((Region1 == -1) || (Region2 == -1))
      return;
   while (Parent[Region1] > 0)
      Region1 = Parent[Region1];
   while (Parent[Region2] > 0)
      Region2 = Parent[Region2];
   if (Region1 == Region2)
      return;

   /* Calculate combined variance */
   NewMean = (Total[0][Region1] + Total[0][Region2]) /
      (Count[Region1] + Count[Region2]);
   NewVar = (Tot2[0][Region1] + Tot2[0][Region2]) /
      (Count[Region1] + Count[Region2]) - NewMean * NewMean;

   /* Combine region statistics */
   if (NewVar < VarThresh)
   {
      Count[Region1] += Count[Region2];
      Total[0][Region1] += Total[0][Region2];
      Tot2[0][Region1] += Tot2[0][Region2];
      Parent[Region2] = Region1;
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates the split_merge of a color image.       */
/*---------------------------------------------------------------------------*/
void Color()
{
   COLOR_TYPE **Data1;
   INT_TYPE **Data2;
   int x, y;

   /* Read input image */
   Data1 = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
   im_read(Image1, COLOR, (char *) &(Data1[0][0]));

   /* Create output image */
   Image2 = im_create(Name2, COLOR, Xdim, Ydim, Zdim);
   Data2 = (INT_TYPE **) im_alloc2D(Image2, INT);

   /* Perform segmentation */
   SplitColor(0, Xdim, 0, Ydim, Data1, Data2);
   if (Debug == TRUE)
      printf("Split count = %d\n", Region);

   /* Mark each region with mean value of region */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Region = Data2[y][x];
	 while (Parent[Region] > 0)
	    Region = Parent[Region];
	 Data1[y][x].r = (BYTE_TYPE) (Total[0][Region] / Count[Region]);
	 Data1[y][x].g = (BYTE_TYPE) (Total[1][Region] / Count[Region]);
	 Data1[y][x].b = (BYTE_TYPE) (Total[2][Region] / Count[Region]);
      }

   /* Write information to output image */
   im_write(Image2, COLOR, (char *) &(Data1[0][0]));
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does the recursive splitting for color images.     */
/*---------------------------------------------------------------------------*/
void SplitColor(int Xlow, int Xhigh, int Ylow, int Yhigh,
		COLOR_TYPE ** Data1, INT_TYPE ** Data2)
{
   int x, y, Xmid, Ymid;
   int Cnt = 0;
   float RValue, GValue, BValue;
   float RTotal = 0, GTotal = 0, BTotal = 0;
   float RTotal2 = 0, GTotal2 = 0, BTotal2 = 0;
   float RMean = 0, GMean = 0, BMean = 0, Variance = 0;
   float RVar = 0, GVar = 0, BVar = 0;

   /* Calculate region mean and variance */
   if ((Xlow == Xhigh) || (Ylow == Yhigh))
      return;
   for (y = Ylow; y < Yhigh; y++)
      for (x = Xlow; x < Xhigh; x++)
      {
	 RValue = Data1[y][x].r;
	 GValue = Data1[y][x].g;
	 BValue = Data1[y][x].b;
	 RTotal += RValue;
	 GTotal += GValue;
	 BTotal += BValue;
	 RTotal2 += RValue * RValue;
	 GTotal2 += GValue * GValue;
	 BTotal2 += BValue * BValue;
	 Cnt++;
      }
   RMean = RTotal / Cnt;
   GMean = GTotal / Cnt;
   BMean = BTotal / Cnt;
   RVar = RTotal2 / Cnt - RMean * RMean;
   GVar = GTotal2 / Cnt - GMean * GMean;
   BVar = BTotal2 / Cnt - BMean * BMean;
   Variance = RVar + GVar + BVar;

   /* Save region information */
   if (Variance <= VarThresh)
   {
      for (y = Ylow; y < Yhigh; y++)
	 for (x = Xlow; x < Xhigh; x++)
	    Data2[y][x] = (int) Region;

      if (Region >= MAXREGION)
	 Error("Too many regions");
      Parent[Region] = -1;
      Count[Region] = Cnt;
      Total[0][Region] = RTotal;
      Total[1][Region] = GTotal;
      Total[2][Region] = BTotal;
      Tot2[0][Region] = RTotal2;
      Tot2[1][Region] = GTotal2;
      Tot2[2][Region] = BTotal2;
      Region++;
   }

   /* Recursively split and merge regions */
   else
   {
      Xmid = (Xhigh + Xlow) / 2;
      Ymid = (Yhigh + Ylow) / 2;
      SplitColor(Xlow, Xmid, Ylow, Ymid, Data1, Data2);
      SplitColor(Xlow, Xmid, Ymid, Yhigh, Data1, Data2);
      SplitColor(Xmid, Xhigh, Ylow, Ymid, Data1, Data2);
      SplitColor(Xmid, Xhigh, Ymid, Yhigh, Data1, Data2);

      if (Xmid > 0)
	 for (y = Ylow; y < Yhigh; y++)
	    MergeColor(Data2[y][Xmid], Data2[y][Xmid - 1]);
      if (Ymid > 0)
	 for (x = Xlow; x < Xhigh; x++)
	    MergeColor(Data2[Ymid][x], Data2[Ymid - 1][x]);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine merges two regions of a Color Image.               */
/*---------------------------------------------------------------------------*/
void MergeColor(int Region1, int Region2)
{
   float NewMeanR, NewMeanG, NewMeanB;
   float NewVarR, NewVarG, NewVarB, NewVariance;

   /* Find parents of regions */
   if ((Region1 == -1) || (Region2 == -1))
      return;
   while (Parent[Region1] > 0)
      Region1 = Parent[Region1];
   while (Parent[Region2] > 0)
      Region2 = Parent[Region2];
   if (Region1 == Region2)
      return;

   /* Calculate combined variance */
   NewMeanR = (Total[0][Region1] + Total[0][Region2]) /
      (Count[Region1] + Count[Region2]);
   NewMeanG = (Total[1][Region1] + Total[1][Region2]) /
      (Count[Region1] + Count[Region2]);
   NewMeanB = (Total[2][Region1] + Total[2][Region2]) /
      (Count[Region1] + Count[Region2]);
   NewVarR = (Tot2[0][Region1] + Tot2[0][Region2]) /
      (Count[Region1] + Count[Region2]) - NewMeanR * NewMeanR;
   NewVarG = (Tot2[1][Region1] + Tot2[1][Region2]) /
      (Count[Region1] + Count[Region2]) - NewMeanG * NewMeanG;
   NewVarB = (Tot2[2][Region1] + Tot2[2][Region2]) /
      (Count[Region1] + Count[Region2]) - NewMeanB * NewMeanB;
   NewVariance = NewVarR + NewVarG + NewVarB;

   /* Combine region statistics */
   if (NewVariance < VarThresh)
   {
      Count[Region1] += Count[Region2];
      Total[0][Region1] += Total[0][Region2];
      Total[1][Region1] += Total[1][Region2];
      Total[2][Region1] += Total[2][Region2];
      Tot2[0][Region1] += Tot2[0][Region2];
      Tot2[1][Region1] += Tot2[1][Region2];
      Tot2[2][Region1] += Tot2[2][Region2];
      Parent[Region2] = Region1;
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates the split_merge of a vector image.      */
/*---------------------------------------------------------------------------*/
void Vector()
{
   FLOAT_TYPE ***Data1;
   INT_TYPE **Data2;
   int x, y;

   /* Read input image */
   Data1 = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data1[0][0][0]));

   /* Create output image */
   Image2 = im_create(Name2, FLOAT, Xdim, Ydim, Zdim);
   Data2 = (INT_TYPE **) im_alloc2D(Image2, INT);

   /* Perform segmentation */
   SplitVector(0, Xdim, 0, Ydim, Data1, Data2);
   if (Debug == TRUE)
      printf("Split count = %d\n", Region);

   /* Mark each region with mean value of region */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Region = Data2[y][x];
	 while (Parent[Region] > 0)
	    Region = Parent[Region];
	 Data1[0][y][x] = Total[0][Region] / Count[Region];
	 Data1[1][y][x] = Total[1][Region] / Count[Region];
	 Data1[2][y][x] = Total[2][Region] / Count[Region];
      }

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data1[0][0][0]));
   im_free3D((char ***) Data1);
   im_free2D((char **) Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does the recursive splitting for vector images.    */
/*---------------------------------------------------------------------------*/
void SplitVector(int Xlow, int Xhigh, int Ylow, int Yhigh,
		 FLOAT_TYPE *** Data1, INT_TYPE ** Data2)
{
   int x, y, z, Xmid, Ymid;
   int Cnt = 0;
   float Value[3], total[3], total2[3];
   float Mean[3], Var[3], Variance;

   /* Initialize totals */
   for (z = 0; z < 3; z++)
      total[z] = total2[z] = 0;

   /* Calculate region mean and variance */
   if ((Xlow == Xhigh) || (Ylow == Yhigh))
      return;
   for (y = Ylow; y < Yhigh; y++)
      for (x = Xlow; x < Xhigh; x++)
	 for (z = 0; z < 3; z++)
	 {
	    Value[z] = Data1[z][y][x];
	    total[z] += Value[z];
	    total2[z] += Value[z] * Value[z];
	 }
   Cnt = (Yhigh - Ylow) * (Xhigh - Xlow);
   for (z = 0; z < 3; z++)
   {
      Mean[z] = total[z] / Cnt;
      Var[z] = total2[z] / Cnt - Mean[z] * Mean[z];
   }
   Variance = Var[0] + Var[1] + Var[2];

   /* Save region information */
   if (Variance <= VarThresh)
   {
      for (y = Ylow; y < Yhigh; y++)
	 for (x = Xlow; x < Xhigh; x++)
	    Data2[y][x] = (int) Region;

      if (Region >= MAXREGION)
	 Error("Too many regions");
      Parent[Region] = -1;
      Count[Region] = Cnt;
      for (z = 0; z < 3; z++)
      {
	 Total[z][Region] = total[z];
	 Tot2[z][Region] = total2[z];
      }
      Region++;
   }

   /* Recursively split and merge regions */
   else
   {
      Xmid = (Xhigh + Xlow) / 2;
      Ymid = (Yhigh + Ylow) / 2;
      SplitVector(Xlow, Xmid, Ylow, Ymid, Data1, Data2);
      SplitVector(Xlow, Xmid, Ymid, Yhigh, Data1, Data2);
      SplitVector(Xmid, Xhigh, Ylow, Ymid, Data1, Data2);
      SplitVector(Xmid, Xhigh, Ymid, Yhigh, Data1, Data2);

      if (Xmid > 0)
	 for (y = Ylow; y < Yhigh; y++)
	    MergeVector(Data2[y][Xmid], Data2[y][Xmid - 1]);
      if (Ymid > 0)
	 for (x = Xlow; x < Xhigh; x++)
	    MergeVector(Data2[Ymid][x], Data2[Ymid - 1][x]);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine merges two regions of a vector Image.              */
/*---------------------------------------------------------------------------*/
void MergeVector(int Region1, int Region2)
{
   float NewMean[3], NewVar[3], NewVariance;

   /* Find parents of regions */
   if ((Region1 == -1) || (Region2 == -1))
      return;
   while (Parent[Region1] > 0)
      Region1 = Parent[Region1];
   while (Parent[Region2] > 0)
      Region2 = Parent[Region2];
   if (Region1 == Region2)
      return;

   /* Calculate combined variance */
   NewMean[0] = (Total[0][Region1] + Total[0][Region2]) /
      (Count[Region1] + Count[Region2]);
   NewMean[1] = (Total[1][Region1] + Total[1][Region2]) /
      (Count[Region1] + Count[Region2]);
   NewMean[2] = (Total[2][Region1] + Total[2][Region2]) /
      (Count[Region1] + Count[Region2]);
   NewVar[0] = (Tot2[0][Region1] + Tot2[0][Region2]) /
      (Count[Region1] + Count[Region2]) - NewMean[0] * NewMean[0];
   NewVar[1] = (Tot2[1][Region1] + Tot2[1][Region2]) /
      (Count[Region1] + Count[Region2]) - NewMean[1] * NewMean[1];
   NewVar[2] = (Tot2[2][Region1] + Tot2[2][Region2]) /
      (Count[Region1] + Count[Region2]) - NewMean[2] * NewMean[2];
   NewVariance = NewVar[0] + NewVar[1] + NewVar[2];
   NewVariance = NewVar[0] + NewVar[1] + NewVar[2];

   /* Combine region statistics */
   if (NewVariance < VarThresh)
   {
      Count[Region1] += Count[Region2];
      Total[0][Region1] += Total[0][Region2];
      Total[1][Region1] += Total[1][Region2];
      Total[2][Region1] += Total[2][Region2];
      Tot2[0][Region1] += Tot2[0][Region2];
      Tot2[1][Region1] += Tot2[1][Region2];
      Tot2[2][Region1] += Tot2[2][Region2];
      Parent[Region2] = Region1;
   }
}
