/*---------------------------------------------------------------------------*/
/* Program:  stats.c                                                         */
/*                                                                           */
/* Purpose:  This program calculates image statistics in a small 2D region   */
/*           about each point or along the Z dimension of 3D an image.       */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     June 2, 1994                                                    */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define AVERAGE  1
#define STD_DEV  2
#define VARIANCE 3
#define MAXVAL   4
#define MINVAL   5
#define RANGE    6
#define MEDIAN   7

/* Global variables */
FLOAT_TYPE **Data1;
FLOAT_TYPE **Data2;
FLOAT_TYPE ***Data3;
int Xdim, Ydim, Zdim, DimCnt;
int Output = AVERAGE;
int dx = 2;
int dy = 2;
double sqrt();

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine does simple selection sorting.                     */
/*---------------------------------------------------------------------------*/
void Sort(float Array[], int Count)
{
   int i, j, half, minpos;
   float minval;

   half = Count / 2;
   for (i = 0; i <= half; i++)
   {
      minpos = i;
      minval = Array[minpos];
      for (j = i + 1; j < Count; j++)
	 if (Array[j] < minval)
	 {
	    minpos = j;
	    minval = Array[j];
	 }
      Array[minpos] = Array[i];
      Array[i] = minval;
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates simple statistics for small regions     */
/*           of the image and outputs the result.                            */
/*---------------------------------------------------------------------------*/
void GetStats()
{
   int x, y, z, i, j, count;
   float temp, min, max, total, total2;
   float array[1000];

   /* Loop over all output pixels */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 total = 0.0;
	 total2 = 0.0;

	 /* Handle 2D images by processing window */
	 if (DimCnt == 2)
	 {
	    count = 0;
	    max = min = Data1[0][0];
	    for (j = y - dy; j <= y + dy; j++)
	       for (i = x - dx; i <= x + dx; i++)
		  if ((i > 0) && (i < Xdim) && (j > 0) && (j < Ydim))
		  {
		     count++;
		     temp = Data1[j][i];
		     total = total + temp;
		     total2 = total2 + temp * temp;
		     if (temp < min)
			min = temp;
		     if (temp > max)
			max = temp;
		  }
	 }

	 /* Handle 3D images by processing slices */
	 else if (DimCnt == 3)
	 {
	    count = Zdim;
	    max = min = Data3[0][y][x];
	    for (z = 0; z < Zdim; z++)
	    {
	       temp = Data3[z][y][x];
	       total = total + temp;
	       total2 = total2 + temp * temp;
	       if (temp < min)
		  min = temp;
	       if (temp > max)
		  max = temp;
	       array[z] = temp;
	    }
	 }

	 /* Calculate output value */
	 total = total / count;
	 if (Output == AVERAGE)
	    Data2[y][x] = total;
	 else if (Output == MINVAL)
	    Data2[y][x] = min;
	 else if (Output == MAXVAL)
	    Data2[y][x] = max;
	 else if (Output == RANGE)
	    Data2[y][x] = max - min;
	 else if (Output == VARIANCE)
	    Data2[y][x] = (total2 / count - total * total);
	 else if (Output == STD_DEV)
	    Data2[y][x] = (float) sqrt((double)
				       (total2 / count - total * total));
	 else if (Output == MEDIAN)
	 {
	    Sort(array, Zdim);
	    Data2[y][x] = array[Zdim / 2];
	 }
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   int PixType;

   /* Program variables */
   int Debug = FALSE;
   int i = 0;

   /* Interpret program options */
   printf("STATS Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      case 'a':
	 Output = AVERAGE;
	 break;
      case 'e':
	 Output = MEDIAN;
	 break;
      case 'm':
	 Output = MINVAL;
	 break;
      case 'M':
	 Output = MAXVAL;
	 break;
      case 'r':
	 Output = RANGE;
	 break;
      case 'v':
	 Output = VARIANCE;
	 break;
      case 's':
	 Output = STD_DEV;
	 break;
      case 'x':
	 if (sscanf(argv[++i], "%d", &dx) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'y':
	 if (sscanf(argv[++i], "%d", &dy) == 0)
	    Error("Could not get integer argument");
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: stats [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-a]    Calculate average\n");
      fprintf(stderr, "       [-e]    Calculate median\n");
      fprintf(stderr, "       [-m]    Calculate minimum\n");
      fprintf(stderr, "       [-M]    Calculate maximum\n");
      fprintf(stderr, "       [-r]    Calculate range (maximum-minimum)\n");
      fprintf(stderr, "       [-v]    Calculate variance\n");
      fprintf(stderr, "       [-s]    Calculate standard deviation\n");
      fprintf(stderr, "       [-x #]  Window half width (2)\n");
      fprintf(stderr, "       [-y #]  Window half height (2)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt == 1)
      Error("Can not process 1D images");
   else if (DimCnt == 2)
   {
      Data1 = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
      im_read(Image1, FLOAT, (char *) &(Data1[0][0]));
   }
   else if (DimCnt == 3)
   {
      Data3 = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
      im_read(Image1, FLOAT, (char *) &(Data3[0][0][0]));
   }

   /* Create output image */
   Image2 = im_create(Name2, FLOAT, Xdim, Ydim, 1);
   Data2 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Calculate statistics */
   GetStats();

   /* Write information to output image */
   im_write(Image2, FLOAT, (char *) &(Data2[0][0]));
   im_free2D((char **) Data2);
   if (DimCnt == 2)
      im_free2D((char **) Data1);
   if (DimCnt == 3)
      im_free3D((char ***) Data3);
   return (0);
}
