/*---------------------------------------------------------------------------*/
/* Program:  track.c                                                         */
/*                                                                           */
/* Purpose:  This program reads in a sequence of images and tracks the       */
/*           positions of blobs from one frame to the next.  The output      */
/*           image contains the paths of all blob centers.                   */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     June 15, 1996    - Rewrote old track program.                   */
/*           January 27, 1997 - Modified output format.                      */
/*           April 9, 1997    - Fixed bug in MI calculation.                 */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1996-1997               */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define SIZE 5000

/* Global variables */
int Debug = FALSE;		/* Flag for debugging output */
int Quarters = FALSE;		/* Flag for output by quarters */
int NewBlobs = FALSE;		/* Flag for tracking new blobs */
int Xdim, Ydim, Zdim;		/* Input image dimensions */
int NewXdim, NewYdim;		/* Output image dimensions */
int Thresh = 128;		/* Cell segmentation threshold */
int MinArea = 10;		/* Smallest blob size allowed (pixels) */
int MaxArea = 1000;		/* Largest blob size allowed (pixels) */
int Area, Xpos, Ypos;		/* Blob size and position */
int Color, MaxColor;		/* Blob color indices */
int Xfirst[SIZE], Xlast[SIZE];	/* X coordinates */
int Yfirst[SIZE], Ylast[SIZE];	/* Y coordinates */
int Tfirst[SIZE], Tlast[SIZE];	/* T coordinates */
int Status[SIZE];		/* Tracking status (1=tracked, 0=ignored) */
float Dist[SIZE];		/* Distance traveled */
float PixelScale = 0.658;	/* Microns per pixel */
float TimeScale = 6.4;		/* Seconds between images */
float MinVelocity = 5.0;	/* Slowest cell allowed (micons/minute) */
float MinDist = 10.0;		/* Smallest displacement allowed (microns) */
float MinTime = 5;		/* Minimim tracking time allowed (frames) */
char Name1[50];			/* Input file name */
char Name2[50];			/* Output file name */

/* Function prototypes */
void Preprocess(SHORT_TYPE ** In, SHORT_TYPE ** Out);
void BlobColor(SHORT_TYPE ** Data, int x, int y, int Color);
void DrawLine(SHORT_TYPE ** Out, int x1, int y1, int x2, int y2, int color);
void PrintStats(int frames, int minX, int maxX);

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine preprocesses the images.                           */
/*---------------------------------------------------------------------------*/
void Preprocess(SHORT_TYPE ** In, SHORT_TYPE ** Out)
{
   int x, y;

   /* Smooth the image */
   for (y = 1; y < Ydim - 1; y++)
      for (x = 1; x < Xdim - 1; x++)
	 Out[y][x] = (In[y - 1][x - 1] + In[y - 1][x] + In[y - 1][x + 1] +
		      In[y][x - 1] + In[y][x] + In[y][x + 1] +
		    In[y + 1][x - 1] + In[y + 1][x] + In[y + 1][x + 1]) / 9;

   /* Threshold the image */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 if (Out[y][x] < Thresh)
	    Out[y][x] = 2;
	 else
	    Out[y][x] = 0;

   /* Zero image boundary */
   for (y = 0; y < Ydim; y++)
   {
      Out[y][0] = 0;
      Out[y][Xdim - 1] = 0;
   }
   for (x = 0; x < Xdim; x++)
   {
      Out[0][x] = 0;
      Out[Ydim - 1][x] = 0;
   }

   /* Remove large and small blobs */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 if (Out[y][x] == 2)
	 {
	    Area = 0;
	    BlobColor(Out, x, y, 1);
	    if ((Area < MinArea) || (Area > MaxArea))
	       BlobColor(Out, x, y, 0);
	 }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine identifies a connected blob.                       */
/*---------------------------------------------------------------------------*/
void BlobColor(SHORT_TYPE ** Data, int x, int y, int Color)
{
   if ((Data[y][x] > 0) && (Data[y][x] != Color))
   {
      Area++;
      Xpos += x;
      Ypos += y;
      Data[y][x] = Color;
      BlobColor(Data, x + 1, y, Color);
      BlobColor(Data, x - 1, y, Color);
      BlobColor(Data, x, y + 1, Color);
      BlobColor(Data, x, y - 1, Color);
      BlobColor(Data, x + 1, y + 1, Color);
      BlobColor(Data, x - 1, y - 1, Color);
      BlobColor(Data, x - 1, y + 1, Color);
      BlobColor(Data, x + 1, y - 1, Color);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine draws a given line segment.                        */
/*---------------------------------------------------------------------------*/
void DrawLine(SHORT_TYPE ** Out, int x1, int y1, int x2, int y2, int color)
{
   int dx, dy, length, i, x, y;

   dx = x2 - x1;
   dy = y2 - y1;
   length = (int) sqrt((double) (dx * dx + dy * dy));
   for (i = 0; i < length; i++)
   {
      x = x1 + dx * i / length;
      y = y1 + dy * i / length;
      Out[y][x] = color;
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine prints the tracking statistics.                    */
/*---------------------------------------------------------------------------*/
void PrintStats(int frames, int minX, int maxX)
{
   int dx, dy, dt;
   float dist, velocity, MI;
   float AveVelocity = 0, AveIndex = 0;
   int TrackCnt = 0;

   if (Debug == TRUE)
   {
      printf("\n\nOBJECTS NOT CONSIDERED IN STATISTICS:\n\n");
      printf("Cell   Position   Frame1 Frame2   Displace   Velocity       M.I.\n");
      printf("----------------------------------------------------------------\n");
   }
   for (Color = 2; Color < MaxColor; Color++)
      if ((Tlast[Color] > Tfirst[Color]) &&
	  (Xfirst[Color] >= minX) && (Xfirst[Color] <= maxX))
      {
	 dx = Xlast[Color] - Xfirst[Color];
	 dy = Ylast[Color] - Yfirst[Color];
	 dt = Tlast[Color] - Tfirst[Color];
	 dist = (float) sqrt((double) (dx * dx + dy * dy)) * PixelScale;
	 if (dt == 0)
	    velocity = 0.0;
	 else
	    velocity = PixelScale * Dist[Color] / (TimeScale * dt / 60.0);
	 if (dist == 0.0)
	    MI = 0.0;
	 else
	    MI = PixelScale * dx / dist;

	 if ((dt <= MinTime) || (dist <= MinDist) || (velocity <= MinVelocity))
	    if (Debug == TRUE)
	       printf("%4d%7d%4d%9d%7d%11.3f%11.3f%11.3f\n",
		      Color, (int) (0.5 + Xfirst[Color]), (int) (0.5 + Yfirst[Color]),
		      Tfirst[Color], Tlast[Color], dist, velocity, MI);
      }

   if (Debug == TRUE)
   {
      printf("\n\nCELLS TRACKED FOR MINIMUM TIME AND DISTANCE:\n\n");
      printf("Cell   Position   Frame1 Frame2   Displace   Velocity       M.I.\n");
      printf("----------------------------------------------------------------\n");
   }
   for (Color = 2; Color < MaxColor; Color++)
      if ((Tlast[Color] > Tfirst[Color]) &&
	  (Xfirst[Color] >= minX) && (Xfirst[Color] <= maxX))
      {
	 dx = Xlast[Color] - Xfirst[Color];
	 dy = Ylast[Color] - Yfirst[Color];
	 dt = Tlast[Color] - Tfirst[Color];
	 dist = (float) sqrt((double) (dx * dx + dy * dy)) * PixelScale;
	 if (dt == 0)
	    velocity = 0.0;
	 else
	    velocity = PixelScale * Dist[Color] / (TimeScale * dt / 60.0);
	 if (dist == 0.0)
	    MI = 0.0;
	 else
	    MI = PixelScale * dx / dist;

	 if ((dt > MinTime) && (dist > MinDist) && (velocity > MinVelocity))
	 {
	    if (Debug == TRUE)
	       printf("%4d%7d%4d%9d%7d%11.3f%11.3f%11.3f\n",
		      Color, (int) (0.5 + Xfirst[Color]), (int) (0.5 + Yfirst[Color]),
		      Tfirst[Color], Tlast[Color], dist, velocity, MI);
	    AveVelocity += velocity;
	    AveIndex += MI;
	    Status[Color] = 1;
	    TrackCnt++;
	 }
      }

   if (TrackCnt > 0)
   {
      printf("\n\nSUMMARY STATISTICS AND PROGRAM PARAMETERS:\n");
      printf("------------------------------------------\n");
      printf("Experiment elapsed time:   %7.3f seconds\n", TimeScale * frames);
      printf("Minimum tracking time:     %7.3f frames\n", MinTime);
      printf("Minimum tracking distance: %7.3f microns\n", MinDist);
      printf("Minimum tracking velocity: %7.3f microns/minute\n", MinVelocity);
      printf("Number of cells detected:  %4d\n", MaxColor - 2);
      printf("Number of cells tracked:   %4d\n", TrackCnt);
      printf("Pixel size used:           %7.3f microns/pixel\n", PixelScale);
      printf("Time between images:       %7.3f seconds\n", TimeScale);
      printf("Average cell velocity:     %7.3f microns/minute\n",
	     AveVelocity / TrackCnt);
      printf("Average McCutcheon index:  %7.3f\n", AveIndex / TrackCnt);
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
   /* Image variables */
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   SHORT_TYPE **Raw;
   SHORT_TYPE **Data1;
   SHORT_TYPE **Data2;
   SHORT_TYPE **Out;
   SHORT_TYPE **Temp;
   int PixType, DimCnt;

   /* Program variables */
   int i = 0, j, x, y, dx, dy;

   /* Interpret program options */
   printf("TRACK Program - KUIM Version 1.2\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 't':
	 if (sscanf(argv[++i], "%d", &Thresh) == 0)
	    Error("Could not get thresh argument");
	 break;
      case 'n':
	 NewBlobs = TRUE;
	 break;
      case 'q':
	 Quarters = TRUE;
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i < 2)
   {
      fprintf(stderr, "Usage: track [options] outfile infile ...\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-q]    Output statistics by quarters\n");
      fprintf(stderr, "       [-n]    Track new blobs as they are found\n");
      fprintf(stderr, "       [-t #]  Image threshold (128)\n");
      exit(1);
   }

   /* Get output file names from argument list */
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Check dimensions of input images */
   for (j = i; j < argc; j++)
   {
      if (sscanf(argv[j], "%s", Name1) == 0)
	 Error("Could not get input file name");
      Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
      if (DimCnt != 2)
	 Error("Can only process 2D images");
      if (j == i)
      {
	 NewXdim = Xdim;
	 NewYdim = Ydim;
      }
      else if ((NewXdim != Xdim) || (NewYdim != Ydim))
	 Error("Input image dimensions do not match");
   }

   /* Read the first image */
   if (sscanf(argv[i], "%s", Name1) == 0)
      Error("Could not get input file name");
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   Raw = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   Data1 = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   Data2 = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   Out = (SHORT_TYPE **) im_alloc2D(Image1, SHORT);
   im_read(Image1, SHORT, (char *) &(Raw[0][0]));
   fprintf(stderr, "Processing %s\n", Name1);

   /* Preprocess the first image */
   Preprocess(Raw, Data1);

   /* Blob color first image */
   MaxColor = 2;
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 if (Data1[y][x] == 1)
	 {
	    Area = Xpos = Ypos = 0;
	    BlobColor(Data1, x, y, MaxColor);
	    Xfirst[MaxColor] = Xlast[MaxColor] = Xpos / Area;
	    Yfirst[MaxColor] = Ylast[MaxColor] = Ypos / Area;
	    Tfirst[MaxColor] = Tlast[MaxColor] = 0;
	    Status[MaxColor] = 0;
	    Dist[MaxColor] = 0.0;
	    MaxColor++;
	 }

   /* Process remaining images in sequence */
   for (j = i + 1; j < argc; j++)
   {
      /* Read next image */
      if (sscanf(argv[j], "%s", Name1) == 0)
	 Error("Could not get input file name");
      Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
      im_read(Image1, SHORT, (char *) &(Raw[0][0]));
      fprintf(stderr, "Processing %s\n", Name1);

      /* Preprocess next image */
      Preprocess(Raw, Data2);

      /* Track blobs based on their centers */
      for (Color = 2; Color < MaxColor; Color++)
	 if (Tlast[Color] == j - i - 1)
	 {
	    x = Xlast[Color];
	    y = Ylast[Color];
	    if ((Data2[y][x] == 1) && (Data1[y][x] == Color))
	    {
	       Area = Xpos = Ypos = 0;
	       BlobColor(Data2, x, y, Color);
	       DrawLine(Out, Xlast[Color], Ylast[Color],
			Xpos / Area, Ypos / Area, Color);
	       dx = Xlast[Color] - Xpos / Area;
	       dy = Ylast[Color] - Ypos / Area;
	       Dist[Color] += (float) sqrt((double) (dx * dx + dy * dy));
	       Xlast[Color] = Xpos / Area;
	       Ylast[Color] = Ypos / Area;
	       Tlast[Color] = j - i;
	       BlobColor(Data1, x, y, 0);
	    }
	 }

      /* Track remaining blobs */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	    if ((Data2[y][x] == 1) && (Data1[y][x] > 1))
	    {
	       Color = Data1[y][x];
	       Area = Xpos = Ypos = 0;
	       BlobColor(Data2, x, y, Color);
	       DrawLine(Out, Xlast[Color], Ylast[Color],
			Xpos / Area, Ypos / Area, Color);
	       dx = Xlast[Color] - Xpos / Area;
	       dy = Ylast[Color] - Ypos / Area;
	       Dist[Color] += (float) sqrt((double) (dx * dx + dy * dy));
	       Xlast[Color] = Xpos / Area;
	       Ylast[Color] = Ypos / Area;
	       Tlast[Color] = j - i;
	       BlobColor(Data1, x, y, 0);
	    }

      /* Color new blobs */
      if (NewBlobs == TRUE)
	 for (y = 0; y < Ydim; y++)
	    for (x = 0; x < Xdim; x++)
	       if (Data2[y][x] == 1)
	       {
		  Area = Xpos = Ypos = 0;
		  BlobColor(Data2, x, y, MaxColor);
		  Xfirst[MaxColor] = Xlast[MaxColor] = Xpos / Area;
		  Yfirst[MaxColor] = Ylast[MaxColor] = Ypos / Area;
		  Tfirst[MaxColor] = Tlast[MaxColor] = j - i;
		  Status[MaxColor] = 0;
		  Dist[MaxColor] = 0.0;
		  MaxColor++;
	       }

      /* Swap images */
      Temp = Data1;
      Data1 = Data2;
      Data2 = Temp;
   }

   /* Print blob motion data */
   if (Quarters == TRUE)
   {
      printf("\n\nStatistics for [%d .. %d] range:\n", 0, Xdim / 4 - 1);
      printf("============================================\n\n");
      PrintStats(j - i, 0, Xdim / 4 - 1);
      printf("\n\nStatistics for [%d .. %d] range:\n", Xdim / 4, Xdim / 2 - 1);
      printf("============================================\n\n");
      PrintStats(j - i, Xdim / 4, Xdim / 2 - 1);
      printf("\n\nStatistics for [%d .. %d] range:\n", Xdim / 2, 3 * Xdim / 4 - 1);
      printf("============================================\n\n");
      PrintStats(j - i, Xdim / 2, 3 * Xdim / 4 - 1);
      printf("\n\nStatistics for [%d .. %d] range:\n", 3 * Xdim / 4, Xdim - 1);
      printf("============================================\n\n");
      PrintStats(j - i, 3 * Xdim / 4, Xdim - 1);
   }
   printf("\n\nStatistics for [%d .. %d] range:\n", 0, Xdim - 1);
   printf("============================================\n\n");
   PrintStats(j - i, 0, Xdim - 1);

   /* Create output image */
   Status[0] = 0;
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 Out[y][x] = Status[Out[y][x]];
   Image2 = im_create(Name2, SHORT, NewXdim, NewYdim, 1);

   /* Write information to output image */
   im_write(Image2, SHORT, (char *) &(Out[0][0]));
   im_free2D((char **) Raw);
   im_free2D((char **) Data1);
   im_free2D((char **) Data2);
   im_free2D((char **) Out);
   return (0);
}
