/*---------------------------------------------------------------------------*/
/* Program:  vector_flow.c                                                   */
/*                                                                           */
/* Purpose:  This program calculates the optic flow field for a pair         */
/*           of vector valued images (such as RGB) taken from a motion       */
/*           sequence.  Horn's regularization algorithms is used.            */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Usage:    vector_flow [options] infile1 infile2 outfileU outfileV\n       */
/*           options: [-d]   Print debugging information                     */
/*                    [-s]   Print statistic information                     */
/*                            sum(|U|+|V|), sum(U), sum(V)                   */
/*                    [-a #] Smoothness weight (default=0.2)                 */
/*                    [-c #] Number of smoothing iterations (default=30)     */
/*                                                                           */
/* Date:     March 24, 1994  -  Original program.                            */
/*           June 3, 1997    -  Updated Algorithm.                           */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994-1995               */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define TEMPLATE   1
#define VARIATIONAL 2
#define G  1
#define C 2

/* common variables */
int Debug = FALSE;
int Stats = FALSE;
int Xdim, Ydim, Zdim;
IM_TYPE *Image1;
IM_TYPE *Image2;
IM_TYPE *ImageU;
IM_TYPE *ImageV;
FLOAT_TYPE **DataU;
FLOAT_TYPE **DataV;
char Name1[50];
char Name2[50];
char NameU[50];
char NameV[50];
int PixType1, PixType2, DimCnt1, DimCnt2;
int Method = VARIATIONAL;
int Color = G;

/* variables for color functions */
COLOR_TYPE **Data1c = NULL, **Data2c = NULL;
COLOR_TYPE ***Data1c3D = NULL, ***Data2c3D = NULL;

/* variables for grey functions */
FLOAT_TYPE **Data1g = NULL, **Data2g = NULL;
FLOAT_TYPE ***Data1g3D = NULL, ***Data2g3D = NULL;

/* Variational variables */
FLOAT_TYPE **AveU;
FLOAT_TYPE **AveV;
float Alpha = 0.2;
int Count = 30;

/* Template Matching variables */
int SearchSize = 3;
int BlockSize = 3;



/*---------------------------------------------------------------------------*/
/* Purpose:  This routine minimizes functional to find (u,v) values.         */
/*---------------------------------------------------------------------------*/
void GreyVariational2d()
{
   int i, x, y, z;
   float Ix, Iy, It;
   float A1, B1, C1, A2, B2, C2;
   float Temp;

   /* Initialize (u,v) to zero */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Iteratively improve (u,v) estimate */
   for (i = 0; i < Count; i++)
   {
      /* Calculate average (u,v) values */
      if (Debug == TRUE)
	 printf("Iteration %d\n", i + 1);
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    AveU[y][x] = (DataU[y - 1][x] + DataU[y + 1][x]
			  + DataU[y][x - 1] + DataU[y][x + 1]) / 4;
	    AveV[y][x] = (DataV[y - 1][x] + DataV[y + 1][x]
			  + DataV[y][x - 1] + DataV[y][x + 1]) / 4;
	 }

      /* Loop over image */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    /* Initialize A1,B1,C1, A2,B2,C2 terms */
	    A1 = Alpha;
	    B1 = 0.0;
	    C1 = Alpha * AveU[y][x];
	    A2 = 0.0;
	    B2 = Alpha;
	    C2 = Alpha * AveV[y][x];


	    /* Calculate partial derivatives */
	    Ix = (Data1g[y][x + 1] - Data1g[y][x]
		  + Data1g[y + 1][x + 1] - Data1g[y + 1][x]
		  + Data2g[y][x + 1] - Data2g[y][x]
		  + Data2g[y + 1][x + 1] - Data2g[y + 1][x]) / 4;
	    Iy = (Data1g[y + 1][x] - Data1g[y][x]
		  + Data1g[y + 1][x + 1] - Data1g[y][x + 1]
		  + Data2g[y + 1][x] - Data2g[y][x]
		  + Data2g[y + 1][x + 1] - Data2g[y][x + 1]) / 4;
	    It = (Data2g[y][x] - Data1g[y][x]
		  + Data2g[y + 1][x] - Data1g[y + 1][x]
		  + Data2g[y][x + 1] - Data1g[y][x + 1]
		  + Data2g[y + 1][x + 1] - Data1g[y + 1][x + 1]) / 4;

	    /* Update A1,B1,C1, A2,B2,C2 terms */
	    A1 += Ix * Ix;
	    B1 += Ix * Iy;
	    C1 -= Ix * It;
	    A2 += Iy * Ix;
	    B2 += Iy * Iy;
	    C2 -= Iy * It;


	    /* Calculate new (u,v) values */
	    Temp = A1 * B2 - A2 * B1;
	    if (Temp == 0.0)
	    {
	       DataU[y][x] = AveU[y][x];
	       DataV[y][x] = AveV[y][x];
	    }
	    else
	    {
	       DataU[y][x] = (C1 * B2 - C2 * B1) / Temp;
	       DataV[y][x] = (C2 * A1 - C1 * A2) / Temp;
	    }
	 }
   }
}



/*---------------------------------------------------------------------------*/
/* Purpose:  This routine minimizes functional to find (u,v) values 3d.      */
/*---------------------------------------------------------------------------*/
GreyVariational3d()
{
   int i, x, y, z;
   float Ix, Iy, It;
   float A1, B1, C1, A2, B2, C2;
   float Temp;

   /* Initialize (u,v) to zero */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Iteratively improve (u,v) estimate */
   for (i = 0; i < Count; i++)
   {
      /* Calculate average (u,v) values */
      if (Debug == TRUE)
	 printf("Iteration %d\n", i + 1);
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    AveU[y][x] = (DataU[y - 1][x] + DataU[y + 1][x]
			  + DataU[y][x - 1] + DataU[y][x + 1]) / 4;
	    AveV[y][x] = (DataV[y - 1][x] + DataV[y + 1][x]
			  + DataV[y][x - 1] + DataV[y][x + 1]) / 4;
	 }

      /* Loop over image */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    /* Initialize A1,B1,C1, A2,B2,C2 terms */
	    A1 = Alpha;
	    B1 = 0.0;
	    C1 = Alpha * AveU[y][x];
	    A2 = 0.0;
	    B2 = Alpha;
	    C2 = Alpha * AveV[y][x];

	    /* Loop over vector components */
	    for (z = 0; z < Zdim; z++)
	    {
	       /* Calculate partial derivatives */
	       Ix = (Data1g3D[z][y][x + 1] - Data1g3D[z][y][x]
		     + Data1g3D[z][y + 1][x + 1] - Data1g3D[z][y + 1][x]
		     + Data2g3D[z][y][x + 1] - Data2g3D[z][y][x]
		   + Data2g3D[z][y + 1][x + 1] - Data2g3D[z][y + 1][x]) / 4;
	       Iy = (Data1g3D[z][y + 1][x] - Data1g3D[z][y][x]
		     + Data1g3D[z][y + 1][x + 1] - Data1g3D[z][y][x + 1]
		     + Data2g3D[z][y + 1][x] - Data2g3D[z][y][x]
		   + Data2g3D[z][y + 1][x + 1] - Data2g3D[z][y][x + 1]) / 4;
	       It = (Data2g3D[z][y][x] - Data1g3D[z][y][x]
		     + Data2g3D[z][y + 1][x] - Data1g3D[z][y + 1][x]
		     + Data2g3D[z][y][x + 1] - Data1g3D[z][y][x + 1]
	       + Data2g3D[z][y + 1][x + 1] - Data1g3D[z][y + 1][x + 1]) / 4;

	       /* Update A1,B1,C1, A2,B2,C2 terms */
	       A1 += Ix * Ix;
	       B1 += Ix * Iy;
	       C1 -= Ix * It;
	       A2 += Iy * Ix;
	       B2 += Iy * Iy;
	       C2 -= Iy * It;
	    }

	    /* Calculate new (u,v) values */
	    Temp = A1 * B2 - A2 * B1;
	    if (Temp == 0.0)
	    {
	       DataU[y][x] = AveU[y][x];
	       DataV[y][x] = AveV[y][x];
	    }
	    else
	    {
	       DataU[y][x] = (C1 * B2 - C2 * B1) / Temp;
	       DataV[y][x] = (C2 * A1 - C1 * A2) / Temp;
	    }
	 }
   }
}


/*---------------------------------------------------------------------------*/
/* Purpose:  This routine minimizes functional to find (u,v) values.         */
/*---------------------------------------------------------------------------*/
void ColorVariational2d()
{
   int i, x, y, z;
   double rIx, rIy, rIt, gIx, gIy, gIt, bIx, bIy, bIt;
   double A1, B1, C1, A2, B2, C2;
   double Temp;

   /* Initialize (u,v) to zero */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Iteratively improve (u,v) estimate */
   for (i = 0; i < Count; i++)
   {
      /* Calculate average (u,v) values */
      if (Debug == TRUE)
	 printf("Iteration %d\n", i + 1);
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    AveU[y][x] = (DataU[y - 1][x] + DataU[y + 1][x]
			  + DataU[y][x - 1] + DataU[y][x + 1]) / 4;
	    AveV[y][x] = (DataV[y - 1][x] + DataV[y + 1][x]
			  + DataV[y][x - 1] + DataV[y][x + 1]) / 4;
	 }

      /* Loop over image */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    /* Initialize A1,B1,C1, A2,B2,C2 terms */
	    A1 = Alpha;
	    B1 = 0.0;
	    C1 = Alpha * AveU[y][x];
	    A2 = 0.0;
	    B2 = Alpha;
	    C2 = Alpha * AveV[y][x];


	    /* Calculate partial derivatives */
	    rIx = (Data1c[y][x + 1].r - Data1c[y][x].r
		   + Data1c[y + 1][x + 1].r - Data1c[y + 1][x].r
		   + Data2c[y][x + 1].r - Data2c[y][x].r
		   + Data2c[y + 1][x + 1].r - Data2c[y + 1][x].r) / 4;
	    rIy = (Data1c[y + 1][x].r - Data1c[y][x].r
		   + Data1c[y + 1][x + 1].r - Data1c[y][x + 1].r
		   + Data2c[y + 1][x].r - Data2c[y][x].r
		   + Data2c[y + 1][x + 1].r - Data2c[y][x + 1].r) / 4;
	    rIt = (Data2c[y][x].r - Data1c[y][x].r
		   + Data2c[y + 1][x].r - Data1c[y + 1][x].r
		   + Data2c[y][x + 1].r - Data1c[y][x + 1].r
		   + Data2c[y + 1][x + 1].r - Data1c[y + 1][x + 1].r) / 4;

	    gIx = (Data1c[y][x + 1].g - Data1c[y][x].g
		   + Data1c[y + 1][x + 1].g - Data1c[y + 1][x].g
		   + Data2c[y][x + 1].g - Data2c[y][x].g
		   + Data2c[y + 1][x + 1].g - Data2c[y + 1][x].g) / 4;
	    gIy = (Data1c[y + 1][x].g - Data1c[y][x].g
		   + Data1c[y + 1][x + 1].g - Data1c[y][x + 1].g
		   + Data2c[y + 1][x].g - Data2c[y][x].g
		   + Data2c[y + 1][x + 1].g - Data2c[y][x + 1].g) / 4;
	    gIt = (Data2c[y][x].g - Data1c[y][x].g
		   + Data2c[y + 1][x].g - Data1c[y + 1][x].g
		   + Data2c[y][x + 1].g - Data1c[y][x + 1].g
		   + Data2c[y + 1][x + 1].g - Data1c[y + 1][x + 1].g) / 4;

	    bIx = (Data1c[y][x + 1].b - Data1c[y][x].b
		   + Data1c[y + 1][x + 1].b - Data1c[y + 1][x].b
		   + Data2c[y][x + 1].b - Data2c[y][x].b
		   + Data2c[y + 1][x + 1].b - Data2c[y + 1][x].b) / 4;
	    bIy = (Data1c[y + 1][x].b - Data1c[y][x].b
		   + Data1c[y + 1][x + 1].b - Data1c[y][x + 1].b
		   + Data2c[y + 1][x].b - Data2c[y][x].b
		   + Data2c[y + 1][x + 1].b - Data2c[y][x + 1].b) / 4;
	    bIt = (Data2c[y][x].b - Data1c[y][x].b
		   + Data2c[y + 1][x].b - Data1c[y + 1][x].b
		   + Data2c[y][x + 1].b - Data1c[y][x + 1].b
		   + Data2c[y + 1][x + 1].b - Data1c[y + 1][x + 1].b) / 4;

	    /* Update A1,B1,C1, A2,B2,C2 terms */
	    A1 += rIx * rIx + gIx * gIx + bIx * bIx;
	    B1 += rIx * rIy + gIx * gIy + bIx * bIy;
	    C1 -= rIx * rIt + gIx * gIt + bIx * bIt;
	    A2 += rIy * rIx + gIy * gIx + bIy * bIx;
	    B2 += rIy * rIy + gIy * gIy + bIy * bIy;
	    C2 -= rIy * rIt + gIy * gIt + bIy * bIt;


	    /* Calculate new (u,v) values */
	    Temp = A1 * B2 - A2 * B1;
	    if (Temp == 0.0)
	    {
	       DataU[y][x] = AveU[y][x];
	       DataV[y][x] = AveV[y][x];
	    }
	    else
	    {
	       DataU[y][x] = (C1 * B2 - C2 * B1) / Temp;
	       DataV[y][x] = (C2 * A1 - C1 * A2) / Temp;
	    }
	 }
   }
}



/*---------------------------------------------------------------------------*/
/* Purpose:  This routine minimizes functional to find (u,v) values 3d.      */
/*---------------------------------------------------------------------------*/
ColorVariational3d()
{
   int i, x, y, z;
   float rIx, rIy, rIt, gIx, gIy, gIt, bIx, bIy, bIt;
   float A1, B1, C1, A2, B2, C2;
   float Temp;

   /* Initialize (u,v) to zero */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Iteratively improve (u,v) estimate */
   for (i = 0; i < Count; i++)
   {
      /* Calculate average (u,v) values */
      if (Debug == TRUE)
	 printf("Iteration %d\n", i + 1);
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    AveU[y][x] = (DataU[y - 1][x] + DataU[y + 1][x]
			  + DataU[y][x - 1] + DataU[y][x + 1]) / 4;
	    AveV[y][x] = (DataV[y - 1][x] + DataV[y + 1][x]
			  + DataV[y][x - 1] + DataV[y][x + 1]) / 4;
	 }

      /* Loop over image */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    /* Initialize A1,B1,C1, A2,B2,C2 terms */
	    A1 = Alpha;
	    B1 = 0.0;
	    C1 = Alpha * AveU[y][x];
	    A2 = 0.0;
	    B2 = Alpha;
	    C2 = Alpha * AveV[y][x];

	    /* Loop over vector components */
	    for (z = 0; z < Zdim; z++)
	    {
	       /* Calculate partial derivatives */
	       rIx = (Data1c3D[z][y][x + 1].r - Data1c3D[z][y][x].r
		      + Data1c3D[z][y + 1][x + 1].r - Data1c3D[z][y + 1][x].r
		      + Data2c3D[z][y][x + 1].r - Data2c3D[z][y][x].r
	       + Data2c3D[z][y + 1][x + 1].r - Data2c3D[z][y + 1][x].r) / 4;
	       rIy = (Data1c3D[z][y + 1][x].r - Data1c3D[z][y][x].r
		      + Data1c3D[z][y + 1][x + 1].r - Data1c3D[z][y][x + 1].r
		      + Data2c3D[z][y + 1][x].r - Data2c3D[z][y][x].r
	       + Data2c3D[z][y + 1][x + 1].r - Data2c3D[z][y][x + 1].r) / 4;
	       rIt = (Data2c3D[z][y][x].r - Data1c3D[z][y][x].r
		      + Data2c3D[z][y + 1][x].r - Data1c3D[z][y + 1][x].r
		      + Data2c3D[z][y][x + 1].r - Data1c3D[z][y][x + 1].r
		      + Data2c3D[z][y + 1][x + 1].r - Data1c3D[z][y + 1][x + 1].r) / 4;

	       gIx = (Data1c3D[z][y][x + 1].g - Data1c3D[z][y][x].g
		      + Data1c3D[z][y + 1][x + 1].g - Data1c3D[z][y + 1][x].g
		      + Data2c3D[z][y][x + 1].g - Data2c3D[z][y][x].g
	       + Data2c3D[z][y + 1][x + 1].g - Data2c3D[z][y + 1][x].g) / 4;
	       gIy = (Data1c3D[z][y + 1][x].g - Data1c3D[z][y][x].g
		      + Data1c3D[z][y + 1][x + 1].g - Data1c3D[z][y][x + 1].g
		      + Data2c3D[z][y + 1][x].g - Data2c3D[z][y][x].g
	       + Data2c3D[z][y + 1][x + 1].g - Data2c3D[z][y][x + 1].g) / 4;
	       gIt = (Data2c3D[z][y][x].g - Data1c3D[z][y][x].g
		      + Data2c3D[z][y + 1][x].g - Data1c3D[z][y + 1][x].g
		      + Data2c3D[z][y][x + 1].g - Data1c3D[z][y][x + 1].g
		      + Data2c3D[z][y + 1][x + 1].g - Data1c3D[z][y + 1][x + 1].g) / 4;

	       bIx = (Data1c3D[z][y][x + 1].b - Data1c3D[z][y][x].b
		      + Data1c3D[z][y + 1][x + 1].b - Data1c3D[z][y + 1][x].b
		      + Data2c3D[z][y][x + 1].b - Data2c3D[z][y][x].b
	       + Data2c3D[z][y + 1][x + 1].b - Data2c3D[z][y + 1][x].b) / 4;
	       bIy = (Data1c3D[z][y + 1][x].b - Data1c3D[z][y][x].b
		      + Data1c3D[z][y + 1][x + 1].b - Data1c3D[z][y][x + 1].b
		      + Data2c3D[z][y + 1][x].b - Data2c3D[z][y][x].b
	       + Data2c3D[z][y + 1][x + 1].b - Data2c3D[z][y][x + 1].b) / 4;
	       bIt = (Data2c3D[z][y][x].b - Data1c3D[z][y][x].b
		      + Data2c3D[z][y + 1][x].b - Data1c3D[z][y + 1][x].b
		      + Data2c3D[z][y][x + 1].b - Data1c3D[z][y][x + 1].b
		      + Data2c3D[z][y + 1][x + 1].b - Data1c3D[z][y + 1][x + 1].b) / 4;

	       /* Update A1,B1,C1, A2,B2,C2 terms */
	       A1 += rIx * rIx + gIx * gIx + bIx * bIx;
	       B1 += rIx * rIy + gIx * gIy + bIx * bIy;
	       C1 -= rIx * rIt + gIx * gIt + bIx * bIt;
	       A2 += rIy * rIx + gIy * gIx + bIy * bIx;
	       B2 += rIy * rIy + gIy * gIy + bIy * bIy;
	       C2 -= rIy * rIt + gIy * gIt + bIy * bIt;

	    }

	    /* Calculate new (u,v) values */
	    Temp = A1 * B2 - A2 * B1;
	    if (Temp == 0.0)
	    {
	       DataU[y][x] = AveU[y][x];
	       DataV[y][x] = AveV[y][x];
	    }
	    else
	    {
	       DataU[y][x] = (C1 * B2 - C2 * B1) / Temp;
	       DataV[y][x] = (C2 * A1 - C1 * A2) / Temp;
	    }
	 }
   }
}





/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates optic flow of Grey images using block   */
/*           matching.                                                       */
/*---------------------------------------------------------------------------*/
void GreyTemplateMatching2d()
{
   int x, y, dx, dy, bx, by, minU, minV;
   float Temp, Diff, MinDiff;

   /* Initialize output images */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Calculate optic flow */
   for (y = SearchSize + BlockSize; y < Ydim - SearchSize - BlockSize; y++)
      for (x = SearchSize + BlockSize; x < Xdim - SearchSize - BlockSize; x++)
      {
	 /* Set initial match error */
	 MinDiff = 0;
	 minU = minV = 0;
	 for (by = -BlockSize; by <= BlockSize; by++)
	    for (bx = -BlockSize; bx <= BlockSize; bx++)
	    {
	       Temp = Data2g[y + by][x + bx]
		  - Data1g[y + by][x + bx];
	       if (Temp < 0)
		  MinDiff -= Temp;
	       else
		  MinDiff += Temp;
	    }

	 /* Search for best block match at each point */
	 for (dy = -SearchSize; dy <= SearchSize; dy++)
	    for (dx = -SearchSize; dx <= SearchSize; dx++)
	    {
	       Diff = 0;
	       for (by = -BlockSize; by <= BlockSize; by++)
		  for (bx = -BlockSize; bx <= BlockSize; bx++)
		  {
		     Temp = Data2g[y + by][x + bx]
			- Data1g[y + by - dy][x + bx - dx];
		     if (Temp < 0)
			Diff -= Temp;
		     else
			Diff += Temp;
		  }
	       if (Diff < MinDiff)
	       {
		  MinDiff = Diff;
		  minU = dx;
		  minV = dy;
	       }
	    }
	 DataU[y][x] = minU;
	 DataV[y][x] = minV;
      }

}


/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates optic flow of Color images using block  */
/*           matching.                                                       */
/*---------------------------------------------------------------------------*/
void ColorTemplateMatching2d()
{
   int x, y, dx, dy, bx, by, minU, minV;
   float TempRed, TempGreen, TempBlue, Diff, MinDiff;


   /* Initialize output images */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
	 DataU[y][x] = DataV[y][x] = 0.0;

   /* Calculate optic flow */
   for (y = SearchSize + BlockSize; y < Ydim - SearchSize - BlockSize; y++)
      for (x = SearchSize + BlockSize; x < Xdim - SearchSize - BlockSize; x++)
      {
	 /* Set initial match error */
	 MinDiff = 0;
	 minU = minV = 0;
	 for (by = -BlockSize; by <= BlockSize; by++)
	    for (bx = -BlockSize; bx <= BlockSize; bx++)
	    {
	       TempRed = (float) Data2c[y + by][x + bx].r -
		  (float) Data1c[y + by][x + bx].r;
	       TempGreen = (float) Data2c[y + by][x + bx].g -
		  (float) Data1c[y + by][x + bx].g;
	       TempBlue = (float) Data2c[y + by][x + bx].b -
		  (float) Data1c[y + by][x + bx].b;
	       if (TempRed < 0)
		  TempRed = -TempRed;
	       if (TempGreen < 0)
		  TempGreen = -TempGreen;
	       if (TempBlue < 0)
		  TempBlue = -TempBlue;
	       MinDiff += TempRed + TempGreen + TempBlue;
	    }

	 /* Search for best block match at each point */
	 for (dy = -SearchSize; dy <= SearchSize; dy++)
	    for (dx = -SearchSize; dx <= SearchSize; dx++)
	    {
	       Diff = 0;
	       for (by = -BlockSize; by <= BlockSize; by++)
		  for (bx = -BlockSize; bx <= BlockSize; bx++)
		  {
		     TempRed = (float) Data2c[y + by][x + bx].r -
			(float) Data1c[y + by - dy][x + bx - dx].r;
		     TempGreen = (float) Data2c[y + by][x + bx].g -
			(float) Data1c[y + by - dy][x + bx - dx].g;
		     TempBlue = (float) Data2c[y + by][x + bx].b -
			(float) Data1c[y + by - dy][x + bx - dx].b;
		     if (TempRed < 0)
			TempRed = -TempRed;
		     if (TempGreen < 0)
			TempGreen = -TempGreen;
		     if (TempBlue < 0)
			TempBlue = -TempBlue;
		     Diff += TempRed + TempGreen + TempBlue;
		  }
	       if (Diff < MinDiff)
	       {
		  MinDiff = Diff;
		  minU = dx;
		  minV = dy;
	       }
	    }
	 DataU[y][x] = minU;
	 DataV[y][x] = minV;
      }

}




/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Program variables */
   int i = 0;
   int x, y;
   double SumUV;
   double SumU;
   double SumV;

   /* Interpret program options */
   printf("OPTIC_FLOW Program - KUIM Version 2.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'v':
	 Method = VARIATIONAL;
	 Color = G;
	 break;
      case 'V':
	 Method = VARIATIONAL;
	 Color = C;
	 break;
      case 't':
	 Method = TEMPLATE;
	 Color = G;
	 break;
      case 'T':
	 Method = TEMPLATE;
	 Color = C;
	 break;
      case 's':
	 if (sscanf(argv[++i], "%f", &SearchSize) == 0)
	    Error("Can not get search size");
	 break;
      case 'b':
	 if (sscanf(argv[++i], "%d", &BlockSize) == 0)
	    Error("Can not get Block Size");
	 break;
      case 'a':
	 if (sscanf(argv[++i], "%f", &Alpha) == 0)
	    Error("Can not get smoothness weight");
	 break;
      case 'c':
	 if (sscanf(argv[++i], "%d", &Count) == 0)
	    Error("Can not get iteration count");
	 break;
      case 'd':
	 Debug = TRUE;
	 break;
      case 'i':
	 Stats = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 4)
   {
      fprintf(stderr,
       "Usage: optic_flow [options] infile1 infile2 outfileU outfileV\n\n");
      fprintf(stderr, "    [-d]     Print debugging information\n");
      fprintf(stderr, "    [-i]     Print statistic information\n\n");
      fprintf(stderr, "    [-t]     grey template matching optical flow\n");
      fprintf(stderr, "    [-T]     color template matching optical flow\n");
      fprintf(stderr, "     [-s #]  Size of search window (3)\n");
      fprintf(stderr, "     [-b #]  Size of the matching block (3)\n\n");
      fprintf(stderr, "    [-v]     grey variational optical flow\n");
      fprintf(stderr, "    [-V]     color variational optical flow\n");
      fprintf(stderr, "     [-a #]  Smoothness weight (0.2)\n");
      fprintf(stderr, "     [-c #]  Number of smoothing iterations (30)\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", NameU) == 0)
      Error("Could not get output file name");
   if (sscanf(argv[i++], "%s", NameV) == 0)
      Error("Could not get output file name");

   /* Read input image1 */
   Image1 = im_open(Name1, &PixType1, &Xdim, &Ydim, &Zdim, &DimCnt1);
   Image2 = im_open(Name2, &PixType2, &Xdim, &Ydim, &Zdim, &DimCnt2);

   if ((DimCnt1 < 2) || (DimCnt2 < 2))
      Error("Can only process 2D or 3d images");

   /* Create output U,V images */
   ImageU = im_create(NameU, FLOAT, Xdim, Ydim, 1);
   ImageV = im_create(NameV, FLOAT, Xdim, Ydim, 1);
   DataU = (FLOAT_TYPE **) im_alloc2D(ImageU, FLOAT);
   DataV = (FLOAT_TYPE **) im_alloc2D(ImageV, FLOAT);
   AveU = (FLOAT_TYPE **) im_alloc2D(ImageU, FLOAT);
   AveV = (FLOAT_TYPE **) im_alloc2D(ImageV, FLOAT);

   if (Color == C)
   {
      if ((DimCnt1 == 2) && (DimCnt2 == 2))
      {
	 Data1c = (COLOR_TYPE **) im_alloc2D(Image1, COLOR);
	 Data2c = (COLOR_TYPE **) im_alloc2D(Image2, COLOR);
	 im_read(Image1, COLOR, (char *) &(Data1c[0][0]));
	 im_read(Image2, COLOR, (char *) &(Data2c[0][0]));
	 if (Method == VARIATIONAL)
	    ColorVariational2d();
	 else
	    ColorTemplateMatching2d();
      }
      else
      {
	 Data1c3D = (COLOR_TYPE ***) im_alloc3D(Image1, COLOR);
	 Data2c3D = (COLOR_TYPE ***) im_alloc3D(Image2, COLOR);
	 im_read(Image1, COLOR, (char *) &(Data1c3D[0][0][0]));
	 im_read(Image2, COLOR, (char *) &(Data2c3D[0][0][0]));
	 if (Method == VARIATIONAL)
	    ColorVariational3d();
	 /* else
	    ColorTemplateMatching3d(); */
      }
   }
   else
   {
      if ((DimCnt1 == 2) && (DimCnt2 == 2))
      {
	 Data1g = (FLOAT_TYPE **) im_alloc2D(Image1, FLOAT);
	 Data2g = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
	 im_read(Image1, FLOAT, (char *) &(Data1g[0][0]));
	 im_read(Image2, FLOAT, (char *) &(Data2g[0][0]));
	 if (Method == VARIATIONAL)
	    GreyVariational2d();
	 else
	    GreyTemplateMatching2d();
      }
      else
      {
	 Data1g3D = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
	 Data2g3D = (FLOAT_TYPE ***) im_alloc3D(Image2, FLOAT);
	 im_read(Image1, FLOAT, (char *) &(Data1g3D[0][0][0]));
	 im_read(Image2, FLOAT, (char *) &(Data2g3D[0][0][0]));
	 if (Method == VARIATIONAL)
	    GreyVariational3d();
	 /*else
	    GreyTemplateMatching3d(); */
      }
   }


   /* print statistic information */
   if (Stats == TRUE)
   {
      SumU = SumV = SumUV = 0.0;
      /* the loop does not include the borders */
      for (y = 1; y < (Ydim - 1); y++)
	 for (x = 1; x < (Xdim - 1); x++)
	 {
	    SumU += DataU[y][x];
	    SumV += DataV[y][x];
	    SumUV += (DataU[y][x] >= 0.0) ? DataU[y][x] : -DataU[y][x];
	    SumUV += (DataV[y][x] >= 0.0) ? DataV[y][x] : -DataV[y][x];
	 }
      printf("sum( |U| + |V| ) = %f per pixel\n", SumUV / ((Xdim - 2) * (Ydim - 2)));
      printf("sum( U ) = %f per pixel\n", SumU / ((Xdim - 2) * (Ydim - 2)));
      printf("sum( V ) = %f per pixel\n", SumV / ((Xdim - 2) * (Ydim - 2)));
   }

   /* Write information to output image */
   im_write(ImageU, FLOAT, (char *) &(DataU[0][0]));
   im_write(ImageV, FLOAT, (char *) &(DataV[0][0]));

   if (Data1g != NULL)
      im_free2D((char **) Data1g);
   if (Data2g != NULL)
      im_free2D((char **) Data2g);
   if (Data1c != NULL)
      im_free2D((char **) Data1c);
   if (Data2c != NULL)
      im_free2D((char **) Data2c);
   if (Data1g3D != NULL)
      im_free3D((char ***) Data1g3D);
   if (Data2g3D != NULL)
      im_free3D((char ***) Data2g3D);
   if (Data1c3D != NULL)
      im_free3D((char ***) Data1c3D);
   if (Data2c3D != NULL)
      im_free3D((char ***) Data2c3D);

   im_free2D((char **) DataU);
   im_free2D((char **) DataV);
   exit(0);
}
