/*---------------------------------------------------------------------------*/
/* Program:  water_merge.c                                                   */
/*                                                                           */
/* Purpose:  This program performs image segmentation by merging watershed   */
/*           regions to minimize the segmentation functional proposed by     */
/*           Mumford and Shah in 1985 CVPR.  A number of input parameters    */
/*           control how many internal iterations are performed and how      */
/*           similar regions must be in order to be merged.                  */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     March 29, 1996                                                  */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1996                    */
/*---------------------------------------------------------------------------*/
#include <IM.h>
#define MAXR 500000
#define MERGE_GRAD 1
#define MERGE_MEAN 2
#define MERGE_VAR  3

/* Global variables */
int Debug = FALSE;
int Method = MERGE_MEAN;
int Xdim, Ydim, Zdim;
int NumSmooth = 5;
int NumMerge = 20;
float Alpha = 0.5;
float Threshold = 20;
struct InfoRec
{
   int R1, R2;
   int Count;
   float Grad;
};
struct InfoRec Info[MAXR];
int TotalCount;
int Parent[MAXR];
int Size[MAXR];
float SumA1[MAXR];
float SumB1[MAXR];
float SumC1[MAXR];
float SumA2[MAXR];
float SumB2[MAXR];
float SumC2[MAXR];
int MinCount = 1;
int HighIndex = 0;

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine presmooths the image.                              */
/*---------------------------------------------------------------------------*/
presmooth(FLOAT_TYPE *** In, FLOAT_TYPE *** Out)
{
   int x, y, X, Y, Count;
   float Sum;

   /* Calculate local average 0th component */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Count = 0;
	 Sum = 0.0;
	 for (Y = (y - 1); Y <= (y + 1); Y++)
	    for (X = (x - 1); X <= (x + 1); X++)
	       if ((X > 0) && (Y > 0) && (X < Xdim) && (Y < Ydim))
	       {
		  Count++;
		  Sum += In[0][Y][X];
	       }
	 Out[0][y][x] = (In[0][y][x] + Alpha * Sum) / (1 + Alpha * Count);
      }

   /* Calculate local average 1st component */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Count = 0;
	 Sum = 0.0;
	 for (Y = (y - 1); Y <= (y + 1); Y++)
	    for (X = (x - 1); X <= (x + 1); X++)
	       if ((X > 0) && (Y > 0) && (X < Xdim) && (Y < Ydim))
	       {
		  Count++;
		  Sum += In[1][Y][X];
	       }
	 Out[1][y][x] = (In[1][y][x] + Alpha * Sum) / (1 + Alpha * Count);
      }

   /* Calculate local average 2nd component */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Count = 0;
	 Sum = 0.0;
	 for (Y = (y - 1); Y <= (y + 1); Y++)
	    for (X = (x - 1); X <= (x + 1); X++)
	       if ((X > 0) && (Y > 0) && (X < Xdim) && (Y < Ydim))
	       {
		  Count++;
		  Sum += In[2][Y][X];
	       }
	 Out[2][y][x] = (In[2][y][x] + Alpha * Sum) / (1 + Alpha * Count);
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates the gradient.                           */
/*---------------------------------------------------------------------------*/
gradient(FLOAT_TYPE *** In, FLOAT_TYPE ** Out)
{
   int x, y;
   float Dx, Dy;

   /* Calculate gradient magnitude 0th component */
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 Dx = (In[0][y + 1][x + 1] + 2 * In[0][y][x + 1] + In[0][y - 1][x + 1]
	       - In[0][y + 1][x - 1] - 2 * In[0][y][x - 1] - In[0][y - 1][x - 1]) / 8.0;
	 Dy = (In[0][y + 1][x + 1] + 2 * In[0][y + 1][x] + In[0][y + 1][x - 1]
	       - In[0][y - 1][x + 1] - 2 * In[0][y - 1][x] - In[0][y - 1][x - 1]) / 8.0;
	 Out[y][x] = (FLOAT_TYPE) sqrt((double) (Dx * Dx + Dy * Dy));
      }

   /* Calculate gradient magnitude 1st component */
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 Dx = (In[1][y + 1][x + 1] + 2 * In[1][y][x + 1] + In[1][y - 1][x + 1]
	       - In[1][y + 1][x - 1] - 2 * In[1][y][x - 1] - In[1][y - 1][x - 1]) / 8.0;
	 Dy = (In[1][y + 1][x + 1] + 2 * In[1][y + 1][x] + In[1][y + 1][x - 1]
	       - In[1][y - 1][x + 1] - 2 * In[1][y - 1][x] - In[1][y - 1][x - 1]) / 8.0;
	 Out[y][x] += (FLOAT_TYPE) sqrt((double) (Dx * Dx + Dy * Dy));
      }

   /* Calculate gradient magnitude 2nd component */
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 Dx = (In[2][y + 1][x + 1] + 2 * In[2][y][x + 1] + In[2][y - 1][x + 1]
	       - In[2][y + 1][x - 1] - 2 * In[2][y][x - 1] - In[2][y - 1][x - 1]) / 8.0;
	 Dy = (In[2][y + 1][x + 1] + 2 * In[2][y + 1][x] + In[2][y + 1][x - 1]
	       - In[2][y - 1][x + 1] - 2 * In[2][y - 1][x] - In[2][y - 1][x - 1]) / 8.0;
	 Out[y][x] += (FLOAT_TYPE) sqrt((double) (Dx * Dx + Dy * Dy));
      }

   /* Fix up the image boundary */
   for (x = 0; x < Xdim; x++)
   {
      Out[0][x] = Out[1][x];
      Out[Ydim - 1][x] = Out[Ydim - 2][x];
   }
   for (y = 0; y < Ydim; y++)
   {
      Out[y][0] = Out[y][1];
      Out[y][Xdim - 1] = Out[y][Xdim - 2];
   }

}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates the watersheds.                         */
/*---------------------------------------------------------------------------*/
watershed(FLOAT_TYPE *** In, FLOAT_TYPE ** Grad, SHORT_TYPE ** Out)
{
   int i, x, y, X, Y;
   int MinIndex;
   float MinValue;
   register int index;
   register float temp0, temp1, temp2;

   /* Initialize region statistics */
   for (i = 0; i < MAXR; i++)
   {
      Size[i] = 0;
      SumA1[i] = SumB1[i] = SumC1[i] = 0.0;
      SumA2[i] = SumB2[i] = SumC2[i] = 0.0;
   }

   /* Mark boundary as pseudo-minima */
   for (y = 0; y < Ydim; y++)
      Out[y][0] = Out[y][Xdim - 1] = MinCount;
   for (x = 0; x < Xdim; x++)
      Out[0][x] = Out[Ydim - 1][x] = MinCount;

   /* Find local minima and gradient for each pixel */
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 MinValue = Grad[y][x];
	 MinIndex = 4;
	 for (Y = 0; Y < 3; Y++)
	    for (X = 0; X < 3; X++)
	       if (Grad[y + Y - 1][x + X - 1] <= MinValue)
	       {
		  MinValue = Grad[y + Y - 1][x + X - 1];
		  MinIndex = Y * 3 + X;
	       }
	 if (MinIndex == 4)
	    Out[y][x] = (++MinCount);
	 else
	    Out[y][x] = -MinIndex;
      }

   /* Follow gradient downhill for each pixel */
   for (y = 1; y < (Ydim - 1); y++)
      for (x = 1; x < (Xdim - 1); x++)
      {
	 X = x;
	 Y = y;
	 while (Out[Y][X] <= 0)
	 {
	    switch (Out[Y][X])
	    {
	    case 0:
	       X--;
	       Y--;
	       break;
	    case -1:
	       Y--;
	       break;
	    case -2:
	       X++;
	       Y--;
	       break;
	    case -3:
	       X--;
	       break;
	    case -5:
	       X++;
	       break;
	    case -6:
	       X--;
	       Y++;
	       break;
	    case -7:
	       Y++;
	       break;
	    case -8:
	       X++;
	       Y++;
	       break;
	    }
	 }
	 index = Out[y][x] = Out[Y][X];
	 temp0 = In[0][y][x];
	 temp1 = In[1][y][x];
	 temp2 = In[2][y][x];
	 Size[index]++;
	 SumA1[index] += temp0;
	 SumB1[index] += temp1;
	 SumC1[index] += temp2;
	 SumA2[index] += temp0 * temp0;
	 SumB2[index] += temp1 * temp1;
	 SumC2[index] += temp2 * temp2;
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine smooths the image within watersheds.               */
/*---------------------------------------------------------------------------*/
smooth(FLOAT_TYPE *** Raw, FLOAT_TYPE *** In,
       SHORT_TYPE ** Region, FLOAT_TYPE ** Out)
{
   int i, x, y, X, Y, Count;
   float Sum, Change;

   /* Perform iterative smoothing */
   for (i = 0, Change = 0.0; i < NumSmooth; i++)
   {
      /* Average 0th component watershed regions */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    Count = -1;
	    Sum = -In[0][y][x];
	    for (Y = (y - 1); Y <= (y + 1); Y++)
	       for (X = (x - 1); X <= (x + 1); X++)
		  if ((X > 0) && (Y > 0) && (X < Xdim) && (Y < Ydim) &&
		      (Region[y][x] == Region[Y][X]))
		  {
		     Count++;
		     Sum += In[0][Y][X];
		  }
	    Out[y][x] = (Raw[0][y][x] + Alpha * Sum) / (1 + Alpha * Count);
	 }

      /* Copy the data back for next iteration */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    if (In[0][y][x] > Out[y][x])
	       Change += (In[0][y][x] - Out[y][x]);
	    else
	       Change -= (In[0][y][x] - Out[y][x]);
	    In[0][y][x] = Out[y][x];
	 }

      /* Average 1st component watershed regions */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    Count = -1;
	    Sum = -In[1][y][x];
	    for (Y = (y - 1); Y <= (y + 1); Y++)
	       for (X = (x - 1); X <= (x + 1); X++)
		  if ((X > 0) && (Y > 0) && (X < Xdim) && (Y < Ydim) &&
		      (Region[y][x] == Region[Y][X]))
		  {
		     Count++;
		     Sum += In[1][Y][X];
		  }
	    Out[y][x] = (Raw[1][y][x] + Alpha * Sum) / (1 + Alpha * Count);
	 }

      /* Copy the data back for next iteration */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    if (In[1][y][x] > Out[y][x])
	       Change += (In[1][y][x] - Out[y][x]);
	    else
	       Change -= (In[1][y][x] - Out[y][x]);
	    In[1][y][x] = Out[y][x];
	 }

      /* Average 2nd component watershed regions */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    Count = -1;
	    Sum = -In[2][y][x];
	    for (Y = (y - 1); Y <= (y + 1); Y++)
	       for (X = (x - 1); X <= (x + 1); X++)
		  if ((X > 0) && (Y > 0) && (X < Xdim) && (Y < Ydim) &&
		      (Region[y][x] == Region[Y][X]))
		  {
		     Count++;
		     Sum += In[2][Y][X];
		  }
	    Out[y][x] = (Raw[2][y][x] + Alpha * Sum) / (1 + Alpha * Count);
	 }

      /* Copy the data back for next iteration */
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	 {
	    if (In[2][y][x] > Out[y][x])
	       Change += (In[2][y][x] - Out[y][x]);
	    else
	       Change -= (In[2][y][x] - Out[y][x]);
	    In[2][y][x] = Out[y][x];
	 }

      /* Print debug info */
      if (Debug == TRUE)
	 printf("Iteration %d had average change %f\n", i, Change / (Xdim * Ydim));
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates difference in region means.             */
/*---------------------------------------------------------------------------*/
float meandiff(int Region1, int Region2)
{
   float temp, diff = 0.0;

   /* Find parents of regions */
   while (Parent[Region1] > 0)
      Region1 = Parent[Region1];
   while (Parent[Region2] > 0)
      Region2 = Parent[Region2];

   /* Calculate difference in means */
   temp = SumA1[Region1] / Size[Region1] - SumA1[Region2] / Size[Region2];
   if (temp > 0)
      diff += temp;
   else
      diff -= temp;
   temp = SumB1[Region1] / Size[Region1] - SumB1[Region2] / Size[Region2];
   if (temp > 0)
      diff += temp;
   else
      diff -= temp;
   temp = SumC1[Region1] / Size[Region1] - SumC1[Region2] / Size[Region2];
   if (temp > 0)
      diff += temp;
   else
      diff -= temp;
   return (diff);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine calculates combined region variance.               */
/*---------------------------------------------------------------------------*/
float variance(int Region1, int Region2)
{
   int count;
   float mean;
   float var;

   /* Find parents of regions */
   while (Parent[Region1] > 0)
      Region1 = Parent[Region1];
   while (Parent[Region2] > 0)
      Region2 = Parent[Region2];

   /* Calculate combined region variance */
   count = Size[Region1] + Size[Region2];
   mean = (SumA1[Region1] + SumA1[Region2]) / count;
   var = (SumA2[Region1] + SumA2[Region2]) / count - mean * mean;
   mean = (SumB1[Region1] + SumB1[Region2]) / count;
   var += (SumB2[Region1] + SumB2[Region2]) / count - mean * mean;
   mean = (SumC1[Region1] + SumC1[Region2]) / count;
   var += (SumC2[Region1] + SumC2[Region2]) / count - mean * mean;
   return (var);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine saves info about adjacent regions.                 */
/*---------------------------------------------------------------------------*/
save(int Region1, int Region2, float Grad)
{
   int Index;

   /* Swap region numbers if needed */
   if (Region1 < Region2)
   {
      Index = Region1;
      Region1 = Region2;
      Region2 = Index;
   }

   /* Search for existing entry */
   Index = Region1 * 10;
   while ((Index < MAXR) && (Info[Index].Count > 0))
   {
      if ((Info[Index].R1 == Region1) && (Info[Index].R2 == Region2))
      {
	 Info[Index].Count++;
	 Info[Index].Grad += Grad;
	 Index = MAXR;
      }
      else
	 Index++;
   }

   /* Add new entry if needed */
   if (Index < MAXR)
   {
      Info[Index].R1 = Region1;
      Info[Index].R2 = Region2;
      Info[Index].Count = 1;
      Info[Index].Grad = Grad;
      TotalCount++;
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine combines watersheds based on edge strengths.       */
/*---------------------------------------------------------------------------*/
combine(SHORT_TYPE ** Region, FLOAT_TYPE ** Grad, float Threshold)
{
   int i, x, y;
   int Count, Region1;

   /* Clear adjacent region data structure */
   TotalCount = 0;
   for (i = 0; i < HighIndex; i++)
   {
      Info[i].Count = Info[i].R1 = Info[i].R2 = 0;
      Info[i].Grad = 0.0;
      Parent[i] = -1;
   }

   /* Save adjacent region edge strengths */
   for (y = 1; y < Ydim; y++)
      for (x = 1; x < Xdim; x++)
      {
	 if (Region[y][x] != Region[y - 1][x])
	    save(Region[y][x], Region[y - 1][x], Grad[y][x]);
	 if (Region[y][x] != Region[y][x - 1])
	    save(Region[y][x], Region[y][x - 1], Grad[y][x]);
      }

   /* Merge adjacent regions */
   Count = 0;
   if (Method == MERGE_MEAN)
   {
      for (i = 0; i < HighIndex; i++)
	 if ((Info[i].Count > 0) &&
	     (meandiff(Info[i].R1, Info[i].R2) < Threshold))
	 {
	    merge(Info[i].R1, Info[i].R2);
	    Count++;
	 }
   }
   else if (Method == MERGE_VAR)
   {
      for (i = 0; i < HighIndex; i++)
	 if ((Info[i].Count > 0) &&
	     (variance(Info[i].R1, Info[i].R2) < Threshold))
	 {
	    merge(Info[i].R1, Info[i].R2);
	    Count++;
	 }
   }
   else if (Method == MERGE_GRAD)
   {
      for (i = 0; i < HighIndex; i++)
	 if ((Info[i].Count > 5) &&
	     ((Info[i].Grad / Info[i].Count) < Threshold))
	 {
	    merge(Info[i].R1, Info[i].R2);
	    Count++;
	 }
   }

   /* Print debug info */
   if (Debug == TRUE)
      printf("Merged %d of %d adjacent regions\n", Count, TotalCount);

   /* Renumber watershed regions */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 Region1 = Region[y][x];
	 while (Parent[Region1] > 0)
	    Region1 = Parent[Region1];
	 Region[y][x] = Region1;
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine merges two regions.                                */
/*---------------------------------------------------------------------------*/
merge(int Region1, int Region2)
{
   /* Find parents of regions */
   while (Parent[Region1] > 0)
      Region1 = Parent[Region1];
   while (Parent[Region2] > 0)
      Region2 = Parent[Region2];

   /* Adjust parent pointers and statistics */
   if (Region1 < Region2)
   {
      Parent[Region2] = Region1;
      Size[Region1] += Size[Region2];
      SumA1[Region1] += SumA1[Region2];
      SumA2[Region1] += SumA2[Region2];
      SumB1[Region1] += SumB1[Region2];
      SumB2[Region1] += SumB2[Region2];
      SumC1[Region1] += SumC1[Region2];
      SumC2[Region1] += SumC2[Region2];
   }
   else if (Region1 > Region2)
   {
      Parent[Region1] = Region2;
      Size[Region2] += Size[Region1];
      SumA1[Region2] += SumA1[Region1];
      SumA2[Region2] += SumA2[Region1];
      SumB1[Region2] += SumB1[Region1];
      SumB2[Region2] += SumB2[Region1];
      SumC1[Region2] += SumC1[Region1];
      SumC2[Region2] += SumC2[Region1];
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine displays watersheds based on average intensity.    */
/*---------------------------------------------------------------------------*/
display(SHORT_TYPE ** Region, FLOAT_TYPE *** Out)
{
   int x, y, r;

   /* Calculate average intensity in each region */
   for (y = 0; y < Ydim; y++)
      for (x = 0; x < Xdim; x++)
      {
	 r = Region[y][x];
	 if (Size[r] > 0)
	 {
	    Out[0][y][x] = SumA1[r] / Size[r];
	    Out[1][y][x] = SumB1[r] / Size[r];
	    Out[2][y][x] = SumC1[r] / Size[r];
	 }
	 else
	    Out[0][y][x] = Out[1][y][x] = Out[2][y][x] = 0;
      }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;
   FLOAT_TYPE ***Data0;
   FLOAT_TYPE ***Data1;
   SHORT_TYPE **Data2;
   FLOAT_TYPE **Data3;
   int PixType, DimCnt;

   /* Program variables */
   int i = 0;

   /* Interpret program options */
   printf("WATER_MERGE Program - KUIM Version 3.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      case 'm':
	 Method = MERGE_MEAN;
	 break;
      case 'g':
	 Method = MERGE_GRAD;
	 break;
      case 'v':
	 Method = MERGE_VAR;
	 break;
      case 'a':
	 if (sscanf(argv[++i], "%f", &Alpha) == 0)
	    Error("Could not get float argument");
	 break;
      case 's':
	 if (sscanf(argv[++i], "%d", &NumSmooth) == 0)
	    Error("Could not get integer argument");
	 break;
      case 'n':
	 if (sscanf(argv[++i], "%d", &NumMerge) == 0)
	    Error("Could not get integer argument");
	 break;
      case 't':
	 if (sscanf(argv[++i], "%f", &Threshold) == 0)
	    Error("Could not get float argument");
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 2)
   {
      fprintf(stderr, "Usage: water_merge [options] infile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      fprintf(stderr, "       [-m]    Merge based on difference in region means\n");
      fprintf(stderr, "       [-g]    Merge based on gradient along regions\n");
      fprintf(stderr, "       [-v]    Merge based on combined region variance\n");
      fprintf(stderr, "       [-a #]  Neighborhood smoothing weight (%f)\n", Alpha);
      fprintf(stderr, "       [-s #]  Number of smoothing iterations (%d)\n", NumSmooth);
      fprintf(stderr, "       [-n #]  Number of merging iterations (%d)\n", NumMerge);
      fprintf(stderr, "       [-t #]  Threshold for region merging (%f)\n", Threshold);
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get output file name");

   /* Read input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt != 3)
      Error("Expecting 3D images");
   Data0 = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
   Data1 = (FLOAT_TYPE ***) im_alloc3D(Image1, FLOAT);
   im_read(Image1, FLOAT, (char *) &(Data0[0][0][0]));

   /* Create output image */
   Image2 = im_create(Name2, SHORT, Xdim, Ydim, Zdim);
   Data2 = (SHORT_TYPE **) im_alloc2D(Image2, SHORT);
   Data3 = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);

   /* Calculate initial watersheds */
   presmooth(Data0, Data1);
   gradient(Data1, Data3);
   watershed(Data1, Data3, Data2);
   HighIndex = 10 * MinCount + 10;
   if (Method == MERGE_VAR)
      Threshold *= Threshold;

   /* Perform iterative smoothing and merging */
   for (i = 1; i <= NumMerge; i++)
   {
      if (Method == MERGE_GRAD)
      {
	 smooth(Data0, Data1, Data2, Data3);
	 gradient(Data1, Data3);
      }
      combine(Data2, Data3, i * Threshold / NumMerge);
      if (Debug == TRUE)
	 printf("Loop %d Threshold %f\n", i, i * Threshold / NumMerge);
   }

   /* Write information to output image */
   display(Data2, Data1);
   im_write(Image2, FLOAT, (char *) &(Data1[0][0][0]));
   im_free3D((char ***) Data0);
   im_free3D((char ***) Data1);
   im_free2D((char **) Data2);
   im_free2D((char **) Data3);
   exit(0);
}
