/*---------------------------------------------------------------------------*/
/* Program:  xdeform.c                                                       */
/*                                                                           */
/* Purpose:  This program implements interactive boundary fitting using      */
/*           a deformable model.                                             */
/*                                                                           */
/* Author:   John Gauch                                                      */
/*                                                                           */
/* Date:     April 25, 1994 - Original program.                              */
/*           September 14, 1994 - Version 2.                                 */
/*           November 3, 1994 - Version 3.                                   */
/*           May 30, 1995 - Version 4.                                       */
/*                                                                           */
/* Note:     Copyright (C) The University of Kansas, 1994-1995               */
/*---------------------------------------------------------------------------*/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <IM.h>

/* Global variables */
BYTE_TYPE **data1;
FLOAT_TYPE **attract;
FLOAT_TYPE **contour;
FLOAT_TYPE **temp;
BYTE_TYPE ***Data1;
FLOAT_TYPE ***Attract;
FLOAT_TYPE ***Contour;
FLOAT_TYPE ***Temp;
int Xdim, Ydim, Zdim;
int Pcnt, Scnt, Pmax = 1000;
int Debug = FALSE;
int FirstPoint = TRUE;
int Expanded = FALSE;
float alpha = 0.2;
float beta = 0.1;
float delta = -10;
float step = 0.2;
int num_iter = 100;

/* Function prototypes */
void PlotLine3D(int x1, int y1, int z1, int x2, int y2, int z2);
void PlotLine2D(int x1, int y1, int x2, int y2);

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine deforms a 2D boundary.                             */
/*---------------------------------------------------------------------------*/
void Deform2D(int num_iter)
{
   int iteration, index;
   int X, Y, L1, L2, R1, R2;
   int pos, length, NewCnt;
   float Gx, Xuu, Xuuuu, Xerr;
   float Gy, Yuu, Yuuuu, Yerr;
   float *Xold, *Yold, *Xnew, *Ynew;
   float dx, dy;

   /* Set up pointers */
   Xold = contour[0];
   Yold = contour[1];
   Xnew = temp[0];
   Ynew = temp[1];

   /* Linearly interpolate between initial points */
   if (Expanded == FALSE)
   {
      Expanded = TRUE;
      NewCnt = 0;
      for (index = 1; index < Pcnt; index++)
      {
	 dx = Xold[index] - Xold[index - 1];
	 dy = Yold[index] - Yold[index - 1];
	 length = (int) sqrt((double) (dx * dx + dy * dy));
	 for (pos = 0; pos < length; pos = pos + 3)
	 {
	    Xnew[NewCnt] = Xold[index - 1] + dx * pos / length;
	    Ynew[NewCnt] = Yold[index - 1] + dy * pos / length;
	    NewCnt++;
	 }
      }
      if (Pcnt > 0)
      {
	 dx = Xold[0] - Xold[Pcnt - 1];
	 dy = Yold[0] - Yold[Pcnt - 1];
	 length = (int) sqrt((double) (dx * dx + dy * dy));
	 for (pos = 0; pos < length; pos = pos + 3)
	 {
	    Xnew[NewCnt] = Xold[Pcnt - 1] + dx * pos / length;
	    Ynew[NewCnt] = Yold[Pcnt - 1] + dy * pos / length;
	    NewCnt++;
	 }
	 Pcnt = NewCnt;
      }

      /* Copy boundary */
      contour[0][Pmax - 1] = (float) NewCnt;
      for (index = 0; index < Pcnt; index++)
      {
	 Xold[index] = Xnew[index];
	 Yold[index] = Ynew[index];
      }
   }

   /* Deform object boundary */
   for (iteration = 0; iteration < num_iter; iteration++)
   {
      /* Deform boundary */
      for (index = 0; index < Pcnt; index++)
      {
	 /* Set up indices */
	 L1 = (index + Pcnt - 1) % Pcnt;
	 L2 = (index + Pcnt - 2) % Pcnt;
	 R1 = (index + Pcnt + 1) % Pcnt;
	 R2 = (index + Pcnt + 2) % Pcnt;

	 /* Calculate gradient of attract function */
	 X = (int) Xold[index];
	 Y = (int) Yold[index];
	 Gx = attract[Y][X] - attract[Y][X - 1];
	 Gy = attract[Y][X] - attract[Y - 1][X];

	 /* Update x,y coordinates */
	 Xuu = Xold[L1] - 2 * Xold[index] + Xold[R1];
	 Yuu = Yold[L1] - 2 * Yold[index] + Yold[R1];
	 Xuuuu = Xold[L2] - 4 * Xold[L1] + 6 * Xold[index] - 4 * Xold[R1] + Xold[R2];
	 Yuuuu = Yold[L2] - 4 * Yold[L1] + 6 * Yold[index] - 4 * Yold[R1] + Yold[R2];
	 Xerr = delta * Gx - alpha * Xuu + beta * Xuuuu;
	 Yerr = delta * Gy - alpha * Yuu + beta * Yuuuu;
	 Xnew[index] = Xold[index] - step * Xerr;
	 Ynew[index] = Yold[index] - step * Yerr;

	 /* Check for range errors */
	 if (Xnew[index] < 1)
	    Xnew[index] = 1.0;
	 if (Ynew[index] < 1)
	    Ynew[index] = 1.0;
	 if (Xnew[index] > Xdim - 2)
	    Xnew[index] = Xdim - 2;
	 if (Ynew[index] > Ydim - 2)
	    Ynew[index] = Ydim - 2;
      }

      /* Copy boundary */
      for (index = 0; index < Pcnt; index++)
      {
	 Xold[index] = Xnew[index];
	 Yold[index] = Ynew[index];
      }
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine interps a 3D boundary.                             */
/*---------------------------------------------------------------------------*/
void Interp3D()
{
   int index, pos, slice, NewCnt, length;
   int firstslice, lastslice, firstcnt, lastcnt;
   float **Xold, **Yold, **Xnew, **Ynew;
   float dx, dy;
   int Align, BestAlign;
   float Error, BestError;

   /* Set up pointers */
   Xold = Contour[0];
   Yold = Contour[1];
   Xnew = Temp[0];
   Ynew = Temp[1];

   /* Linearly interpolate between initial points */
   if (Expanded == FALSE)
      for (slice = 0; slice < Scnt; slice++)
      {
	 Expanded = TRUE;
	 NewCnt = 0;
	 Pcnt = (int) Contour[0][slice][Pmax - 1];
	 for (index = 1; index < Pcnt; index++)
	 {
	    dx = Xold[slice][index] - Xold[slice][index - 1];
	    dy = Yold[slice][index] - Yold[slice][index - 1];
	    length = (int) sqrt((double) (dx * dx + dy * dy));
	    for (pos = 0; pos < length; pos = pos + 3)
	    {
	       Xnew[slice][NewCnt] = Xold[slice][index - 1] + dx * pos / length;
	       Ynew[slice][NewCnt] = Yold[slice][index - 1] + dy * pos / length;
	       NewCnt++;
	    }
	 }
	 if (Pcnt > 0)
	 {
	    dx = Xold[slice][0] - Xold[slice][Pcnt - 1];
	    dy = Yold[slice][0] - Yold[slice][Pcnt - 1];
	    length = (int) sqrt((double) (dx * dx + dy * dy));
	    for (pos = 0; pos < length; pos = pos + 3)
	    {
	       Xnew[slice][NewCnt] = Xold[slice][Pcnt - 1] + dx * pos / length;
	       Ynew[slice][NewCnt] = Yold[slice][Pcnt - 1] + dy * pos / length;
	       NewCnt++;
	    }
	    Pcnt = NewCnt;
	    Contour[0][slice][Pmax - 1] = (float) NewCnt;
	 }

	 /* Copy boundary */
	 for (index = 0; index < Pcnt; index++)
	 {
	    Xold[slice][index] = Xnew[slice][index];
	    Yold[slice][index] = Ynew[slice][index];
	 }
      }

   /* Find contour size */
   firstslice = 0;
   while (Contour[0][firstslice][Pmax - 1] < 1)
      firstslice++;
   lastslice = Scnt - 1;
   while (Contour[0][lastslice][Pmax - 1] < 1)
      lastslice--;
   firstcnt = Contour[0][firstslice][Pmax - 1];
   lastcnt = Contour[0][lastslice][Pmax - 1];

   /* Scale up the smaller contour */
   if (lastcnt < firstcnt)
   {
      slice = lastslice;
      Contour[0][lastslice][Pmax - 1] = Pcnt = firstcnt;
      for (index = 0; index < Pcnt; index++)
      {
	 Xnew[lastslice][index] = Xold[lastslice][index * lastcnt / firstcnt];
	 Ynew[lastslice][index] = Yold[lastslice][index * lastcnt / firstcnt];
      }
   }
   else
   {
      slice = firstslice;
      Contour[0][firstslice][Pmax - 1] = Pcnt = lastcnt;
      for (index = 0; index < Pcnt; index++)
      {
	 Xnew[firstslice][index] = Xold[firstslice][index * firstcnt / lastcnt];
	 Ynew[firstslice][index] = Yold[firstslice][index * firstcnt / lastcnt];
      }
   }

   /* Copy data back again */
   for (index = 0; index < Pcnt; index++)
   {
      Xold[slice][index] = Xnew[slice][index];
      Yold[slice][index] = Ynew[slice][index];
   }

   /* Calculate initial alignment error */
   BestAlign = 0;
   BestError = 0.0;
   for (index = 0; index < Pcnt; index++)
   {
      dx = Xold[firstslice][index] - Xold[lastslice][index];
      dy = Yold[firstslice][index] - Yold[lastslice][index];
      BestError += (float) sqrt((double) (dx * dx + dy * dy));
   }

   /* Search for better alignments */
   for (Align = 1; Align < Pcnt; Align++)
   {
      Error = 0.0;
      for (index = 0; index < Pcnt; index++)
      {
	 dx = Xold[firstslice][index] - Xold[lastslice][(index + Align) % Pcnt];
	 dy = Yold[firstslice][index] - Yold[lastslice][(index + Align) % Pcnt];
	 Error += (float) sqrt((double) (dx * dx + dy * dy));
      }
      if (Error < BestError)
      {
	 BestError = Error;
	 BestAlign = Align;
      }
   }

   /* Align data */
   for (index = 0; index < Pcnt; index++)
   {
      Xnew[lastslice][index] = Xold[lastslice][(index + BestAlign) % Pcnt];
      Ynew[lastslice][index] = Yold[lastslice][(index + BestAlign) % Pcnt];
   }
   for (index = 0; index < Pcnt; index++)
   {
      Xold[lastslice][index] = Xnew[lastslice][index];
      Yold[lastslice][index] = Ynew[lastslice][index];
   }

   /* fill in intermediate contours */
   for (slice = firstslice + 1; slice < lastslice; slice++)
      for (index = 0; index < Pcnt; index++)
      {
	 Xold[slice][index] = (Xold[firstslice][index] * (lastslice - slice)
			    + Xold[lastslice][index] * (slice - firstslice))
	    / (lastslice - firstslice);
	 Yold[slice][index] = (Yold[firstslice][index] * (lastslice - slice)
			    + Yold[lastslice][index] * (slice - firstslice))
	    / (lastslice - firstslice);
	 Contour[0][slice][Pmax - 1] = Pcnt;
      }
}


/*---------------------------------------------------------------------------*/
/* Purpose:  This routine deforms a 3D boundary.                             */
/*---------------------------------------------------------------------------*/
void Deform3D(int num_iter)
{
   int iteration, u, v, X, Y, Z, L1, L2, R1, R2, U1, U2, D1, D2;
   int firstslice, lastslice;
   float Gx, Xuu, Xvv, Xuuuu, Xuuvv, Xvvvv, Xerr;
   float Gy, Yuu, Yvv, Yuuuu, Yuuvv, Yvvvv, Yerr;
   float Gz, Zuu, Zvv, Zuuuu, Zuuvv, Zvvvv, Zerr;
   float **Xold, **Yold, **Zold, **Xnew, **Ynew, **Znew;

   /* Set up pointers */
   Xold = Contour[0];
   Yold = Contour[1];
   Zold = Contour[2];
   Xnew = Temp[0];
   Ynew = Temp[1];
   Znew = Temp[2];

   /* Get range to deform */
   firstslice = 0;
   while (Contour[0][firstslice][Pmax - 1] < 1)
      firstslice++;
   lastslice = Scnt - 1;
   while (Contour[0][lastslice][Pmax - 1] < 1)
      lastslice--;

   /* Deform object boundary */
   for (iteration = 0; iteration < num_iter; iteration++)
   {
      /* Deform boundary */
      for (v = firstslice; v <= lastslice; v++)
	 for (u = 0; u < Pcnt; u++)
	 {
	    /* Set up indices */
	    L1 = (u + Pcnt - 1) % Pcnt;
	    L2 = (u + Pcnt - 2) % Pcnt;
	    R1 = (u + Pcnt + 1) % Pcnt;
	    R2 = (u + Pcnt + 2) % Pcnt;
	    U1 = (v < firstslice + 1) ? firstslice : v - 1;
	    U2 = (v < firstslice + 2) ? firstslice : v - 2;
	    D1 = (v > lastslice - 1) ? lastslice : v + 1;
	    D2 = (v > lastslice - 2) ? lastslice : v + 2;

	    /* Calculate gradient of attract function */
	    X = (int) Xold[v][u];
	    Y = (int) Yold[v][u];
	    Z = (int) Zold[v][u];
	    Gx = Gy = Gz = 0.0;
	    if (X > 0 && Y > 0 && X < Xdim && Y < Ydim)
	    {
	       Gx = Attract[Z][Y][X] - Attract[Z][Y][X - 1];
	       Gy = Attract[Z][Y][X] - Attract[Z][Y - 1][X];
	    }

	    /* Calculate surface derivatives */
	    Xuu = Xold[v][L1] - 2 * Xold[v][u] + Xold[v][R1];
	    Xvv = Xold[U1][u] - 2 * Xold[v][u] + Xold[D1][u];
	    Xuuuu = Xold[v][L2] - 4 * Xold[v][L1] + 6 * Xold[v][u] - 4 * Xold[v][R1] + Xold[v][R2];
	    Xvvvv = Xold[U2][u] - 4 * Xold[U1][u] + 6 * Xold[v][u] - 4 * Xold[D1][u] + Xold[D2][u];
	    Xuuvv = Xold[U1][L1] - 2 * Xold[U1][u] + Xold[U1][R1]
	       - 2 * Xold[v][L1] + 4 * Xold[v][u] - 2 * Xold[v][R1]
	       + Xold[D1][L1] - 2 * Xold[D1][u] + Xold[D1][R1];
	    Yuu = Yold[v][L1] - 2 * Yold[v][u] + Yold[v][R1];
	    Yvv = Yold[U1][u] - 2 * Yold[v][u] + Yold[D1][u];
	    Yuuuu = Yold[v][L2] - 4 * Yold[v][L1] + 6 * Yold[v][u] - 4 * Yold[v][R1] + Yold[v][R2];
	    Yvvvv = Yold[U2][u] - 4 * Yold[U1][u] + 6 * Yold[v][u] - 4 * Yold[D1][u] + Yold[D2][u];
	    Yuuvv = Yold[U1][L1] - 2 * Yold[U1][u] + Yold[U1][R1]
	       - 2 * Yold[v][L1] + 4 * Yold[v][u] - 2 * Yold[v][R1]
	       + Yold[D1][L1] - 2 * Yold[D1][u] + Yold[D1][R1];

	    /* Update x,y coordinates */
	    /* Xvv = Yvv = Xuuvv = Yuuvv = Xvvvv = Yvvvv = 0.0; */
	    Xerr = delta * Gx - alpha * (Xuu + Xvv) + beta * (Xuuuu + 2 * Xuuvv + Xvvvv);
	    Yerr = delta * Gy - alpha * (Yuu + Yvv) + beta * (Yuuuu + 2 * Yuuvv + Yvvvv);
	    Xnew[v][u] = Xold[v][u] - step * Xerr;
	    Ynew[v][u] = Yold[v][u] - step * Yerr;

	    /* Check for range errors */
	    if (Xnew[v][u] < 1)
	       Xnew[v][u] = 1.0;
	    if (Ynew[v][u] < 1)
	       Ynew[v][u] = 1.0;
	    if (Xnew[v][u] > Xdim - 2)
	       Xnew[v][u] = Xdim - 2;
	    if (Ynew[v][u] > Ydim - 2)
	       Ynew[v][u] = Ydim - 2;
	 }

      /* Copy boundary */
      for (v = 0; v < Scnt; v++)
	 for (u = 0; u < Pcnt; u++)
	 {
	    Xold[v][u] = Xnew[v][u];
	    Yold[v][u] = Ynew[v][u];
	 }
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose: This routine displays a 2D image.                                */
/*---------------------------------------------------------------------------*/
void Display2D(char *Name)
{
   Display *display;
   char *display_name = NULL;
   int screen_num;
   int depth;
   unsigned long foreground;
   unsigned long background;
   Window window;
   GC gc;
   Visual *visual;
   Colormap old_colormap;
   Colormap new_colormap;
   int i;
   XColor color;
   double Gamma = 0.01;
   XImage *ximage;
   int done = FALSE;
   XEvent event;
   char text[50];
   KeySym key;
   float *X, *Y;

   /* Set up window for display */
   if ((display = XOpenDisplay(display_name)) == NULL)
      Error("Can not connect to X server");
   screen_num = DefaultScreen(display);
   depth = DefaultDepth(display, screen_num);
   if (depth != 8)
      Error("Can only display on 8 bit displays");
   background = BlackPixel(display, screen_num);
   foreground = WhitePixel(display, screen_num);
   window = XCreateSimpleWindow(display, DefaultRootWindow(display),
			   200, 200, Xdim, Ydim, 5, foreground, background);
   XStoreName(display, window, Name);
   XSelectInput(display, window, ButtonPressMask | KeyPressMask | ExposureMask);
   gc = XCreateGC(display, window, 0, 0);
   XSetBackground(display, gc, background);
   XSetForeground(display, gc, foreground);
   visual = XDefaultVisualOfScreen(DefaultScreenOfDisplay(display));
   XMapRaised(display, window);

   /* Set up colormap */
   old_colormap = DefaultColormap(display, screen_num);
   new_colormap = XCreateColormap(display, window, visual, AllocAll);
   for (i = 0; i <= 255; i++)
   {
      color.pixel = i;
      XQueryColor(display, old_colormap, &color);
      color.flags = DoRed | DoGreen | DoBlue;
      XStoreColor(display, new_colormap, &color);
   }
   for (i = 50; i <= 250; i++)
   {
      color.pixel = i;
      color.blue = color.green = color.red = (int) (65535
				    * log((double) (1.0 + Gamma * (i - 50)))
				       / log((double) (1.0 + Gamma * 200)));
      XStoreColor(display, new_colormap, &color);
   }

   /* Add line color */
   color.pixel = 49;
   color.red = 65535;
   color.blue = 0;
   color.green = 0;
   XStoreColor(display, new_colormap, &color);

   /* Copy image to display window */
   XSetWindowColormap(display, window, new_colormap);
   ximage = XCreateImage(display, visual, 8, ZPixmap, 0,
			 &(data1[0][0]), Xdim, Ydim, 8, 0);
   XSetForeground(display, gc, 49);

   /* Set up pointers */
   X = contour[0];
   Y = contour[1];

   /* Loop handling events */
   while (done == FALSE)
   {
      XNextEvent(display, &event);
      switch (event.type)
      {
      case Expose:
	 if (event.xexpose.count == 0)
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	 for (i = 1; i < Pcnt; i++)
	    XDrawLine(display, window, gc,
		    (int) X[i], (int) Y[i], (int) X[i - 1], (int) Y[i - 1]);
	 if ((FirstPoint == TRUE) && (Pcnt > 0))
	    XDrawLine(display, window, gc,
	      (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	 break;

      case MappingNotify:
	 XRefreshKeyboardMapping((XMappingEvent *) & event);
	 break;

      case ButtonPress:
	 /* Handle first button press */
	 if ((event.xbutton.button == 1) && (FirstPoint == TRUE))
	 {
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	    FirstPoint = FALSE;
	    Pcnt = 0;
	    X[Pcnt] = (float) event.xbutton.x;
	    Y[Pcnt] = (float) event.xbutton.y;
	    Pcnt++;
	 }

	 /* Handle subsequent button presses */
	 else if (event.xbutton.button == 1)
	 {
	    X[Pcnt] = (float) event.xbutton.x;
	    Y[Pcnt] = (float) event.xbutton.y;
	    XDrawLine(display, window, gc,
		      (int) X[Pcnt], (int) Y[Pcnt], (int) X[Pcnt - 1], (int) Y[Pcnt - 1]);
	    Pcnt++;
	 }

	 /* Handle last button press */
	 else if (event.xbutton.button > 1)
	 {
	    Expanded = FALSE;
	    FirstPoint = TRUE;
	    XDrawLine(display, window, gc,
	      (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	 }
	 break;

      case KeyPress:
	 /* Exit if 'q' was pressed */
	 i = XLookupString((XKeyEvent *) & event, text, 50, &key, 0);
	 if (i == 1 && text[0] == 'q')
	    done = TRUE;

	 /* Modify iterations if 'i' was pressed */
	 if (i == 1 && text[0] == 'i')
	 {
	    printf("Current iteration count: %d\n", num_iter);
	    printf("Enter new iteration count: ");
	    scanf("%d", &num_iter);
	    if (num_iter < 0)
	       num_iter = 0;
	 }

	 /* Modify weights if 'w' was pressed */
	 if (i == 1 && text[0] == 'w')
	 {
	    printf("Current alpha, beta and delta weights: %f %f %f\n",
		   alpha, beta, delta);
	    printf("Enter new alpha, beta and delta weights: ");
	    scanf("%f %f %f", &alpha, &beta, &delta);
	    if (alpha > 5)
	       alpha = 5;
	    if (beta > 1.25)
	       beta = 1.25;
	 }

	 /* Start deforming if 'd' was pressed */
	 if ((i == 1 && text[0] == 'd') &&
	     (FirstPoint == TRUE) && (Pcnt > 0))
	 {
	    Deform2D(num_iter);
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	    for (i = 1; i < Pcnt; i++)
	       XDrawLine(display, window, gc,
		    (int) X[i], (int) Y[i], (int) X[i - 1], (int) Y[i - 1]);
	    XDrawLine(display, window, gc,
	      (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	 }
	 break;
      }
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose: This routine displays a 3D image.                                */
/*---------------------------------------------------------------------------*/
void Display3D(char *Name)
{
   Display *display;
   char *display_name = NULL;
   int screen_num;
   int depth;
   unsigned long foreground;
   unsigned long background;
   Window window;
   GC gc;
   Visual *visual;
   Colormap old_colormap;
   Colormap new_colormap;
   int i;
   XColor color;
   double Gamma = 0.01;
   XImage *ximage;
   int done = FALSE;
   XEvent event;
   char text[50];
   KeySym key;
   int slice = 0;
   float *X, *Y;

   /* Set up window for display */
   if ((display = XOpenDisplay(display_name)) == NULL)
      Error("Can not connect to X server");
   screen_num = DefaultScreen(display);
   depth = DefaultDepth(display, screen_num);
   if (depth != 8)
      Error("Can only display on 8 bit displays");
   background = BlackPixel(display, screen_num);
   foreground = WhitePixel(display, screen_num);
   window = XCreateSimpleWindow(display, DefaultRootWindow(display),
			   200, 200, Xdim, Ydim, 5, foreground, background);
   XStoreName(display, window, Name);
   XSelectInput(display, window, ButtonPressMask | KeyPressMask | ExposureMask);
   gc = XCreateGC(display, window, 0, 0);
   XSetBackground(display, gc, background);
   XSetForeground(display, gc, foreground);
   visual = XDefaultVisualOfScreen(DefaultScreenOfDisplay(display));
   XMapRaised(display, window);

   /* Set up colormap */
   old_colormap = DefaultColormap(display, screen_num);
   new_colormap = XCreateColormap(display, window, visual, AllocAll);
   for (i = 0; i <= 255; i++)
   {
      color.pixel = i;
      XQueryColor(display, old_colormap, &color);
      color.flags = DoRed | DoGreen | DoBlue;
      XStoreColor(display, new_colormap, &color);
   }
   for (i = 50; i <= 250; i++)
   {
      color.pixel = i;
      color.blue = color.green = color.red = (int) (65535
				    * log((double) (1.0 + Gamma * (i - 50)))
				       / log((double) (1.0 + Gamma * 200)));
      XStoreColor(display, new_colormap, &color);
   }

   /* Add line color */
   color.pixel = 49;
   color.red = 65535;
   color.blue = 0;
   color.green = 0;
   XStoreColor(display, new_colormap, &color);

   /* Copy first slice to display window */
   XSetWindowColormap(display, window, new_colormap);
   ximage = XCreateImage(display, visual, 8, ZPixmap, 0,
			 &(Data1[0][0][0]), Xdim, Ydim, 8, 0);
   XSetForeground(display, gc, 49);

   /* Loop handling events */
   while (done == FALSE)
   {
      XNextEvent(display, &event);
      switch (event.type)
      {
      case Expose:
	 X = Contour[0][slice];
	 Y = Contour[1][slice];
	 Pcnt = Contour[0][slice][Pmax - 1];
	 if (event.xexpose.count == 0)
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	 for (i = 1; i < Pcnt; i++)
	    XDrawLine(display, window, gc,
		    (int) X[i], (int) Y[i], (int) X[i - 1], (int) Y[i - 1]);
	 if ((FirstPoint == TRUE) && (Pcnt > 0))
	    XDrawLine(display, window, gc,
	      (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	 break;

      case MappingNotify:
	 XRefreshKeyboardMapping((XMappingEvent *) & event);
	 break;

      case ButtonPress:
	 /* Handle first button press */
	 if ((event.xbutton.button == 1) && (FirstPoint == TRUE))
	 {
	    X = Contour[0][slice];
	    Y = Contour[1][slice];
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	    FirstPoint = FALSE;
	    Pcnt = 0;
	    X[Pcnt] = (float) event.xbutton.x;
	    Y[Pcnt] = (float) event.xbutton.y;
	    Pcnt++;
	 }

	 /* Handle subsequent button presses */
	 else if (event.xbutton.button == 1)
	 {
	    X[Pcnt] = (float) event.xbutton.x;
	    Y[Pcnt] = (float) event.xbutton.y;
	    XDrawLine(display, window, gc,
		      (int) X[Pcnt], (int) Y[Pcnt], (int) X[Pcnt - 1], (int) Y[Pcnt - 1]);
	    Pcnt++;
	 }

	 /* Handle last button press */
	 else if (event.xbutton.button > 1)
	 {
	    Expanded = FALSE;
	    FirstPoint = TRUE;
	    XDrawLine(display, window, gc,
	      (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	    Contour[0][slice][Pmax - 1] = Pcnt;
	 }
	 break;

      case KeyPress:
	 /* Exit if 'q' was pressed */
	 i = XLookupString((XKeyEvent *) & event, text, 50, &key, 0);
	 if (i == 1 && text[0] == 'q')
	    done = TRUE;

	 /* Single step forward if 'f' pressed */
	 if (i == 1 && text[0] == 'f')
	 {
	    slice = (slice + 1) % Scnt;
	    X = Contour[0][slice];
	    Y = Contour[1][slice];
	    Pcnt = Contour[0][slice][Pmax - 1];
	    ximage = XCreateImage(display, visual, 8, ZPixmap, 0,
				  &(Data1[slice][0][0]), Xdim, Ydim, 8, 0);
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	    for (i = 1; i < Pcnt; i++)
	       XDrawLine(display, window, gc,
		    (int) X[i], (int) Y[i], (int) X[i - 1], (int) Y[i - 1]);
	    if ((FirstPoint == TRUE) && (Pcnt > 0))
	       XDrawLine(display, window, gc,
	       (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	    sprintf(text, "slice %d", slice);
	    XStoreName(display, window, text);
	 }

	 /* Single step backward if 'b' pressed */
	 if (i == 1 && text[0] == 'b')
	 {
	    slice = (slice + Scnt - 1) % Scnt;
	    X = Contour[0][slice];
	    Y = Contour[1][slice];
	    Pcnt = Contour[0][slice][Pmax - 1];
	    ximage = XCreateImage(display, visual, 8, ZPixmap, 0,
				  &(Data1[slice][0][0]), Xdim, Ydim, 8, 0);
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	    for (i = 1; i < Pcnt; i++)
	       XDrawLine(display, window, gc,
		    (int) X[i], (int) Y[i], (int) X[i - 1], (int) Y[i - 1]);
	    if ((FirstPoint == TRUE) && (Pcnt > 0))
	       XDrawLine(display, window, gc,
	       (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	    sprintf(text, "slice %d", slice);
	    XStoreName(display, window, text);
	 }

	 /* Play movie forward if 'F' was pressed */
	 if (i == 1 && text[0] == 'F')
	 {
	    for (slice = 0; slice < Scnt; slice++)
	    {
	       X = Contour[0][slice];
	       Y = Contour[1][slice];
	       Pcnt = Contour[0][slice][Pmax - 1];
	       ximage = XCreateImage(display, visual, 8, ZPixmap, 0,
				   &(Data1[slice][0][0]), Xdim, Ydim, 8, 0);
	       XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	       for (i = 1; i < Pcnt; i++)
		  XDrawLine(display, window, gc,
		    (int) X[i], (int) Y[i], (int) X[i - 1], (int) Y[i - 1]);
	       if ((FirstPoint == TRUE) && (Pcnt > 0))
		  XDrawLine(display, window, gc,
			    (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	       sprintf(text, "slice %d", slice);
	       XStoreName(display, window, text);
	    }
	    slice = Scnt - 1;
	 }

	 /* Play movie backward if 'B' was pressed */
	 if (i == 1 && text[0] == 'B')
	 {
	    for (slice = (Scnt - 1); slice >= 0; slice--)
	    {
	       X = Contour[0][slice];
	       Y = Contour[1][slice];
	       Pcnt = Contour[0][slice][Pmax - 1];
	       ximage = XCreateImage(display, visual, 8, ZPixmap, 0,
				   &(Data1[slice][0][0]), Xdim, Ydim, 8, 0);
	       XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	       for (i = 1; i < Pcnt; i++)
		  XDrawLine(display, window, gc,
		    (int) X[i], (int) Y[i], (int) X[i - 1], (int) Y[i - 1]);
	       if ((FirstPoint == TRUE) && (Pcnt > 0))
		  XDrawLine(display, window, gc,
			    (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	       sprintf(text, "slice %d", slice);
	       XStoreName(display, window, text);
	    }
	    slice = 0;
	 }

	 /* Modify iterations if 'i' was pressed */
	 if (i == 1 && text[0] == 'i')
	 {
	    printf("Current iteration count: %d\n", num_iter);
	    printf("Enter new iteration count: ");
	    scanf("%d", &num_iter);
	    if (num_iter < 0)
	       num_iter = 0;
	 }

	 /* Modify weights if 'w' was pressed */
	 if (i == 1 && text[0] == 'w')
	 {
	    printf("Current alpha, beta and delta weights: %f %f %f\n",
		   alpha, beta, delta);
	    printf("Enter new alpha, beta and delta weights: ");
	    scanf("%f %f %f", &alpha, &beta, &delta);
	    if (alpha > 5)
	       alpha = 5;
	    if (beta > 1.25)
	       beta = 1.25;
	 }

	 /* Start deforming if 'd' was pressed */
	 if ((i == 1 && text[0] == 'd') &&
	     (FirstPoint == TRUE) && (Pcnt > 0))
	 {
	    Interp3D();
	    Deform3D(num_iter);
	    X = Contour[0][slice];
	    Y = Contour[1][slice];
	    Pcnt = Contour[0][slice][Pmax - 1];
	    XPutImage(display, window, gc, ximage, 0, 0, 0, 0, Xdim, Ydim);
	    for (i = 1; i < Pcnt; i++)
	       XDrawLine(display, window, gc,
		    (int) X[i], (int) Y[i], (int) X[i - 1], (int) Y[i - 1]);
	    XDrawLine(display, window, gc,
	      (int) X[Pcnt - 1], (int) Y[Pcnt - 1], (int) X[0], (int) Y[0]);
	 }
	 break;
      }
   }
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This is the main program.                                       */
/*---------------------------------------------------------------------------*/
main(int argc, char *argv[])
{
   /* Image variables */
   char Name1[50];
   char Name2[50];
   char Name3[50];
   char Name4[50];
   IM_TYPE *Image1;
   IM_TYPE *Image2;
   IM_TYPE *Image3;

   /* Program variables */
   int PixType, DimCnt;
   int Min, Max;
   int i = 0, x, y, z, u, v;

   /* Interpret program options */
   printf("XDEFORM Program - KUIM Version 4.0\n\n");
   while ((++i < argc) && (argv[i][0] == '-'))
      switch (argv[i][1])
      {
      case 'd':
	 Debug = TRUE;
	 break;
      default:
	 Error("Invalid option encountered");
	 break;
      }

   /* Check number of file names */
   if (argc - i != 3)
   {
      fprintf(stderr, "Usage: xdeform [options] infile attractfile outfile\n");
      fprintf(stderr, "       [-d]    Print debugging information\n");
      exit(1);
   }

   /* Get image file names from argument list */
   if (sscanf(argv[i++], "%s", Name1) == 0)
      Error("Could not get input file name");
   if (sscanf(argv[i++], "%s", Name2) == 0)
      Error("Could not get attract file name");
   if (sscanf(argv[i++], "%s", Name3) == 0)
      Error("Could not get output file name");

   /* Open input image */
   Image1 = im_open(Name1, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
   if (DimCnt == 1)
      Error("Can not handle 1D images");

   /* Handle 2D images */
   else if (DimCnt == 2)
   {
      /* Read input image */
      data1 = (BYTE_TYPE **) im_alloc2D(Image1, BYTE);
      im_read(Image1, BYTE, (char *) &(data1[0][0]));

      /* Scale image to 50..250 range for display */
      Min = Max = data1[0][0];
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	    if (Min > data1[y][x])
	       Min = data1[y][x];
	    else if (Max < data1[y][x])
	       Max = data1[y][x];
      for (y = 0; y < Ydim; y++)
	 for (x = 0; x < Xdim; x++)
	    data1[y][x] = 50 + ((data1[y][x] - Min) * 200) / (Max - Min);

      /* Read attraction function image */
      Image2 = im_open(Name2, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
      if (DimCnt != 2)
	 Error("Expecting 2D attraction image");
      attract = (FLOAT_TYPE **) im_alloc2D(Image2, FLOAT);
      im_read(Image2, FLOAT, (char *) &(attract[0][0]));

      /* Create contour image */
      Pcnt = 1000;
      Image3 = im_create(Name3, FLOAT, Pcnt, 2, 1);
      contour = (FLOAT_TYPE **) im_alloc2D(Image3, FLOAT);
      temp = (FLOAT_TYPE **) im_alloc2D(Image3, FLOAT);

      /* Initialize deformable contour */
      for (u = 0; u < Pcnt; u++)
      {
	 contour[0][u] = temp[0][u] = 0.0;
	 contour[1][u] = temp[1][u] = 0.0;
      }

      /* Print instructions */
      printf("\nMouse commands:\n");
      printf("   <left button> - add points to contour\n");
      printf("   <right button> - close current contour\n");
      printf("\nKeyboard commands:\n");
      printf("   'd' - deform surface\n");
      printf("   'i' - enter new iteration count\n");
      printf("   'w' - enter new deformation weights\n");
      printf("   'q' - quit xdeform program\n");

      /* Deform image */
      Display2D(Name1);

      /* Draw lines on optional output image */
      if (Debug == TRUE)
      {
	 printf("Enter name for debug file: ");
	 scanf("%s", Name1);
	 Image1 = im_create(Name1, BYTE, Xdim, Ydim, Zdim);
	 for (u = 1; u < Pcnt; u++)
	    PlotLine2D((int) contour[0][u - 1], (int) contour[1][u - 1],
		       (int) contour[0][u], (int) contour[1][u]);
	 PlotLine2D((int) contour[0][0], (int) contour[1][0],
		    (int) contour[0][Pcnt - 1], (int) contour[1][Pcnt - 1]);
	 im_write(Image1, BYTE, (char *) &(data1[0][0]));
      }

      /* Write contour to output image */
      printf("Extract first %d points from contour\n", Pcnt);
      im_write(Image3, FLOAT, (char *) &(contour[0][0]));
      im_free2D((char **) data1);
      im_free2D((char **) attract);
      im_free2D((char **) contour);
   }

   /* Handle 3D images */
   else if (DimCnt == 3)
   {
      /* Read input image */
      Data1 = (BYTE_TYPE ***) im_alloc3D(Image1, BYTE);
      im_read(Image1, BYTE, (char *) &(Data1[0][0][0]));

      /* Scale image to 50..250 range for display */
      Min = Max = Data1[0][0][0];
      for (z = 0; z < Zdim; z++)
	 for (y = 0; y < Ydim; y++)
	    for (x = 0; x < Xdim; x++)
	       if (Min > Data1[z][y][x])
		  Min = Data1[z][y][x];
	       else if (Max < Data1[z][y][x])
		  Max = Data1[z][y][x];
      for (z = 0; z < Zdim; z++)
	 for (y = 0; y < Ydim; y++)
	    for (x = 0; x < Xdim; x++)
	       Data1[z][y][x] = 50 + ((Data1[z][y][x] - Min) * 200) / (Max - Min);

      /* Read attraction function image */
      Image2 = im_open(Name2, &PixType, &Xdim, &Ydim, &Zdim, &DimCnt);
      if (DimCnt != 3)
	 Error("Expecting 3D attraction image");
      Attract = (FLOAT_TYPE ***) im_alloc3D(Image2, FLOAT);
      im_read(Image2, FLOAT, (char *) &(Attract[0][0][0]));

      /* Create contour image */
      Pcnt = 1000;
      Scnt = Zdim;
      Image3 = im_create(Name3, FLOAT, Pcnt, Scnt, 3);
      Contour = (FLOAT_TYPE ***) im_alloc3D(Image3, FLOAT);
      Temp = (FLOAT_TYPE ***) im_alloc3D(Image3, FLOAT);

      /* Initialize deformable surface */
      for (v = 0; v < Scnt; v++)
	 for (u = 0; u < Pcnt; u++)
	 {
	    Contour[0][v][u] = Temp[0][v][u] = 0.0;
	    Contour[1][v][u] = Temp[1][v][u] = 0.0;
	    Contour[2][v][u] = Temp[2][v][u] = v;
	 }

      /* Print instructions */
      printf("\nMouse commands:\n");
      printf("   <left button> - add points to contour\n");
      printf("   <right button> - close current contour\n");
      printf("\nKeyboard commands:\n");
      printf("   'd' - deform surface\n");
      printf("   'i' - enter new iteration count\n");
      printf("   'w' - enter new deformation weights\n");
      printf("   'f' - go forward to next slice\n");
      printf("   'b' - go backward to previous slice\n");
      printf("   'F' - play sequence of images forwards\n");
      printf("   'B' - play sequence of images backwards\n");
      printf("   'q' - quit xdeform program\n");

      /* Deform image */
      Display3D(Name1);

      /* Draw lines on optional output image */
      if (Debug == TRUE)
      {
	 printf("Enter name for debug file: ");
	 scanf("%s", Name1);
	 Image1 = im_create(Name1, BYTE, Xdim, Ydim, Zdim);
	 for (v = 0; v < Scnt; v++)
	    if (Contour[0][v][Pmax - 1] > 0)
	    {
	       for (u = 1; u < Pcnt; u++)
		  PlotLine3D((int) Contour[0][v][u - 1], (int) Contour[1][v][u - 1], v,
			 (int) Contour[0][v][u], (int) Contour[1][v][u], v);
	       PlotLine3D((int) Contour[0][v][0], (int) Contour[1][v][0], v,
			  (int) Contour[0][v][Pcnt - 1], (int) Contour[1][v][Pcnt - 1], v);
	    }
	 im_write(Image1, BYTE, (char *) &(Data1[0][0][0]));
      }

      /* Write contour to output image */
      printf("Extract first %d points from surface\n", Pcnt);
      im_write(Image3, FLOAT, (char *) &(Contour[0][0][0]));
      im_free3D((char ***) Data1);
      im_free3D((char ***) Attract);
      im_free3D((char ***) Contour);
   }

   exit(0);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine plots a given line segment.                        */
/*---------------------------------------------------------------------------*/
void PlotLine2D(int x1, int y1, int x2, int y2)
{
   int dx, dy, length, i, x, y;
   int Thick = FALSE;
   int Intensity = 255;

   printf("%d %d\n", x1, y1);
   dx = x2 - x1;
   dy = y2 - y1;
   length = (int) sqrt((double) (dx * dx + dy * dy));
   for (i = 0; i < length; i++)
   {
      x = x1 + dx * i / length;
      y = y1 + dy * i / length;
      data1[y][x] = Intensity;
      if (Thick == TRUE)
      {
	 data1[y + 1][x] = Intensity;
	 data1[y][x + 1] = Intensity;
	 data1[y + 1][x + 1] = Intensity;
      }
   }
}
/*---------------------------------------------------------------------------*/
/* Purpose:  This routine plots a given line segment.                        */
/*---------------------------------------------------------------------------*/
void PlotLine3D(int x1, int y1, int z1, int x2, int y2, int z2)
{
   int dx, dy, length, i, x, y;
   int Thick = FALSE;
   int Intensity = 255;

   printf("%d %d %d\n", x1, y1, z1);
   dx = x2 - x1;
   dy = y2 - y1;
   length = (int) sqrt((double) (dx * dx + dy * dy));
   for (i = 0; i < length; i++)
   {
      x = x1 + dx * i / length;
      y = y1 + dy * i / length;
      Data1[z1][y][x] = Intensity;
      if (Thick == TRUE)
      {
	 Data1[z1][y + 1][x] = Intensity;
	 Data1[z1][y][x + 1] = Intensity;
	 Data1[z1][y + 1][x + 1] = Intensity;
      }
   }
}
